// SPDX-FileCopyrightText: None
// SPDX-License-Identifier: LGPL-2.0-or-later
// This file is auto-generated. All changes will be lost. See tools/update-emojis.py
// clang-format off
{u"waving hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44B\U0001F3FB"), u"wave_tone1"_s, u"waving hand: light skin tone"_s})},
{u"waving hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44B\U0001F3FC"), u"wave_tone2"_s, u"waving hand: medium-light skin tone"_s})},
{u"waving hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44B\U0001F3FD"), u"wave_tone3"_s, u"waving hand: medium skin tone"_s})},
{u"waving hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44B\U0001F3FE"), u"wave_tone4"_s, u"waving hand: medium-dark skin tone"_s})},
{u"waving hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44B\U0001F3FF"), u"wave_tone5"_s, u"waving hand: dark skin tone"_s})},
{u"raised back of hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91A\U0001F3FB"), u"raised_back_of_hand_tone1"_s, u"raised back of hand: light skin tone"_s})},
{u"raised back of hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91A\U0001F3FC"), u"raised_back_of_hand_tone2"_s, u"raised back of hand: medium-light skin tone"_s})},
{u"raised back of hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91A\U0001F3FD"), u"raised_back_of_hand_tone3"_s, u"raised back of hand: medium skin tone"_s})},
{u"raised back of hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91A\U0001F3FE"), u"raised_back_of_hand_tone4"_s, u"raised back of hand: medium-dark skin tone"_s})},
{u"raised back of hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91A\U0001F3FF"), u"raised_back_of_hand_tone5"_s, u"raised back of hand: dark skin tone"_s})},
{u"hand with fingers splayed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F590\U0001F3FB"), u"hand_splayed_tone1"_s, u"hand with fingers splayed: light skin tone"_s})},
{u"hand with fingers splayed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F590\U0001F3FC"), u"hand_splayed_tone2"_s, u"hand with fingers splayed: medium-light skin tone"_s})},
{u"hand with fingers splayed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F590\U0001F3FD"), u"hand_splayed_tone3"_s, u"hand with fingers splayed: medium skin tone"_s})},
{u"hand with fingers splayed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F590\U0001F3FE"), u"hand_splayed_tone4"_s, u"hand with fingers splayed: medium-dark skin tone"_s})},
{u"hand with fingers splayed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F590\U0001F3FF"), u"hand_splayed_tone5"_s, u"hand with fingers splayed: dark skin tone"_s})},
{u"raised hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270B\U0001F3FB"), u"raised_hand_tone1"_s, u"raised hand: light skin tone"_s})},
{u"raised hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270B\U0001F3FC"), u"raised_hand_tone2"_s, u"raised hand: medium-light skin tone"_s})},
{u"raised hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270B\U0001F3FD"), u"raised_hand_tone3"_s, u"raised hand: medium skin tone"_s})},
{u"raised hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270B\U0001F3FE"), u"raised_hand_tone4"_s, u"raised hand: medium-dark skin tone"_s})},
{u"raised hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270B\U0001F3FF"), u"raised_hand_tone5"_s, u"raised hand: dark skin tone"_s})},
{u"vulcan salute"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F596\U0001F3FB"), u"vulcan_tone1"_s, u"vulcan salute: light skin tone"_s})},
{u"vulcan salute"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F596\U0001F3FC"), u"vulcan_tone2"_s, u"vulcan salute: medium-light skin tone"_s})},
{u"vulcan salute"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F596\U0001F3FD"), u"vulcan_tone3"_s, u"vulcan salute: medium skin tone"_s})},
{u"vulcan salute"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F596\U0001F3FE"), u"vulcan_tone4"_s, u"vulcan salute: medium-dark skin tone"_s})},
{u"vulcan salute"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F596\U0001F3FF"), u"vulcan_tone5"_s, u"vulcan salute: dark skin tone"_s})},
{u"rightwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FB"), u"rightwards hand: light skin tone"_s, u"rightwards hand: light skin tone"_s})},
{u"rightwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FC"), u"rightwards hand: medium-light skin tone"_s, u"rightwards hand: medium-light skin tone"_s})},
{u"rightwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FD"), u"rightwards hand: medium skin tone"_s, u"rightwards hand: medium skin tone"_s})},
{u"rightwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FE"), u"rightwards hand: medium-dark skin tone"_s, u"rightwards hand: medium-dark skin tone"_s})},
{u"rightwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FF"), u"rightwards hand: dark skin tone"_s, u"rightwards hand: dark skin tone"_s})},
{u"leftwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF2\U0001F3FB"), u"leftwards hand: light skin tone"_s, u"leftwards hand: light skin tone"_s})},
{u"leftwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF2\U0001F3FC"), u"leftwards hand: medium-light skin tone"_s, u"leftwards hand: medium-light skin tone"_s})},
{u"leftwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF2\U0001F3FD"), u"leftwards hand: medium skin tone"_s, u"leftwards hand: medium skin tone"_s})},
{u"leftwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF2\U0001F3FE"), u"leftwards hand: medium-dark skin tone"_s, u"leftwards hand: medium-dark skin tone"_s})},
{u"leftwards hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF2\U0001F3FF"), u"leftwards hand: dark skin tone"_s, u"leftwards hand: dark skin tone"_s})},
{u"palm down hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF3\U0001F3FB"), u"palm down hand: light skin tone"_s, u"palm down hand: light skin tone"_s})},
{u"palm down hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF3\U0001F3FC"), u"palm down hand: medium-light skin tone"_s, u"palm down hand: medium-light skin tone"_s})},
{u"palm down hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF3\U0001F3FD"), u"palm down hand: medium skin tone"_s, u"palm down hand: medium skin tone"_s})},
{u"palm down hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF3\U0001F3FE"), u"palm down hand: medium-dark skin tone"_s, u"palm down hand: medium-dark skin tone"_s})},
{u"palm down hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF3\U0001F3FF"), u"palm down hand: dark skin tone"_s, u"palm down hand: dark skin tone"_s})},
{u"palm up hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF4\U0001F3FB"), u"palm up hand: light skin tone"_s, u"palm up hand: light skin tone"_s})},
{u"palm up hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF4\U0001F3FC"), u"palm up hand: medium-light skin tone"_s, u"palm up hand: medium-light skin tone"_s})},
{u"palm up hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF4\U0001F3FD"), u"palm up hand: medium skin tone"_s, u"palm up hand: medium skin tone"_s})},
{u"palm up hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF4\U0001F3FE"), u"palm up hand: medium-dark skin tone"_s, u"palm up hand: medium-dark skin tone"_s})},
{u"palm up hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF4\U0001F3FF"), u"palm up hand: dark skin tone"_s, u"palm up hand: dark skin tone"_s})},
{u"OK hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44C\U0001F3FB"), u"ok_hand_tone1"_s, u"OK hand: light skin tone"_s})},
{u"OK hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44C\U0001F3FC"), u"ok_hand_tone2"_s, u"OK hand: medium-light skin tone"_s})},
{u"OK hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44C\U0001F3FD"), u"ok_hand_tone3"_s, u"OK hand: medium skin tone"_s})},
{u"OK hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44C\U0001F3FE"), u"ok_hand_tone4"_s, u"OK hand: medium-dark skin tone"_s})},
{u"OK hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44C\U0001F3FF"), u"ok_hand_tone5"_s, u"OK hand: dark skin tone"_s})},
{u"pinched fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90C\U0001F3FB"), u"pinched fingers: light skin tone"_s, u"pinched fingers: light skin tone"_s})},
{u"pinched fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90C\U0001F3FC"), u"pinched fingers: medium-light skin tone"_s, u"pinched fingers: medium-light skin tone"_s})},
{u"pinched fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90C\U0001F3FD"), u"pinched fingers: medium skin tone"_s, u"pinched fingers: medium skin tone"_s})},
{u"pinched fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90C\U0001F3FE"), u"pinched fingers: medium-dark skin tone"_s, u"pinched fingers: medium-dark skin tone"_s})},
{u"pinched fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90C\U0001F3FF"), u"pinched fingers: dark skin tone"_s, u"pinched fingers: dark skin tone"_s})},
{u"pinching hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90F\U0001F3FB"), u"pinching hand: light skin tone"_s, u"pinching hand: light skin tone"_s})},
{u"pinching hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90F\U0001F3FC"), u"pinching hand: medium-light skin tone"_s, u"pinching hand: medium-light skin tone"_s})},
{u"pinching hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90F\U0001F3FD"), u"pinching hand: medium skin tone"_s, u"pinching hand: medium skin tone"_s})},
{u"pinching hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90F\U0001F3FE"), u"pinching hand: medium-dark skin tone"_s, u"pinching hand: medium-dark skin tone"_s})},
{u"pinching hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F90F\U0001F3FF"), u"pinching hand: dark skin tone"_s, u"pinching hand: dark skin tone"_s})},
{u"victory hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270C\U0001F3FB"), u"v_tone1"_s, u"victory hand: light skin tone"_s})},
{u"victory hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270C\U0001F3FC"), u"v_tone2"_s, u"victory hand: medium-light skin tone"_s})},
{u"victory hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270C\U0001F3FD"), u"v_tone3"_s, u"victory hand: medium skin tone"_s})},
{u"victory hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270C\U0001F3FE"), u"v_tone4"_s, u"victory hand: medium-dark skin tone"_s})},
{u"victory hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270C\U0001F3FF"), u"v_tone5"_s, u"victory hand: dark skin tone"_s})},
{u"crossed fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91E\U0001F3FB"), u"fingers_crossed_tone1"_s, u"crossed fingers: light skin tone"_s})},
{u"crossed fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91E\U0001F3FC"), u"fingers_crossed_tone2"_s, u"crossed fingers: medium-light skin tone"_s})},
{u"crossed fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91E\U0001F3FD"), u"fingers_crossed_tone3"_s, u"crossed fingers: medium skin tone"_s})},
{u"crossed fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91E\U0001F3FE"), u"fingers_crossed_tone4"_s, u"crossed fingers: medium-dark skin tone"_s})},
{u"crossed fingers"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91E\U0001F3FF"), u"fingers_crossed_tone5"_s, u"crossed fingers: dark skin tone"_s})},
{u"hand with index finger and thumb crossed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF0\U0001F3FB"), u"hand with index finger and thumb crossed: light skin tone"_s, u"hand with index finger and thumb crossed: light skin tone"_s})},
{u"hand with index finger and thumb crossed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF0\U0001F3FC"), u"hand with index finger and thumb crossed: medium-light skin tone"_s, u"hand with index finger and thumb crossed: medium-light skin tone"_s})},
{u"hand with index finger and thumb crossed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF0\U0001F3FD"), u"hand with index finger and thumb crossed: medium skin tone"_s, u"hand with index finger and thumb crossed: medium skin tone"_s})},
{u"hand with index finger and thumb crossed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF0\U0001F3FE"), u"hand with index finger and thumb crossed: medium-dark skin tone"_s, u"hand with index finger and thumb crossed: medium-dark skin tone"_s})},
{u"hand with index finger and thumb crossed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF0\U0001F3FF"), u"hand with index finger and thumb crossed: dark skin tone"_s, u"hand with index finger and thumb crossed: dark skin tone"_s})},
{u"love-you gesture"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91F\U0001F3FB"), u"love_you_gesture_tone1"_s, u"love-you gesture: light skin tone"_s})},
{u"love-you gesture"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91F\U0001F3FC"), u"love_you_gesture_tone2"_s, u"love-you gesture: medium-light skin tone"_s})},
{u"love-you gesture"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91F\U0001F3FD"), u"love_you_gesture_tone3"_s, u"love-you gesture: medium skin tone"_s})},
{u"love-you gesture"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91F\U0001F3FE"), u"love_you_gesture_tone4"_s, u"love-you gesture: medium-dark skin tone"_s})},
{u"love-you gesture"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91F\U0001F3FF"), u"love_you_gesture_tone5"_s, u"love-you gesture: dark skin tone"_s})},
{u"sign of the horns"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F918\U0001F3FB"), u"metal_tone1"_s, u"sign of the horns: light skin tone"_s})},
{u"sign of the horns"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F918\U0001F3FC"), u"metal_tone2"_s, u"sign of the horns: medium-light skin tone"_s})},
{u"sign of the horns"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F918\U0001F3FD"), u"metal_tone3"_s, u"sign of the horns: medium skin tone"_s})},
{u"sign of the horns"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F918\U0001F3FE"), u"metal_tone4"_s, u"sign of the horns: medium-dark skin tone"_s})},
{u"sign of the horns"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F918\U0001F3FF"), u"metal_tone5"_s, u"sign of the horns: dark skin tone"_s})},
{u"call me hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F919\U0001F3FB"), u"call_me_tone1"_s, u"call me hand: light skin tone"_s})},
{u"call me hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F919\U0001F3FC"), u"call_me_tone2"_s, u"call me hand: medium-light skin tone"_s})},
{u"call me hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F919\U0001F3FD"), u"call_me_tone3"_s, u"call me hand: medium skin tone"_s})},
{u"call me hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F919\U0001F3FE"), u"call_me_tone4"_s, u"call me hand: medium-dark skin tone"_s})},
{u"call me hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F919\U0001F3FF"), u"call_me_tone5"_s, u"call me hand: dark skin tone"_s})},
{u"backhand index pointing left"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F448\U0001F3FB"), u"point_left_tone1"_s, u"backhand index pointing left: light skin tone"_s})},
{u"backhand index pointing left"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F448\U0001F3FC"), u"point_left_tone2"_s, u"backhand index pointing left: medium-light skin tone"_s})},
{u"backhand index pointing left"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F448\U0001F3FD"), u"point_left_tone3"_s, u"backhand index pointing left: medium skin tone"_s})},
{u"backhand index pointing left"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F448\U0001F3FE"), u"point_left_tone4"_s, u"backhand index pointing left: medium-dark skin tone"_s})},
{u"backhand index pointing left"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F448\U0001F3FF"), u"point_left_tone5"_s, u"backhand index pointing left: dark skin tone"_s})},
{u"backhand index pointing right"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F449\U0001F3FB"), u"point_right_tone1"_s, u"backhand index pointing right: light skin tone"_s})},
{u"backhand index pointing right"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F449\U0001F3FC"), u"point_right_tone2"_s, u"backhand index pointing right: medium-light skin tone"_s})},
{u"backhand index pointing right"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F449\U0001F3FD"), u"point_right_tone3"_s, u"backhand index pointing right: medium skin tone"_s})},
{u"backhand index pointing right"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F449\U0001F3FE"), u"point_right_tone4"_s, u"backhand index pointing right: medium-dark skin tone"_s})},
{u"backhand index pointing right"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F449\U0001F3FF"), u"point_right_tone5"_s, u"backhand index pointing right: dark skin tone"_s})},
{u"backhand index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F446\U0001F3FB"), u"point_up_2_tone1"_s, u"backhand index pointing up: light skin tone"_s})},
{u"backhand index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F446\U0001F3FC"), u"point_up_2_tone2"_s, u"backhand index pointing up: medium-light skin tone"_s})},
{u"backhand index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F446\U0001F3FD"), u"point_up_2_tone3"_s, u"backhand index pointing up: medium skin tone"_s})},
{u"backhand index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F446\U0001F3FE"), u"point_up_2_tone4"_s, u"backhand index pointing up: medium-dark skin tone"_s})},
{u"backhand index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F446\U0001F3FF"), u"point_up_2_tone5"_s, u"backhand index pointing up: dark skin tone"_s})},
{u"middle finger"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F595\U0001F3FB"), u"middle_finger_tone1"_s, u"middle finger: light skin tone"_s})},
{u"middle finger"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F595\U0001F3FC"), u"middle_finger_tone2"_s, u"middle finger: medium-light skin tone"_s})},
{u"middle finger"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F595\U0001F3FD"), u"middle_finger_tone3"_s, u"middle finger: medium skin tone"_s})},
{u"middle finger"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F595\U0001F3FE"), u"middle_finger_tone4"_s, u"middle finger: medium-dark skin tone"_s})},
{u"middle finger"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F595\U0001F3FF"), u"middle_finger_tone5"_s, u"middle finger: dark skin tone"_s})},
{u"backhand index pointing down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F447\U0001F3FB"), u"point_down_tone1"_s, u"backhand index pointing down: light skin tone"_s})},
{u"backhand index pointing down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F447\U0001F3FC"), u"point_down_tone2"_s, u"backhand index pointing down: medium-light skin tone"_s})},
{u"backhand index pointing down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F447\U0001F3FD"), u"point_down_tone3"_s, u"backhand index pointing down: medium skin tone"_s})},
{u"backhand index pointing down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F447\U0001F3FE"), u"point_down_tone4"_s, u"backhand index pointing down: medium-dark skin tone"_s})},
{u"backhand index pointing down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F447\U0001F3FF"), u"point_down_tone5"_s, u"backhand index pointing down: dark skin tone"_s})},
{u"index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000261D\U0001F3FB"), u"point_up_tone1"_s, u"index pointing up: light skin tone"_s})},
{u"index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000261D\U0001F3FC"), u"point_up_tone2"_s, u"index pointing up: medium-light skin tone"_s})},
{u"index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000261D\U0001F3FD"), u"point_up_tone3"_s, u"index pointing up: medium skin tone"_s})},
{u"index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000261D\U0001F3FE"), u"point_up_tone4"_s, u"index pointing up: medium-dark skin tone"_s})},
{u"index pointing up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000261D\U0001F3FF"), u"point_up_tone5"_s, u"index pointing up: dark skin tone"_s})},
{u"index pointing at the viewer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF5\U0001F3FB"), u"index pointing at the viewer: light skin tone"_s, u"index pointing at the viewer: light skin tone"_s})},
{u"index pointing at the viewer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF5\U0001F3FC"), u"index pointing at the viewer: medium-light skin tone"_s, u"index pointing at the viewer: medium-light skin tone"_s})},
{u"index pointing at the viewer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF5\U0001F3FD"), u"index pointing at the viewer: medium skin tone"_s, u"index pointing at the viewer: medium skin tone"_s})},
{u"index pointing at the viewer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF5\U0001F3FE"), u"index pointing at the viewer: medium-dark skin tone"_s, u"index pointing at the viewer: medium-dark skin tone"_s})},
{u"index pointing at the viewer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF5\U0001F3FF"), u"index pointing at the viewer: dark skin tone"_s, u"index pointing at the viewer: dark skin tone"_s})},
{u"thumbs up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44D\U0001F3FB"), u"thumbsup_tone1"_s, u"thumbs up: light skin tone"_s})},
{u"thumbs up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44D\U0001F3FC"), u"thumbsup_tone2"_s, u"thumbs up: medium-light skin tone"_s})},
{u"thumbs up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44D\U0001F3FD"), u"thumbsup_tone3"_s, u"thumbs up: medium skin tone"_s})},
{u"thumbs up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44D\U0001F3FE"), u"thumbsup_tone4"_s, u"thumbs up: medium-dark skin tone"_s})},
{u"thumbs up"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44D\U0001F3FF"), u"thumbsup_tone5"_s, u"thumbs up: dark skin tone"_s})},
{u"thumbs down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44E\U0001F3FB"), u"thumbsdown_tone1"_s, u"thumbs down: light skin tone"_s})},
{u"thumbs down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44E\U0001F3FC"), u"thumbsdown_tone2"_s, u"thumbs down: medium-light skin tone"_s})},
{u"thumbs down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44E\U0001F3FD"), u"thumbsdown_tone3"_s, u"thumbs down: medium skin tone"_s})},
{u"thumbs down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44E\U0001F3FE"), u"thumbsdown_tone4"_s, u"thumbs down: medium-dark skin tone"_s})},
{u"thumbs down"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44E\U0001F3FF"), u"thumbsdown_tone5"_s, u"thumbs down: dark skin tone"_s})},
{u"raised fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270A\U0001F3FB"), u"fist_tone1"_s, u"raised fist: light skin tone"_s})},
{u"raised fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270A\U0001F3FC"), u"fist_tone2"_s, u"raised fist: medium-light skin tone"_s})},
{u"raised fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270A\U0001F3FD"), u"fist_tone3"_s, u"raised fist: medium skin tone"_s})},
{u"raised fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270A\U0001F3FE"), u"fist_tone4"_s, u"raised fist: medium-dark skin tone"_s})},
{u"raised fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270A\U0001F3FF"), u"fist_tone5"_s, u"raised fist: dark skin tone"_s})},
{u"oncoming fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44A\U0001F3FB"), u"punch_tone1"_s, u"oncoming fist: light skin tone"_s})},
{u"oncoming fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44A\U0001F3FC"), u"punch_tone2"_s, u"oncoming fist: medium-light skin tone"_s})},
{u"oncoming fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44A\U0001F3FD"), u"punch_tone3"_s, u"oncoming fist: medium skin tone"_s})},
{u"oncoming fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44A\U0001F3FE"), u"punch_tone4"_s, u"oncoming fist: medium-dark skin tone"_s})},
{u"oncoming fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44A\U0001F3FF"), u"punch_tone5"_s, u"oncoming fist: dark skin tone"_s})},
{u"left-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91B\U0001F3FB"), u"left_facing_fist_tone1"_s, u"left-facing fist: light skin tone"_s})},
{u"left-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91B\U0001F3FC"), u"left_facing_fist_tone2"_s, u"left-facing fist: medium-light skin tone"_s})},
{u"left-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91B\U0001F3FD"), u"left_facing_fist_tone3"_s, u"left-facing fist: medium skin tone"_s})},
{u"left-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91B\U0001F3FE"), u"left_facing_fist_tone4"_s, u"left-facing fist: medium-dark skin tone"_s})},
{u"left-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91B\U0001F3FF"), u"left_facing_fist_tone5"_s, u"left-facing fist: dark skin tone"_s})},
{u"right-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91C\U0001F3FB"), u"right_facing_fist_tone1"_s, u"right-facing fist: light skin tone"_s})},
{u"right-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91C\U0001F3FC"), u"right_facing_fist_tone2"_s, u"right-facing fist: medium-light skin tone"_s})},
{u"right-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91C\U0001F3FD"), u"right_facing_fist_tone3"_s, u"right-facing fist: medium skin tone"_s})},
{u"right-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91C\U0001F3FE"), u"right_facing_fist_tone4"_s, u"right-facing fist: medium-dark skin tone"_s})},
{u"right-facing fist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91C\U0001F3FF"), u"right_facing_fist_tone5"_s, u"right-facing fist: dark skin tone"_s})},
{u"clapping hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44F\U0001F3FB"), u"clap_tone1"_s, u"clapping hands: light skin tone"_s})},
{u"clapping hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44F\U0001F3FC"), u"clap_tone2"_s, u"clapping hands: medium-light skin tone"_s})},
{u"clapping hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44F\U0001F3FD"), u"clap_tone3"_s, u"clapping hands: medium skin tone"_s})},
{u"clapping hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44F\U0001F3FE"), u"clap_tone4"_s, u"clapping hands: medium-dark skin tone"_s})},
{u"clapping hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F44F\U0001F3FF"), u"clap_tone5"_s, u"clapping hands: dark skin tone"_s})},
{u"raising hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64C\U0001F3FB"), u"raised_hands_tone1"_s, u"raising hands: light skin tone"_s})},
{u"raising hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64C\U0001F3FC"), u"raised_hands_tone2"_s, u"raising hands: medium-light skin tone"_s})},
{u"raising hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64C\U0001F3FD"), u"raised_hands_tone3"_s, u"raising hands: medium skin tone"_s})},
{u"raising hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64C\U0001F3FE"), u"raised_hands_tone4"_s, u"raising hands: medium-dark skin tone"_s})},
{u"raising hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64C\U0001F3FF"), u"raised_hands_tone5"_s, u"raising hands: dark skin tone"_s})},
{u"heart hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF6\U0001F3FB"), u"heart hands: light skin tone"_s, u"heart hands: light skin tone"_s})},
{u"heart hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF6\U0001F3FC"), u"heart hands: medium-light skin tone"_s, u"heart hands: medium-light skin tone"_s})},
{u"heart hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF6\U0001F3FD"), u"heart hands: medium skin tone"_s, u"heart hands: medium skin tone"_s})},
{u"heart hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF6\U0001F3FE"), u"heart hands: medium-dark skin tone"_s, u"heart hands: medium-dark skin tone"_s})},
{u"heart hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF6\U0001F3FF"), u"heart hands: dark skin tone"_s, u"heart hands: dark skin tone"_s})},
{u"open hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F450\U0001F3FB"), u"open_hands_tone1"_s, u"open hands: light skin tone"_s})},
{u"open hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F450\U0001F3FC"), u"open_hands_tone2"_s, u"open hands: medium-light skin tone"_s})},
{u"open hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F450\U0001F3FD"), u"open_hands_tone3"_s, u"open hands: medium skin tone"_s})},
{u"open hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F450\U0001F3FE"), u"open_hands_tone4"_s, u"open hands: medium-dark skin tone"_s})},
{u"open hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F450\U0001F3FF"), u"open_hands_tone5"_s, u"open hands: dark skin tone"_s})},
{u"palms up together"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F932\U0001F3FB"), u"palms_up_together_tone1"_s, u"palms up together: light skin tone"_s})},
{u"palms up together"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F932\U0001F3FC"), u"palms_up_together_tone2"_s, u"palms up together: medium-light skin tone"_s})},
{u"palms up together"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F932\U0001F3FD"), u"palms_up_together_tone3"_s, u"palms up together: medium skin tone"_s})},
{u"palms up together"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F932\U0001F3FE"), u"palms_up_together_tone4"_s, u"palms up together: medium-dark skin tone"_s})},
{u"palms up together"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F932\U0001F3FF"), u"palms_up_together_tone5"_s, u"palms up together: dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91D\U0001F3FB"), u"handshake: light skin tone"_s, u"handshake: light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91D\U0001F3FC"), u"handshake: medium-light skin tone"_s, u"handshake: medium-light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91D\U0001F3FD"), u"handshake: medium skin tone"_s, u"handshake: medium skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91D\U0001F3FE"), u"handshake: medium-dark skin tone"_s, u"handshake: medium-dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F91D\U0001F3FF"), u"handshake: dark skin tone"_s, u"handshake: dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FC"), u"handshake: light skin tone, medium-light skin tone"_s, u"handshake: light skin tone, medium-light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FD"), u"handshake: light skin tone, medium skin tone"_s, u"handshake: light skin tone, medium skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FE"), u"handshake: light skin tone, medium-dark skin tone"_s, u"handshake: light skin tone, medium-dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FB\U0000200D\U0001FAF2\U0001F3FF"), u"handshake: light skin tone, dark skin tone"_s, u"handshake: light skin tone, dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FB"), u"handshake: medium-light skin tone, light skin tone"_s, u"handshake: medium-light skin tone, light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FD"), u"handshake: medium-light skin tone, medium skin tone"_s, u"handshake: medium-light skin tone, medium skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FE"), u"handshake: medium-light skin tone, medium-dark skin tone"_s, u"handshake: medium-light skin tone, medium-dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FC\U0000200D\U0001FAF2\U0001F3FF"), u"handshake: medium-light skin tone, dark skin tone"_s, u"handshake: medium-light skin tone, dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FB"), u"handshake: medium skin tone, light skin tone"_s, u"handshake: medium skin tone, light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FC"), u"handshake: medium skin tone, medium-light skin tone"_s, u"handshake: medium skin tone, medium-light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FE"), u"handshake: medium skin tone, medium-dark skin tone"_s, u"handshake: medium skin tone, medium-dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FD\U0000200D\U0001FAF2\U0001F3FF"), u"handshake: medium skin tone, dark skin tone"_s, u"handshake: medium skin tone, dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FB"), u"handshake: medium-dark skin tone, light skin tone"_s, u"handshake: medium-dark skin tone, light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FC"), u"handshake: medium-dark skin tone, medium-light skin tone"_s, u"handshake: medium-dark skin tone, medium-light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FD"), u"handshake: medium-dark skin tone, medium skin tone"_s, u"handshake: medium-dark skin tone, medium skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FE\U0000200D\U0001FAF2\U0001F3FF"), u"handshake: medium-dark skin tone, dark skin tone"_s, u"handshake: medium-dark skin tone, dark skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FB"), u"handshake: dark skin tone, light skin tone"_s, u"handshake: dark skin tone, light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FC"), u"handshake: dark skin tone, medium-light skin tone"_s, u"handshake: dark skin tone, medium-light skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FD"), u"handshake: dark skin tone, medium skin tone"_s, u"handshake: dark skin tone, medium skin tone"_s})},
{u"handshake"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAF1\U0001F3FF\U0000200D\U0001FAF2\U0001F3FE"), u"handshake: dark skin tone, medium-dark skin tone"_s, u"handshake: dark skin tone, medium-dark skin tone"_s})},
{u"folded hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64F\U0001F3FB"), u"pray_tone1"_s, u"folded hands: light skin tone"_s})},
{u"folded hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64F\U0001F3FC"), u"pray_tone2"_s, u"folded hands: medium-light skin tone"_s})},
{u"folded hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64F\U0001F3FD"), u"pray_tone3"_s, u"folded hands: medium skin tone"_s})},
{u"folded hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64F\U0001F3FE"), u"pray_tone4"_s, u"folded hands: medium-dark skin tone"_s})},
{u"folded hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64F\U0001F3FF"), u"pray_tone5"_s, u"folded hands: dark skin tone"_s})},
{u"writing hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270D\U0001F3FB"), u"writing_hand_tone1"_s, u"writing hand: light skin tone"_s})},
{u"writing hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270D\U0001F3FC"), u"writing_hand_tone2"_s, u"writing hand: medium-light skin tone"_s})},
{u"writing hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270D\U0001F3FD"), u"writing_hand_tone3"_s, u"writing hand: medium skin tone"_s})},
{u"writing hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270D\U0001F3FE"), u"writing_hand_tone4"_s, u"writing hand: medium-dark skin tone"_s})},
{u"writing hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0000270D\U0001F3FF"), u"writing_hand_tone5"_s, u"writing hand: dark skin tone"_s})},
{u"nail polish"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F485\U0001F3FB"), u"nail_care_tone1"_s, u"nail polish: light skin tone"_s})},
{u"nail polish"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F485\U0001F3FC"), u"nail_care_tone2"_s, u"nail polish: medium-light skin tone"_s})},
{u"nail polish"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F485\U0001F3FD"), u"nail_care_tone3"_s, u"nail polish: medium skin tone"_s})},
{u"nail polish"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F485\U0001F3FE"), u"nail_care_tone4"_s, u"nail polish: medium-dark skin tone"_s})},
{u"nail polish"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F485\U0001F3FF"), u"nail_care_tone5"_s, u"nail polish: dark skin tone"_s})},
{u"selfie"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F933\U0001F3FB"), u"selfie_tone1"_s, u"selfie: light skin tone"_s})},
{u"selfie"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F933\U0001F3FC"), u"selfie_tone2"_s, u"selfie: medium-light skin tone"_s})},
{u"selfie"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F933\U0001F3FD"), u"selfie_tone3"_s, u"selfie: medium skin tone"_s})},
{u"selfie"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F933\U0001F3FE"), u"selfie_tone4"_s, u"selfie: medium-dark skin tone"_s})},
{u"selfie"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F933\U0001F3FF"), u"selfie_tone5"_s, u"selfie: dark skin tone"_s})},
{u"flexed biceps"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F4AA\U0001F3FB"), u"muscle_tone1"_s, u"flexed biceps: light skin tone"_s})},
{u"flexed biceps"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F4AA\U0001F3FC"), u"muscle_tone2"_s, u"flexed biceps: medium-light skin tone"_s})},
{u"flexed biceps"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F4AA\U0001F3FD"), u"muscle_tone3"_s, u"flexed biceps: medium skin tone"_s})},
{u"flexed biceps"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F4AA\U0001F3FE"), u"muscle_tone4"_s, u"flexed biceps: medium-dark skin tone"_s})},
{u"flexed biceps"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F4AA\U0001F3FF"), u"muscle_tone5"_s, u"flexed biceps: dark skin tone"_s})},
{u"leg"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B5\U0001F3FB"), u"leg_tone1"_s, u"leg: light skin tone"_s})},
{u"leg"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B5\U0001F3FC"), u"leg_tone2"_s, u"leg: medium-light skin tone"_s})},
{u"leg"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B5\U0001F3FD"), u"leg_tone3"_s, u"leg: medium skin tone"_s})},
{u"leg"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B5\U0001F3FE"), u"leg_tone4"_s, u"leg: medium-dark skin tone"_s})},
{u"leg"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B5\U0001F3FF"), u"leg_tone5"_s, u"leg: dark skin tone"_s})},
{u"foot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B6\U0001F3FB"), u"foot_tone1"_s, u"foot: light skin tone"_s})},
{u"foot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B6\U0001F3FC"), u"foot_tone2"_s, u"foot: medium-light skin tone"_s})},
{u"foot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B6\U0001F3FD"), u"foot_tone3"_s, u"foot: medium skin tone"_s})},
{u"foot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B6\U0001F3FE"), u"foot_tone4"_s, u"foot: medium-dark skin tone"_s})},
{u"foot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B6\U0001F3FF"), u"foot_tone5"_s, u"foot: dark skin tone"_s})},
{u"ear"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F442\U0001F3FB"), u"ear_tone1"_s, u"ear: light skin tone"_s})},
{u"ear"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F442\U0001F3FC"), u"ear_tone2"_s, u"ear: medium-light skin tone"_s})},
{u"ear"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F442\U0001F3FD"), u"ear_tone3"_s, u"ear: medium skin tone"_s})},
{u"ear"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F442\U0001F3FE"), u"ear_tone4"_s, u"ear: medium-dark skin tone"_s})},
{u"ear"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F442\U0001F3FF"), u"ear_tone5"_s, u"ear: dark skin tone"_s})},
{u"ear with hearing aid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9BB\U0001F3FB"), u"ear with hearing aid: light skin tone"_s, u"ear with hearing aid: light skin tone"_s})},
{u"ear with hearing aid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9BB\U0001F3FC"), u"ear with hearing aid: medium-light skin tone"_s, u"ear with hearing aid: medium-light skin tone"_s})},
{u"ear with hearing aid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9BB\U0001F3FD"), u"ear with hearing aid: medium skin tone"_s, u"ear with hearing aid: medium skin tone"_s})},
{u"ear with hearing aid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9BB\U0001F3FE"), u"ear with hearing aid: medium-dark skin tone"_s, u"ear with hearing aid: medium-dark skin tone"_s})},
{u"ear with hearing aid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9BB\U0001F3FF"), u"ear with hearing aid: dark skin tone"_s, u"ear with hearing aid: dark skin tone"_s})},
{u"nose"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F443\U0001F3FB"), u"nose_tone1"_s, u"nose: light skin tone"_s})},
{u"nose"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F443\U0001F3FC"), u"nose_tone2"_s, u"nose: medium-light skin tone"_s})},
{u"nose"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F443\U0001F3FD"), u"nose_tone3"_s, u"nose: medium skin tone"_s})},
{u"nose"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F443\U0001F3FE"), u"nose_tone4"_s, u"nose: medium-dark skin tone"_s})},
{u"nose"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F443\U0001F3FF"), u"nose_tone5"_s, u"nose: dark skin tone"_s})},
{u"baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F476\U0001F3FB"), u"baby_tone1"_s, u"baby: light skin tone"_s})},
{u"baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F476\U0001F3FC"), u"baby_tone2"_s, u"baby: medium-light skin tone"_s})},
{u"baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F476\U0001F3FD"), u"baby_tone3"_s, u"baby: medium skin tone"_s})},
{u"baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F476\U0001F3FE"), u"baby_tone4"_s, u"baby: medium-dark skin tone"_s})},
{u"baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F476\U0001F3FF"), u"baby_tone5"_s, u"baby: dark skin tone"_s})},
{u"child"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D2\U0001F3FB"), u"child_tone1"_s, u"child: light skin tone"_s})},
{u"child"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D2\U0001F3FC"), u"child_tone2"_s, u"child: medium-light skin tone"_s})},
{u"child"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D2\U0001F3FD"), u"child_tone3"_s, u"child: medium skin tone"_s})},
{u"child"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D2\U0001F3FE"), u"child_tone4"_s, u"child: medium-dark skin tone"_s})},
{u"child"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D2\U0001F3FF"), u"child_tone5"_s, u"child: dark skin tone"_s})},
{u"boy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F466\U0001F3FB"), u"boy_tone1"_s, u"boy: light skin tone"_s})},
{u"boy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F466\U0001F3FC"), u"boy_tone2"_s, u"boy: medium-light skin tone"_s})},
{u"boy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F466\U0001F3FD"), u"boy_tone3"_s, u"boy: medium skin tone"_s})},
{u"boy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F466\U0001F3FE"), u"boy_tone4"_s, u"boy: medium-dark skin tone"_s})},
{u"boy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F466\U0001F3FF"), u"boy_tone5"_s, u"boy: dark skin tone"_s})},
{u"girl"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F467\U0001F3FB"), u"girl_tone1"_s, u"girl: light skin tone"_s})},
{u"girl"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F467\U0001F3FC"), u"girl_tone2"_s, u"girl: medium-light skin tone"_s})},
{u"girl"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F467\U0001F3FD"), u"girl_tone3"_s, u"girl: medium skin tone"_s})},
{u"girl"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F467\U0001F3FE"), u"girl_tone4"_s, u"girl: medium-dark skin tone"_s})},
{u"girl"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F467\U0001F3FF"), u"girl_tone5"_s, u"girl: dark skin tone"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB"), u"adult_tone1"_s, u"person: light skin tone"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC"), u"adult_tone2"_s, u"person: medium-light skin tone"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD"), u"adult_tone3"_s, u"person: medium skin tone"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE"), u"adult_tone4"_s, u"person: medium-dark skin tone"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF"), u"adult_tone5"_s, u"person: dark skin tone"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FB"), u"person_with_blond_hair_tone1"_s, u"person: light skin tone, blond hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FC"), u"person_with_blond_hair_tone2"_s, u"person: medium-light skin tone, blond hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FD"), u"person_with_blond_hair_tone3"_s, u"person: medium skin tone, blond hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FE"), u"person_with_blond_hair_tone4"_s, u"person: medium-dark skin tone, blond hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FF"), u"person_with_blond_hair_tone5"_s, u"person: dark skin tone, blond hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB"), u"man_tone1"_s, u"man: light skin tone"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC"), u"man_tone2"_s, u"man: medium-light skin tone"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD"), u"man_tone3"_s, u"man: medium skin tone"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE"), u"man_tone4"_s, u"man: medium-dark skin tone"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF"), u"man_tone5"_s, u"man: dark skin tone"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FB"), u"bearded_person_tone1"_s, u"person: light skin tone, beard"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FC"), u"bearded_person_tone2"_s, u"person: medium-light skin tone, beard"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FD"), u"bearded_person_tone3"_s, u"person: medium skin tone, beard"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FE"), u"bearded_person_tone4"_s, u"person: medium-dark skin tone, beard"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FF"), u"bearded_person_tone5"_s, u"person: dark skin tone, beard"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man: light skin tone, beard"_s, u"man: light skin tone, beard"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man: medium-light skin tone, beard"_s, u"man: medium-light skin tone, beard"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man: medium skin tone, beard"_s, u"man: medium skin tone, beard"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man: medium-dark skin tone, beard"_s, u"man: medium-dark skin tone, beard"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man: dark skin tone, beard"_s, u"man: dark skin tone, beard"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman: light skin tone, beard"_s, u"woman: light skin tone, beard"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman: medium-light skin tone, beard"_s, u"woman: medium-light skin tone, beard"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman: medium skin tone, beard"_s, u"woman: medium skin tone, beard"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman: medium-dark skin tone, beard"_s, u"woman: medium-dark skin tone, beard"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D4\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman: dark skin tone, beard"_s, u"woman: dark skin tone, beard"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F9B0"), u"man: light skin tone, red hair"_s, u"man: light skin tone, red hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F9B0"), u"man: medium-light skin tone, red hair"_s, u"man: medium-light skin tone, red hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F9B0"), u"man: medium skin tone, red hair"_s, u"man: medium skin tone, red hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F9B0"), u"man: medium-dark skin tone, red hair"_s, u"man: medium-dark skin tone, red hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F9B0"), u"man: dark skin tone, red hair"_s, u"man: dark skin tone, red hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F9B1"), u"man: light skin tone, curly hair"_s, u"man: light skin tone, curly hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F9B1"), u"man: medium-light skin tone, curly hair"_s, u"man: medium-light skin tone, curly hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F9B1"), u"man: medium skin tone, curly hair"_s, u"man: medium skin tone, curly hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F9B1"), u"man: medium-dark skin tone, curly hair"_s, u"man: medium-dark skin tone, curly hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F9B1"), u"man: dark skin tone, curly hair"_s, u"man: dark skin tone, curly hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F9B3"), u"man: light skin tone, white hair"_s, u"man: light skin tone, white hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F9B3"), u"man: medium-light skin tone, white hair"_s, u"man: medium-light skin tone, white hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F9B3"), u"man: medium skin tone, white hair"_s, u"man: medium skin tone, white hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F9B3"), u"man: medium-dark skin tone, white hair"_s, u"man: medium-dark skin tone, white hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F9B3"), u"man: dark skin tone, white hair"_s, u"man: dark skin tone, white hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F9B2"), u"man: light skin tone, bald"_s, u"man: light skin tone, bald"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F9B2"), u"man: medium-light skin tone, bald"_s, u"man: medium-light skin tone, bald"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F9B2"), u"man: medium skin tone, bald"_s, u"man: medium skin tone, bald"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F9B2"), u"man: medium-dark skin tone, bald"_s, u"man: medium-dark skin tone, bald"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F9B2"), u"man: dark skin tone, bald"_s, u"man: dark skin tone, bald"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB"), u"woman_tone1"_s, u"woman: light skin tone"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC"), u"woman_tone2"_s, u"woman: medium-light skin tone"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD"), u"woman_tone3"_s, u"woman: medium skin tone"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE"), u"woman_tone4"_s, u"woman: medium-dark skin tone"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF"), u"woman_tone5"_s, u"woman: dark skin tone"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F9B0"), u"woman: light skin tone, red hair"_s, u"woman: light skin tone, red hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F9B0"), u"woman: medium-light skin tone, red hair"_s, u"woman: medium-light skin tone, red hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F9B0"), u"woman: medium skin tone, red hair"_s, u"woman: medium skin tone, red hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F9B0"), u"woman: medium-dark skin tone, red hair"_s, u"woman: medium-dark skin tone, red hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F9B0"), u"woman: dark skin tone, red hair"_s, u"woman: dark skin tone, red hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F9B0"), u"person: light skin tone, red hair"_s, u"person: light skin tone, red hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F9B0"), u"person: medium-light skin tone, red hair"_s, u"person: medium-light skin tone, red hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F9B0"), u"person: medium skin tone, red hair"_s, u"person: medium skin tone, red hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F9B0"), u"person: medium-dark skin tone, red hair"_s, u"person: medium-dark skin tone, red hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F9B0"), u"person: dark skin tone, red hair"_s, u"person: dark skin tone, red hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F9B1"), u"woman: light skin tone, curly hair"_s, u"woman: light skin tone, curly hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F9B1"), u"woman: medium-light skin tone, curly hair"_s, u"woman: medium-light skin tone, curly hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F9B1"), u"woman: medium skin tone, curly hair"_s, u"woman: medium skin tone, curly hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F9B1"), u"woman: medium-dark skin tone, curly hair"_s, u"woman: medium-dark skin tone, curly hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F9B1"), u"woman: dark skin tone, curly hair"_s, u"woman: dark skin tone, curly hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F9B1"), u"person: light skin tone, curly hair"_s, u"person: light skin tone, curly hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F9B1"), u"person: medium-light skin tone, curly hair"_s, u"person: medium-light skin tone, curly hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F9B1"), u"person: medium skin tone, curly hair"_s, u"person: medium skin tone, curly hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F9B1"), u"person: medium-dark skin tone, curly hair"_s, u"person: medium-dark skin tone, curly hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F9B1"), u"person: dark skin tone, curly hair"_s, u"person: dark skin tone, curly hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F9B3"), u"woman: light skin tone, white hair"_s, u"woman: light skin tone, white hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F9B3"), u"woman: medium-light skin tone, white hair"_s, u"woman: medium-light skin tone, white hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F9B3"), u"woman: medium skin tone, white hair"_s, u"woman: medium skin tone, white hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F9B3"), u"woman: medium-dark skin tone, white hair"_s, u"woman: medium-dark skin tone, white hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F9B3"), u"woman: dark skin tone, white hair"_s, u"woman: dark skin tone, white hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F9B3"), u"person: light skin tone, white hair"_s, u"person: light skin tone, white hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F9B3"), u"person: medium-light skin tone, white hair"_s, u"person: medium-light skin tone, white hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F9B3"), u"person: medium skin tone, white hair"_s, u"person: medium skin tone, white hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F9B3"), u"person: medium-dark skin tone, white hair"_s, u"person: medium-dark skin tone, white hair"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F9B3"), u"person: dark skin tone, white hair"_s, u"person: dark skin tone, white hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F9B2"), u"woman: light skin tone, bald"_s, u"woman: light skin tone, bald"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F9B2"), u"woman: medium-light skin tone, bald"_s, u"woman: medium-light skin tone, bald"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F9B2"), u"woman: medium skin tone, bald"_s, u"woman: medium skin tone, bald"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F9B2"), u"woman: medium-dark skin tone, bald"_s, u"woman: medium-dark skin tone, bald"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F9B2"), u"woman: dark skin tone, bald"_s, u"woman: dark skin tone, bald"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F9B2"), u"person: light skin tone, bald"_s, u"person: light skin tone, bald"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F9B2"), u"person: medium-light skin tone, bald"_s, u"person: medium-light skin tone, bald"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F9B2"), u"person: medium skin tone, bald"_s, u"person: medium skin tone, bald"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F9B2"), u"person: medium-dark skin tone, bald"_s, u"person: medium-dark skin tone, bald"_s})},
{u"person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F9B2"), u"person: dark skin tone, bald"_s, u"person: dark skin tone, bald"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman: light skin tone, blond hair"_s, u"woman: light skin tone, blond hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman: medium-light skin tone, blond hair"_s, u"woman: medium-light skin tone, blond hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman: medium skin tone, blond hair"_s, u"woman: medium skin tone, blond hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman: medium-dark skin tone, blond hair"_s, u"woman: medium-dark skin tone, blond hair"_s})},
{u"woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman: dark skin tone, blond hair"_s, u"woman: dark skin tone, blond hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man: light skin tone, blond hair"_s, u"man: light skin tone, blond hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man: medium-light skin tone, blond hair"_s, u"man: medium-light skin tone, blond hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man: medium skin tone, blond hair"_s, u"man: medium skin tone, blond hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man: medium-dark skin tone, blond hair"_s, u"man: medium-dark skin tone, blond hair"_s})},
{u"man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F471\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man: dark skin tone, blond hair"_s, u"man: dark skin tone, blond hair"_s})},
{u"older person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D3\U0001F3FB"), u"older_adult_tone1"_s, u"older person: light skin tone"_s})},
{u"older person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D3\U0001F3FC"), u"older_adult_tone2"_s, u"older person: medium-light skin tone"_s})},
{u"older person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D3\U0001F3FD"), u"older_adult_tone3"_s, u"older person: medium skin tone"_s})},
{u"older person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D3\U0001F3FE"), u"older_adult_tone4"_s, u"older person: medium-dark skin tone"_s})},
{u"older person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D3\U0001F3FF"), u"older_adult_tone5"_s, u"older person: dark skin tone"_s})},
{u"old man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F474\U0001F3FB"), u"older_man_tone1"_s, u"old man: light skin tone"_s})},
{u"old man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F474\U0001F3FC"), u"older_man_tone2"_s, u"old man: medium-light skin tone"_s})},
{u"old man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F474\U0001F3FD"), u"older_man_tone3"_s, u"old man: medium skin tone"_s})},
{u"old man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F474\U0001F3FE"), u"older_man_tone4"_s, u"old man: medium-dark skin tone"_s})},
{u"old man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F474\U0001F3FF"), u"older_man_tone5"_s, u"old man: dark skin tone"_s})},
{u"old woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F475\U0001F3FB"), u"older_woman_tone1"_s, u"old woman: light skin tone"_s})},
{u"old woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F475\U0001F3FC"), u"older_woman_tone2"_s, u"old woman: medium-light skin tone"_s})},
{u"old woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F475\U0001F3FD"), u"older_woman_tone3"_s, u"old woman: medium skin tone"_s})},
{u"old woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F475\U0001F3FE"), u"older_woman_tone4"_s, u"old woman: medium-dark skin tone"_s})},
{u"old woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F475\U0001F3FF"), u"older_woman_tone5"_s, u"old woman: dark skin tone"_s})},
{u"person frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FB"), u"person_frowning_tone1"_s, u"person frowning: light skin tone"_s})},
{u"person frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FC"), u"person_frowning_tone2"_s, u"person frowning: medium-light skin tone"_s})},
{u"person frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FD"), u"person_frowning_tone3"_s, u"person frowning: medium skin tone"_s})},
{u"person frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FE"), u"person_frowning_tone4"_s, u"person frowning: medium-dark skin tone"_s})},
{u"person frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FF"), u"person_frowning_tone5"_s, u"person frowning: dark skin tone"_s})},
{u"man frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man frowning: light skin tone"_s, u"man frowning: light skin tone"_s})},
{u"man frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man frowning: medium-light skin tone"_s, u"man frowning: medium-light skin tone"_s})},
{u"man frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man frowning: medium skin tone"_s, u"man frowning: medium skin tone"_s})},
{u"man frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man frowning: medium-dark skin tone"_s, u"man frowning: medium-dark skin tone"_s})},
{u"man frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man frowning: dark skin tone"_s, u"man frowning: dark skin tone"_s})},
{u"woman frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman frowning: light skin tone"_s, u"woman frowning: light skin tone"_s})},
{u"woman frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman frowning: medium-light skin tone"_s, u"woman frowning: medium-light skin tone"_s})},
{u"woman frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman frowning: medium skin tone"_s, u"woman frowning: medium skin tone"_s})},
{u"woman frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman frowning: medium-dark skin tone"_s, u"woman frowning: medium-dark skin tone"_s})},
{u"woman frowning"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64D\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman frowning: dark skin tone"_s, u"woman frowning: dark skin tone"_s})},
{u"person pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FB"), u"person_with_pouting_face_tone1"_s, u"person pouting: light skin tone"_s})},
{u"person pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FC"), u"person_with_pouting_face_tone2"_s, u"person pouting: medium-light skin tone"_s})},
{u"person pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FD"), u"person_with_pouting_face_tone3"_s, u"person pouting: medium skin tone"_s})},
{u"person pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FE"), u"person_with_pouting_face_tone4"_s, u"person pouting: medium-dark skin tone"_s})},
{u"person pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FF"), u"person_with_pouting_face_tone5"_s, u"person pouting: dark skin tone"_s})},
{u"man pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man pouting: light skin tone"_s, u"man pouting: light skin tone"_s})},
{u"man pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man pouting: medium-light skin tone"_s, u"man pouting: medium-light skin tone"_s})},
{u"man pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man pouting: medium skin tone"_s, u"man pouting: medium skin tone"_s})},
{u"man pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man pouting: medium-dark skin tone"_s, u"man pouting: medium-dark skin tone"_s})},
{u"man pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man pouting: dark skin tone"_s, u"man pouting: dark skin tone"_s})},
{u"woman pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman pouting: light skin tone"_s, u"woman pouting: light skin tone"_s})},
{u"woman pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman pouting: medium-light skin tone"_s, u"woman pouting: medium-light skin tone"_s})},
{u"woman pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman pouting: medium skin tone"_s, u"woman pouting: medium skin tone"_s})},
{u"woman pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman pouting: medium-dark skin tone"_s, u"woman pouting: medium-dark skin tone"_s})},
{u"woman pouting"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64E\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman pouting: dark skin tone"_s, u"woman pouting: dark skin tone"_s})},
{u"person gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FB"), u"no_good_tone1"_s, u"person gesturing NO: light skin tone"_s})},
{u"person gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FC"), u"no_good_tone2"_s, u"person gesturing NO: medium-light skin tone"_s})},
{u"person gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FD"), u"no_good_tone3"_s, u"person gesturing NO: medium skin tone"_s})},
{u"person gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FE"), u"no_good_tone4"_s, u"person gesturing NO: medium-dark skin tone"_s})},
{u"person gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FF"), u"no_good_tone5"_s, u"person gesturing NO: dark skin tone"_s})},
{u"man gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man gesturing NO: light skin tone"_s, u"man gesturing NO: light skin tone"_s})},
{u"man gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man gesturing NO: medium-light skin tone"_s, u"man gesturing NO: medium-light skin tone"_s})},
{u"man gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man gesturing NO: medium skin tone"_s, u"man gesturing NO: medium skin tone"_s})},
{u"man gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man gesturing NO: medium-dark skin tone"_s, u"man gesturing NO: medium-dark skin tone"_s})},
{u"man gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man gesturing NO: dark skin tone"_s, u"man gesturing NO: dark skin tone"_s})},
{u"woman gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman gesturing NO: light skin tone"_s, u"woman gesturing NO: light skin tone"_s})},
{u"woman gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman gesturing NO: medium-light skin tone"_s, u"woman gesturing NO: medium-light skin tone"_s})},
{u"woman gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman gesturing NO: medium skin tone"_s, u"woman gesturing NO: medium skin tone"_s})},
{u"woman gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman gesturing NO: medium-dark skin tone"_s, u"woman gesturing NO: medium-dark skin tone"_s})},
{u"woman gesturing NO"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F645\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman gesturing NO: dark skin tone"_s, u"woman gesturing NO: dark skin tone"_s})},
{u"person gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FB"), u"ok_woman_tone1"_s, u"person gesturing OK: light skin tone"_s})},
{u"person gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FC"), u"ok_woman_tone2"_s, u"person gesturing OK: medium-light skin tone"_s})},
{u"person gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FD"), u"ok_woman_tone3"_s, u"person gesturing OK: medium skin tone"_s})},
{u"person gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FE"), u"ok_woman_tone4"_s, u"person gesturing OK: medium-dark skin tone"_s})},
{u"person gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FF"), u"ok_woman_tone5"_s, u"person gesturing OK: dark skin tone"_s})},
{u"man gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man gesturing OK: light skin tone"_s, u"man gesturing OK: light skin tone"_s})},
{u"man gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man gesturing OK: medium-light skin tone"_s, u"man gesturing OK: medium-light skin tone"_s})},
{u"man gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man gesturing OK: medium skin tone"_s, u"man gesturing OK: medium skin tone"_s})},
{u"man gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man gesturing OK: medium-dark skin tone"_s, u"man gesturing OK: medium-dark skin tone"_s})},
{u"man gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man gesturing OK: dark skin tone"_s, u"man gesturing OK: dark skin tone"_s})},
{u"woman gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman gesturing OK: light skin tone"_s, u"woman gesturing OK: light skin tone"_s})},
{u"woman gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman gesturing OK: medium-light skin tone"_s, u"woman gesturing OK: medium-light skin tone"_s})},
{u"woman gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman gesturing OK: medium skin tone"_s, u"woman gesturing OK: medium skin tone"_s})},
{u"woman gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman gesturing OK: medium-dark skin tone"_s, u"woman gesturing OK: medium-dark skin tone"_s})},
{u"woman gesturing OK"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F646\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman gesturing OK: dark skin tone"_s, u"woman gesturing OK: dark skin tone"_s})},
{u"person tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FB"), u"information_desk_person_tone1"_s, u"person tipping hand: light skin tone"_s})},
{u"person tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FC"), u"information_desk_person_tone2"_s, u"person tipping hand: medium-light skin tone"_s})},
{u"person tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FD"), u"information_desk_person_tone3"_s, u"person tipping hand: medium skin tone"_s})},
{u"person tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FE"), u"information_desk_person_tone4"_s, u"person tipping hand: medium-dark skin tone"_s})},
{u"person tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FF"), u"information_desk_person_tone5"_s, u"person tipping hand: dark skin tone"_s})},
{u"man tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man tipping hand: light skin tone"_s, u"man tipping hand: light skin tone"_s})},
{u"man tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man tipping hand: medium-light skin tone"_s, u"man tipping hand: medium-light skin tone"_s})},
{u"man tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man tipping hand: medium skin tone"_s, u"man tipping hand: medium skin tone"_s})},
{u"man tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man tipping hand: medium-dark skin tone"_s, u"man tipping hand: medium-dark skin tone"_s})},
{u"man tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man tipping hand: dark skin tone"_s, u"man tipping hand: dark skin tone"_s})},
{u"woman tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman tipping hand: light skin tone"_s, u"woman tipping hand: light skin tone"_s})},
{u"woman tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman tipping hand: medium-light skin tone"_s, u"woman tipping hand: medium-light skin tone"_s})},
{u"woman tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman tipping hand: medium skin tone"_s, u"woman tipping hand: medium skin tone"_s})},
{u"woman tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman tipping hand: medium-dark skin tone"_s, u"woman tipping hand: medium-dark skin tone"_s})},
{u"woman tipping hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F481\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman tipping hand: dark skin tone"_s, u"woman tipping hand: dark skin tone"_s})},
{u"person raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FB"), u"raising_hand_tone1"_s, u"person raising hand: light skin tone"_s})},
{u"person raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FC"), u"raising_hand_tone2"_s, u"person raising hand: medium-light skin tone"_s})},
{u"person raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FD"), u"raising_hand_tone3"_s, u"person raising hand: medium skin tone"_s})},
{u"person raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FE"), u"raising_hand_tone4"_s, u"person raising hand: medium-dark skin tone"_s})},
{u"person raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FF"), u"raising_hand_tone5"_s, u"person raising hand: dark skin tone"_s})},
{u"man raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man raising hand: light skin tone"_s, u"man raising hand: light skin tone"_s})},
{u"man raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man raising hand: medium-light skin tone"_s, u"man raising hand: medium-light skin tone"_s})},
{u"man raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man raising hand: medium skin tone"_s, u"man raising hand: medium skin tone"_s})},
{u"man raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man raising hand: medium-dark skin tone"_s, u"man raising hand: medium-dark skin tone"_s})},
{u"man raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man raising hand: dark skin tone"_s, u"man raising hand: dark skin tone"_s})},
{u"woman raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman raising hand: light skin tone"_s, u"woman raising hand: light skin tone"_s})},
{u"woman raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman raising hand: medium-light skin tone"_s, u"woman raising hand: medium-light skin tone"_s})},
{u"woman raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman raising hand: medium skin tone"_s, u"woman raising hand: medium skin tone"_s})},
{u"woman raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman raising hand: medium-dark skin tone"_s, u"woman raising hand: medium-dark skin tone"_s})},
{u"woman raising hand"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F64B\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman raising hand: dark skin tone"_s, u"woman raising hand: dark skin tone"_s})},
{u"deaf person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FB"), u"deaf person: light skin tone"_s, u"deaf person: light skin tone"_s})},
{u"deaf person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FC"), u"deaf person: medium-light skin tone"_s, u"deaf person: medium-light skin tone"_s})},
{u"deaf person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FD"), u"deaf person: medium skin tone"_s, u"deaf person: medium skin tone"_s})},
{u"deaf person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FE"), u"deaf person: medium-dark skin tone"_s, u"deaf person: medium-dark skin tone"_s})},
{u"deaf person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FF"), u"deaf person: dark skin tone"_s, u"deaf person: dark skin tone"_s})},
{u"deaf man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"deaf man: light skin tone"_s, u"deaf man: light skin tone"_s})},
{u"deaf man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"deaf man: medium-light skin tone"_s, u"deaf man: medium-light skin tone"_s})},
{u"deaf man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"deaf man: medium skin tone"_s, u"deaf man: medium skin tone"_s})},
{u"deaf man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"deaf man: medium-dark skin tone"_s, u"deaf man: medium-dark skin tone"_s})},
{u"deaf man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"deaf man: dark skin tone"_s, u"deaf man: dark skin tone"_s})},
{u"deaf woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"deaf woman: light skin tone"_s, u"deaf woman: light skin tone"_s})},
{u"deaf woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"deaf woman: medium-light skin tone"_s, u"deaf woman: medium-light skin tone"_s})},
{u"deaf woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"deaf woman: medium skin tone"_s, u"deaf woman: medium skin tone"_s})},
{u"deaf woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"deaf woman: medium-dark skin tone"_s, u"deaf woman: medium-dark skin tone"_s})},
{u"deaf woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CF\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"deaf woman: dark skin tone"_s, u"deaf woman: dark skin tone"_s})},
{u"person bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FB"), u"bow_tone1"_s, u"person bowing: light skin tone"_s})},
{u"person bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FC"), u"bow_tone2"_s, u"person bowing: medium-light skin tone"_s})},
{u"person bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FD"), u"bow_tone3"_s, u"person bowing: medium skin tone"_s})},
{u"person bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FE"), u"bow_tone4"_s, u"person bowing: medium-dark skin tone"_s})},
{u"person bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FF"), u"bow_tone5"_s, u"person bowing: dark skin tone"_s})},
{u"man bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man bowing: light skin tone"_s, u"man bowing: light skin tone"_s})},
{u"man bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man bowing: medium-light skin tone"_s, u"man bowing: medium-light skin tone"_s})},
{u"man bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man bowing: medium skin tone"_s, u"man bowing: medium skin tone"_s})},
{u"man bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man bowing: medium-dark skin tone"_s, u"man bowing: medium-dark skin tone"_s})},
{u"man bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man bowing: dark skin tone"_s, u"man bowing: dark skin tone"_s})},
{u"woman bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman bowing: light skin tone"_s, u"woman bowing: light skin tone"_s})},
{u"woman bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman bowing: medium-light skin tone"_s, u"woman bowing: medium-light skin tone"_s})},
{u"woman bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman bowing: medium skin tone"_s, u"woman bowing: medium skin tone"_s})},
{u"woman bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman bowing: medium-dark skin tone"_s, u"woman bowing: medium-dark skin tone"_s})},
{u"woman bowing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F647\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman bowing: dark skin tone"_s, u"woman bowing: dark skin tone"_s})},
{u"person facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FB"), u"face_palm_tone1"_s, u"person facepalming: light skin tone"_s})},
{u"person facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FC"), u"face_palm_tone2"_s, u"person facepalming: medium-light skin tone"_s})},
{u"person facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FD"), u"face_palm_tone3"_s, u"person facepalming: medium skin tone"_s})},
{u"person facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FE"), u"face_palm_tone4"_s, u"person facepalming: medium-dark skin tone"_s})},
{u"person facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FF"), u"face_palm_tone5"_s, u"person facepalming: dark skin tone"_s})},
{u"man facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man facepalming: light skin tone"_s, u"man facepalming: light skin tone"_s})},
{u"man facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man facepalming: medium-light skin tone"_s, u"man facepalming: medium-light skin tone"_s})},
{u"man facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man facepalming: medium skin tone"_s, u"man facepalming: medium skin tone"_s})},
{u"man facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man facepalming: medium-dark skin tone"_s, u"man facepalming: medium-dark skin tone"_s})},
{u"man facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man facepalming: dark skin tone"_s, u"man facepalming: dark skin tone"_s})},
{u"woman facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman facepalming: light skin tone"_s, u"woman facepalming: light skin tone"_s})},
{u"woman facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman facepalming: medium-light skin tone"_s, u"woman facepalming: medium-light skin tone"_s})},
{u"woman facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman facepalming: medium skin tone"_s, u"woman facepalming: medium skin tone"_s})},
{u"woman facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman facepalming: medium-dark skin tone"_s, u"woman facepalming: medium-dark skin tone"_s})},
{u"woman facepalming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F926\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman facepalming: dark skin tone"_s, u"woman facepalming: dark skin tone"_s})},
{u"person shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FB"), u"shrug_tone1"_s, u"person shrugging: light skin tone"_s})},
{u"person shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FC"), u"shrug_tone2"_s, u"person shrugging: medium-light skin tone"_s})},
{u"person shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FD"), u"shrug_tone3"_s, u"person shrugging: medium skin tone"_s})},
{u"person shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FE"), u"shrug_tone4"_s, u"person shrugging: medium-dark skin tone"_s})},
{u"person shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FF"), u"shrug_tone5"_s, u"person shrugging: dark skin tone"_s})},
{u"man shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man shrugging: light skin tone"_s, u"man shrugging: light skin tone"_s})},
{u"man shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man shrugging: medium-light skin tone"_s, u"man shrugging: medium-light skin tone"_s})},
{u"man shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man shrugging: medium skin tone"_s, u"man shrugging: medium skin tone"_s})},
{u"man shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man shrugging: medium-dark skin tone"_s, u"man shrugging: medium-dark skin tone"_s})},
{u"man shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man shrugging: dark skin tone"_s, u"man shrugging: dark skin tone"_s})},
{u"woman shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman shrugging: light skin tone"_s, u"woman shrugging: light skin tone"_s})},
{u"woman shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman shrugging: medium-light skin tone"_s, u"woman shrugging: medium-light skin tone"_s})},
{u"woman shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman shrugging: medium skin tone"_s, u"woman shrugging: medium skin tone"_s})},
{u"woman shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman shrugging: medium-dark skin tone"_s, u"woman shrugging: medium-dark skin tone"_s})},
{u"woman shrugging"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F937\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman shrugging: dark skin tone"_s, u"woman shrugging: dark skin tone"_s})},
{u"health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002695\U0000FE0F"), u"health worker: light skin tone"_s, u"health worker: light skin tone"_s})},
{u"health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002695\U0000FE0F"), u"health worker: medium-light skin tone"_s, u"health worker: medium-light skin tone"_s})},
{u"health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002695\U0000FE0F"), u"health worker: medium skin tone"_s, u"health worker: medium skin tone"_s})},
{u"health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002695\U0000FE0F"), u"health worker: medium-dark skin tone"_s, u"health worker: medium-dark skin tone"_s})},
{u"health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002695\U0000FE0F"), u"health worker: dark skin tone"_s, u"health worker: dark skin tone"_s})},
{u"man health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002695\U0000FE0F"), u"man health worker: light skin tone"_s, u"man health worker: light skin tone"_s})},
{u"man health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002695\U0000FE0F"), u"man health worker: medium-light skin tone"_s, u"man health worker: medium-light skin tone"_s})},
{u"man health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002695\U0000FE0F"), u"man health worker: medium skin tone"_s, u"man health worker: medium skin tone"_s})},
{u"man health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002695\U0000FE0F"), u"man health worker: medium-dark skin tone"_s, u"man health worker: medium-dark skin tone"_s})},
{u"man health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002695\U0000FE0F"), u"man health worker: dark skin tone"_s, u"man health worker: dark skin tone"_s})},
{u"woman health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002695\U0000FE0F"), u"woman health worker: light skin tone"_s, u"woman health worker: light skin tone"_s})},
{u"woman health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002695\U0000FE0F"), u"woman health worker: medium-light skin tone"_s, u"woman health worker: medium-light skin tone"_s})},
{u"woman health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002695\U0000FE0F"), u"woman health worker: medium skin tone"_s, u"woman health worker: medium skin tone"_s})},
{u"woman health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002695\U0000FE0F"), u"woman health worker: medium-dark skin tone"_s, u"woman health worker: medium-dark skin tone"_s})},
{u"woman health worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002695\U0000FE0F"), u"woman health worker: dark skin tone"_s, u"woman health worker: dark skin tone"_s})},
{u"student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F393"), u"student: light skin tone"_s, u"student: light skin tone"_s})},
{u"student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F393"), u"student: medium-light skin tone"_s, u"student: medium-light skin tone"_s})},
{u"student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F393"), u"student: medium skin tone"_s, u"student: medium skin tone"_s})},
{u"student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F393"), u"student: medium-dark skin tone"_s, u"student: medium-dark skin tone"_s})},
{u"student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F393"), u"student: dark skin tone"_s, u"student: dark skin tone"_s})},
{u"man student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F393"), u"man student: light skin tone"_s, u"man student: light skin tone"_s})},
{u"man student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F393"), u"man student: medium-light skin tone"_s, u"man student: medium-light skin tone"_s})},
{u"man student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F393"), u"man student: medium skin tone"_s, u"man student: medium skin tone"_s})},
{u"man student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F393"), u"man student: medium-dark skin tone"_s, u"man student: medium-dark skin tone"_s})},
{u"man student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F393"), u"man student: dark skin tone"_s, u"man student: dark skin tone"_s})},
{u"woman student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F393"), u"woman student: light skin tone"_s, u"woman student: light skin tone"_s})},
{u"woman student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F393"), u"woman student: medium-light skin tone"_s, u"woman student: medium-light skin tone"_s})},
{u"woman student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F393"), u"woman student: medium skin tone"_s, u"woman student: medium skin tone"_s})},
{u"woman student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F393"), u"woman student: medium-dark skin tone"_s, u"woman student: medium-dark skin tone"_s})},
{u"woman student"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F393"), u"woman student: dark skin tone"_s, u"woman student: dark skin tone"_s})},
{u"teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F3EB"), u"teacher: light skin tone"_s, u"teacher: light skin tone"_s})},
{u"teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F3EB"), u"teacher: medium-light skin tone"_s, u"teacher: medium-light skin tone"_s})},
{u"teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F3EB"), u"teacher: medium skin tone"_s, u"teacher: medium skin tone"_s})},
{u"teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F3EB"), u"teacher: medium-dark skin tone"_s, u"teacher: medium-dark skin tone"_s})},
{u"teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F3EB"), u"teacher: dark skin tone"_s, u"teacher: dark skin tone"_s})},
{u"man teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F3EB"), u"man teacher: light skin tone"_s, u"man teacher: light skin tone"_s})},
{u"man teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F3EB"), u"man teacher: medium-light skin tone"_s, u"man teacher: medium-light skin tone"_s})},
{u"man teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F3EB"), u"man teacher: medium skin tone"_s, u"man teacher: medium skin tone"_s})},
{u"man teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F3EB"), u"man teacher: medium-dark skin tone"_s, u"man teacher: medium-dark skin tone"_s})},
{u"man teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F3EB"), u"man teacher: dark skin tone"_s, u"man teacher: dark skin tone"_s})},
{u"woman teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F3EB"), u"woman teacher: light skin tone"_s, u"woman teacher: light skin tone"_s})},
{u"woman teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F3EB"), u"woman teacher: medium-light skin tone"_s, u"woman teacher: medium-light skin tone"_s})},
{u"woman teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F3EB"), u"woman teacher: medium skin tone"_s, u"woman teacher: medium skin tone"_s})},
{u"woman teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F3EB"), u"woman teacher: medium-dark skin tone"_s, u"woman teacher: medium-dark skin tone"_s})},
{u"woman teacher"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F3EB"), u"woman teacher: dark skin tone"_s, u"woman teacher: dark skin tone"_s})},
{u"judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002696\U0000FE0F"), u"judge: light skin tone"_s, u"judge: light skin tone"_s})},
{u"judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002696\U0000FE0F"), u"judge: medium-light skin tone"_s, u"judge: medium-light skin tone"_s})},
{u"judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002696\U0000FE0F"), u"judge: medium skin tone"_s, u"judge: medium skin tone"_s})},
{u"judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002696\U0000FE0F"), u"judge: medium-dark skin tone"_s, u"judge: medium-dark skin tone"_s})},
{u"judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002696\U0000FE0F"), u"judge: dark skin tone"_s, u"judge: dark skin tone"_s})},
{u"man judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002696\U0000FE0F"), u"man judge: light skin tone"_s, u"man judge: light skin tone"_s})},
{u"man judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002696\U0000FE0F"), u"man judge: medium-light skin tone"_s, u"man judge: medium-light skin tone"_s})},
{u"man judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002696\U0000FE0F"), u"man judge: medium skin tone"_s, u"man judge: medium skin tone"_s})},
{u"man judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002696\U0000FE0F"), u"man judge: medium-dark skin tone"_s, u"man judge: medium-dark skin tone"_s})},
{u"man judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002696\U0000FE0F"), u"man judge: dark skin tone"_s, u"man judge: dark skin tone"_s})},
{u"woman judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002696\U0000FE0F"), u"woman judge: light skin tone"_s, u"woman judge: light skin tone"_s})},
{u"woman judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002696\U0000FE0F"), u"woman judge: medium-light skin tone"_s, u"woman judge: medium-light skin tone"_s})},
{u"woman judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002696\U0000FE0F"), u"woman judge: medium skin tone"_s, u"woman judge: medium skin tone"_s})},
{u"woman judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002696\U0000FE0F"), u"woman judge: medium-dark skin tone"_s, u"woman judge: medium-dark skin tone"_s})},
{u"woman judge"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002696\U0000FE0F"), u"woman judge: dark skin tone"_s, u"woman judge: dark skin tone"_s})},
{u"farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F33E"), u"farmer: light skin tone"_s, u"farmer: light skin tone"_s})},
{u"farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F33E"), u"farmer: medium-light skin tone"_s, u"farmer: medium-light skin tone"_s})},
{u"farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F33E"), u"farmer: medium skin tone"_s, u"farmer: medium skin tone"_s})},
{u"farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F33E"), u"farmer: medium-dark skin tone"_s, u"farmer: medium-dark skin tone"_s})},
{u"farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F33E"), u"farmer: dark skin tone"_s, u"farmer: dark skin tone"_s})},
{u"man farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F33E"), u"man farmer: light skin tone"_s, u"man farmer: light skin tone"_s})},
{u"man farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F33E"), u"man farmer: medium-light skin tone"_s, u"man farmer: medium-light skin tone"_s})},
{u"man farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F33E"), u"man farmer: medium skin tone"_s, u"man farmer: medium skin tone"_s})},
{u"man farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F33E"), u"man farmer: medium-dark skin tone"_s, u"man farmer: medium-dark skin tone"_s})},
{u"man farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F33E"), u"man farmer: dark skin tone"_s, u"man farmer: dark skin tone"_s})},
{u"woman farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F33E"), u"woman farmer: light skin tone"_s, u"woman farmer: light skin tone"_s})},
{u"woman farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F33E"), u"woman farmer: medium-light skin tone"_s, u"woman farmer: medium-light skin tone"_s})},
{u"woman farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F33E"), u"woman farmer: medium skin tone"_s, u"woman farmer: medium skin tone"_s})},
{u"woman farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F33E"), u"woman farmer: medium-dark skin tone"_s, u"woman farmer: medium-dark skin tone"_s})},
{u"woman farmer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F33E"), u"woman farmer: dark skin tone"_s, u"woman farmer: dark skin tone"_s})},
{u"cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F373"), u"cook: light skin tone"_s, u"cook: light skin tone"_s})},
{u"cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F373"), u"cook: medium-light skin tone"_s, u"cook: medium-light skin tone"_s})},
{u"cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F373"), u"cook: medium skin tone"_s, u"cook: medium skin tone"_s})},
{u"cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F373"), u"cook: medium-dark skin tone"_s, u"cook: medium-dark skin tone"_s})},
{u"cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F373"), u"cook: dark skin tone"_s, u"cook: dark skin tone"_s})},
{u"man cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F373"), u"man cook: light skin tone"_s, u"man cook: light skin tone"_s})},
{u"man cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F373"), u"man cook: medium-light skin tone"_s, u"man cook: medium-light skin tone"_s})},
{u"man cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F373"), u"man cook: medium skin tone"_s, u"man cook: medium skin tone"_s})},
{u"man cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F373"), u"man cook: medium-dark skin tone"_s, u"man cook: medium-dark skin tone"_s})},
{u"man cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F373"), u"man cook: dark skin tone"_s, u"man cook: dark skin tone"_s})},
{u"woman cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F373"), u"woman cook: light skin tone"_s, u"woman cook: light skin tone"_s})},
{u"woman cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F373"), u"woman cook: medium-light skin tone"_s, u"woman cook: medium-light skin tone"_s})},
{u"woman cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F373"), u"woman cook: medium skin tone"_s, u"woman cook: medium skin tone"_s})},
{u"woman cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F373"), u"woman cook: medium-dark skin tone"_s, u"woman cook: medium-dark skin tone"_s})},
{u"woman cook"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F373"), u"woman cook: dark skin tone"_s, u"woman cook: dark skin tone"_s})},
{u"mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F527"), u"mechanic: light skin tone"_s, u"mechanic: light skin tone"_s})},
{u"mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F527"), u"mechanic: medium-light skin tone"_s, u"mechanic: medium-light skin tone"_s})},
{u"mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F527"), u"mechanic: medium skin tone"_s, u"mechanic: medium skin tone"_s})},
{u"mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F527"), u"mechanic: medium-dark skin tone"_s, u"mechanic: medium-dark skin tone"_s})},
{u"mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F527"), u"mechanic: dark skin tone"_s, u"mechanic: dark skin tone"_s})},
{u"man mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F527"), u"man mechanic: light skin tone"_s, u"man mechanic: light skin tone"_s})},
{u"man mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F527"), u"man mechanic: medium-light skin tone"_s, u"man mechanic: medium-light skin tone"_s})},
{u"man mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F527"), u"man mechanic: medium skin tone"_s, u"man mechanic: medium skin tone"_s})},
{u"man mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F527"), u"man mechanic: medium-dark skin tone"_s, u"man mechanic: medium-dark skin tone"_s})},
{u"man mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F527"), u"man mechanic: dark skin tone"_s, u"man mechanic: dark skin tone"_s})},
{u"woman mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F527"), u"woman mechanic: light skin tone"_s, u"woman mechanic: light skin tone"_s})},
{u"woman mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F527"), u"woman mechanic: medium-light skin tone"_s, u"woman mechanic: medium-light skin tone"_s})},
{u"woman mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F527"), u"woman mechanic: medium skin tone"_s, u"woman mechanic: medium skin tone"_s})},
{u"woman mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F527"), u"woman mechanic: medium-dark skin tone"_s, u"woman mechanic: medium-dark skin tone"_s})},
{u"woman mechanic"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F527"), u"woman mechanic: dark skin tone"_s, u"woman mechanic: dark skin tone"_s})},
{u"factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F3ED"), u"factory worker: light skin tone"_s, u"factory worker: light skin tone"_s})},
{u"factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F3ED"), u"factory worker: medium-light skin tone"_s, u"factory worker: medium-light skin tone"_s})},
{u"factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F3ED"), u"factory worker: medium skin tone"_s, u"factory worker: medium skin tone"_s})},
{u"factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F3ED"), u"factory worker: medium-dark skin tone"_s, u"factory worker: medium-dark skin tone"_s})},
{u"factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F3ED"), u"factory worker: dark skin tone"_s, u"factory worker: dark skin tone"_s})},
{u"man factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F3ED"), u"man factory worker: light skin tone"_s, u"man factory worker: light skin tone"_s})},
{u"man factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F3ED"), u"man factory worker: medium-light skin tone"_s, u"man factory worker: medium-light skin tone"_s})},
{u"man factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F3ED"), u"man factory worker: medium skin tone"_s, u"man factory worker: medium skin tone"_s})},
{u"man factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F3ED"), u"man factory worker: medium-dark skin tone"_s, u"man factory worker: medium-dark skin tone"_s})},
{u"man factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F3ED"), u"man factory worker: dark skin tone"_s, u"man factory worker: dark skin tone"_s})},
{u"woman factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F3ED"), u"woman factory worker: light skin tone"_s, u"woman factory worker: light skin tone"_s})},
{u"woman factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F3ED"), u"woman factory worker: medium-light skin tone"_s, u"woman factory worker: medium-light skin tone"_s})},
{u"woman factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F3ED"), u"woman factory worker: medium skin tone"_s, u"woman factory worker: medium skin tone"_s})},
{u"woman factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F3ED"), u"woman factory worker: medium-dark skin tone"_s, u"woman factory worker: medium-dark skin tone"_s})},
{u"woman factory worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F3ED"), u"woman factory worker: dark skin tone"_s, u"woman factory worker: dark skin tone"_s})},
{u"office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F4BC"), u"office worker: light skin tone"_s, u"office worker: light skin tone"_s})},
{u"office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F4BC"), u"office worker: medium-light skin tone"_s, u"office worker: medium-light skin tone"_s})},
{u"office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F4BC"), u"office worker: medium skin tone"_s, u"office worker: medium skin tone"_s})},
{u"office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F4BC"), u"office worker: medium-dark skin tone"_s, u"office worker: medium-dark skin tone"_s})},
{u"office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F4BC"), u"office worker: dark skin tone"_s, u"office worker: dark skin tone"_s})},
{u"man office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F4BC"), u"man office worker: light skin tone"_s, u"man office worker: light skin tone"_s})},
{u"man office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F4BC"), u"man office worker: medium-light skin tone"_s, u"man office worker: medium-light skin tone"_s})},
{u"man office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F4BC"), u"man office worker: medium skin tone"_s, u"man office worker: medium skin tone"_s})},
{u"man office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F4BC"), u"man office worker: medium-dark skin tone"_s, u"man office worker: medium-dark skin tone"_s})},
{u"man office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F4BC"), u"man office worker: dark skin tone"_s, u"man office worker: dark skin tone"_s})},
{u"woman office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F4BC"), u"woman office worker: light skin tone"_s, u"woman office worker: light skin tone"_s})},
{u"woman office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F4BC"), u"woman office worker: medium-light skin tone"_s, u"woman office worker: medium-light skin tone"_s})},
{u"woman office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F4BC"), u"woman office worker: medium skin tone"_s, u"woman office worker: medium skin tone"_s})},
{u"woman office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F4BC"), u"woman office worker: medium-dark skin tone"_s, u"woman office worker: medium-dark skin tone"_s})},
{u"woman office worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F4BC"), u"woman office worker: dark skin tone"_s, u"woman office worker: dark skin tone"_s})},
{u"scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F52C"), u"scientist: light skin tone"_s, u"scientist: light skin tone"_s})},
{u"scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F52C"), u"scientist: medium-light skin tone"_s, u"scientist: medium-light skin tone"_s})},
{u"scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F52C"), u"scientist: medium skin tone"_s, u"scientist: medium skin tone"_s})},
{u"scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F52C"), u"scientist: medium-dark skin tone"_s, u"scientist: medium-dark skin tone"_s})},
{u"scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F52C"), u"scientist: dark skin tone"_s, u"scientist: dark skin tone"_s})},
{u"man scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F52C"), u"man scientist: light skin tone"_s, u"man scientist: light skin tone"_s})},
{u"man scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F52C"), u"man scientist: medium-light skin tone"_s, u"man scientist: medium-light skin tone"_s})},
{u"man scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F52C"), u"man scientist: medium skin tone"_s, u"man scientist: medium skin tone"_s})},
{u"man scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F52C"), u"man scientist: medium-dark skin tone"_s, u"man scientist: medium-dark skin tone"_s})},
{u"man scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F52C"), u"man scientist: dark skin tone"_s, u"man scientist: dark skin tone"_s})},
{u"woman scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F52C"), u"woman scientist: light skin tone"_s, u"woman scientist: light skin tone"_s})},
{u"woman scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F52C"), u"woman scientist: medium-light skin tone"_s, u"woman scientist: medium-light skin tone"_s})},
{u"woman scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F52C"), u"woman scientist: medium skin tone"_s, u"woman scientist: medium skin tone"_s})},
{u"woman scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F52C"), u"woman scientist: medium-dark skin tone"_s, u"woman scientist: medium-dark skin tone"_s})},
{u"woman scientist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F52C"), u"woman scientist: dark skin tone"_s, u"woman scientist: dark skin tone"_s})},
{u"technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F4BB"), u"technologist: light skin tone"_s, u"technologist: light skin tone"_s})},
{u"technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F4BB"), u"technologist: medium-light skin tone"_s, u"technologist: medium-light skin tone"_s})},
{u"technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F4BB"), u"technologist: medium skin tone"_s, u"technologist: medium skin tone"_s})},
{u"technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F4BB"), u"technologist: medium-dark skin tone"_s, u"technologist: medium-dark skin tone"_s})},
{u"technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F4BB"), u"technologist: dark skin tone"_s, u"technologist: dark skin tone"_s})},
{u"man technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F4BB"), u"man technologist: light skin tone"_s, u"man technologist: light skin tone"_s})},
{u"man technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F4BB"), u"man technologist: medium-light skin tone"_s, u"man technologist: medium-light skin tone"_s})},
{u"man technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F4BB"), u"man technologist: medium skin tone"_s, u"man technologist: medium skin tone"_s})},
{u"man technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F4BB"), u"man technologist: medium-dark skin tone"_s, u"man technologist: medium-dark skin tone"_s})},
{u"man technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F4BB"), u"man technologist: dark skin tone"_s, u"man technologist: dark skin tone"_s})},
{u"woman technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F4BB"), u"woman technologist: light skin tone"_s, u"woman technologist: light skin tone"_s})},
{u"woman technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F4BB"), u"woman technologist: medium-light skin tone"_s, u"woman technologist: medium-light skin tone"_s})},
{u"woman technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F4BB"), u"woman technologist: medium skin tone"_s, u"woman technologist: medium skin tone"_s})},
{u"woman technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F4BB"), u"woman technologist: medium-dark skin tone"_s, u"woman technologist: medium-dark skin tone"_s})},
{u"woman technologist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F4BB"), u"woman technologist: dark skin tone"_s, u"woman technologist: dark skin tone"_s})},
{u"singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F3A4"), u"singer: light skin tone"_s, u"singer: light skin tone"_s})},
{u"singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F3A4"), u"singer: medium-light skin tone"_s, u"singer: medium-light skin tone"_s})},
{u"singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F3A4"), u"singer: medium skin tone"_s, u"singer: medium skin tone"_s})},
{u"singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F3A4"), u"singer: medium-dark skin tone"_s, u"singer: medium-dark skin tone"_s})},
{u"singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F3A4"), u"singer: dark skin tone"_s, u"singer: dark skin tone"_s})},
{u"man singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F3A4"), u"man singer: light skin tone"_s, u"man singer: light skin tone"_s})},
{u"man singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F3A4"), u"man singer: medium-light skin tone"_s, u"man singer: medium-light skin tone"_s})},
{u"man singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F3A4"), u"man singer: medium skin tone"_s, u"man singer: medium skin tone"_s})},
{u"man singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F3A4"), u"man singer: medium-dark skin tone"_s, u"man singer: medium-dark skin tone"_s})},
{u"man singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F3A4"), u"man singer: dark skin tone"_s, u"man singer: dark skin tone"_s})},
{u"woman singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F3A4"), u"woman singer: light skin tone"_s, u"woman singer: light skin tone"_s})},
{u"woman singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F3A4"), u"woman singer: medium-light skin tone"_s, u"woman singer: medium-light skin tone"_s})},
{u"woman singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F3A4"), u"woman singer: medium skin tone"_s, u"woman singer: medium skin tone"_s})},
{u"woman singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F3A4"), u"woman singer: medium-dark skin tone"_s, u"woman singer: medium-dark skin tone"_s})},
{u"woman singer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F3A4"), u"woman singer: dark skin tone"_s, u"woman singer: dark skin tone"_s})},
{u"artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F3A8"), u"artist: light skin tone"_s, u"artist: light skin tone"_s})},
{u"artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F3A8"), u"artist: medium-light skin tone"_s, u"artist: medium-light skin tone"_s})},
{u"artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F3A8"), u"artist: medium skin tone"_s, u"artist: medium skin tone"_s})},
{u"artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F3A8"), u"artist: medium-dark skin tone"_s, u"artist: medium-dark skin tone"_s})},
{u"artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F3A8"), u"artist: dark skin tone"_s, u"artist: dark skin tone"_s})},
{u"man artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F3A8"), u"man artist: light skin tone"_s, u"man artist: light skin tone"_s})},
{u"man artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F3A8"), u"man artist: medium-light skin tone"_s, u"man artist: medium-light skin tone"_s})},
{u"man artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F3A8"), u"man artist: medium skin tone"_s, u"man artist: medium skin tone"_s})},
{u"man artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F3A8"), u"man artist: medium-dark skin tone"_s, u"man artist: medium-dark skin tone"_s})},
{u"man artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F3A8"), u"man artist: dark skin tone"_s, u"man artist: dark skin tone"_s})},
{u"woman artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F3A8"), u"woman artist: light skin tone"_s, u"woman artist: light skin tone"_s})},
{u"woman artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F3A8"), u"woman artist: medium-light skin tone"_s, u"woman artist: medium-light skin tone"_s})},
{u"woman artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F3A8"), u"woman artist: medium skin tone"_s, u"woman artist: medium skin tone"_s})},
{u"woman artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F3A8"), u"woman artist: medium-dark skin tone"_s, u"woman artist: medium-dark skin tone"_s})},
{u"woman artist"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F3A8"), u"woman artist: dark skin tone"_s, u"woman artist: dark skin tone"_s})},
{u"pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002708\U0000FE0F"), u"pilot: light skin tone"_s, u"pilot: light skin tone"_s})},
{u"pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002708\U0000FE0F"), u"pilot: medium-light skin tone"_s, u"pilot: medium-light skin tone"_s})},
{u"pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002708\U0000FE0F"), u"pilot: medium skin tone"_s, u"pilot: medium skin tone"_s})},
{u"pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002708\U0000FE0F"), u"pilot: medium-dark skin tone"_s, u"pilot: medium-dark skin tone"_s})},
{u"pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002708\U0000FE0F"), u"pilot: dark skin tone"_s, u"pilot: dark skin tone"_s})},
{u"man pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002708\U0000FE0F"), u"man pilot: light skin tone"_s, u"man pilot: light skin tone"_s})},
{u"man pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002708\U0000FE0F"), u"man pilot: medium-light skin tone"_s, u"man pilot: medium-light skin tone"_s})},
{u"man pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002708\U0000FE0F"), u"man pilot: medium skin tone"_s, u"man pilot: medium skin tone"_s})},
{u"man pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002708\U0000FE0F"), u"man pilot: medium-dark skin tone"_s, u"man pilot: medium-dark skin tone"_s})},
{u"man pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002708\U0000FE0F"), u"man pilot: dark skin tone"_s, u"man pilot: dark skin tone"_s})},
{u"woman pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002708\U0000FE0F"), u"woman pilot: light skin tone"_s, u"woman pilot: light skin tone"_s})},
{u"woman pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002708\U0000FE0F"), u"woman pilot: medium-light skin tone"_s, u"woman pilot: medium-light skin tone"_s})},
{u"woman pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002708\U0000FE0F"), u"woman pilot: medium skin tone"_s, u"woman pilot: medium skin tone"_s})},
{u"woman pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002708\U0000FE0F"), u"woman pilot: medium-dark skin tone"_s, u"woman pilot: medium-dark skin tone"_s})},
{u"woman pilot"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002708\U0000FE0F"), u"woman pilot: dark skin tone"_s, u"woman pilot: dark skin tone"_s})},
{u"astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F680"), u"astronaut: light skin tone"_s, u"astronaut: light skin tone"_s})},
{u"astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F680"), u"astronaut: medium-light skin tone"_s, u"astronaut: medium-light skin tone"_s})},
{u"astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F680"), u"astronaut: medium skin tone"_s, u"astronaut: medium skin tone"_s})},
{u"astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F680"), u"astronaut: medium-dark skin tone"_s, u"astronaut: medium-dark skin tone"_s})},
{u"astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F680"), u"astronaut: dark skin tone"_s, u"astronaut: dark skin tone"_s})},
{u"man astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F680"), u"man astronaut: light skin tone"_s, u"man astronaut: light skin tone"_s})},
{u"man astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F680"), u"man astronaut: medium-light skin tone"_s, u"man astronaut: medium-light skin tone"_s})},
{u"man astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F680"), u"man astronaut: medium skin tone"_s, u"man astronaut: medium skin tone"_s})},
{u"man astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F680"), u"man astronaut: medium-dark skin tone"_s, u"man astronaut: medium-dark skin tone"_s})},
{u"man astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F680"), u"man astronaut: dark skin tone"_s, u"man astronaut: dark skin tone"_s})},
{u"woman astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F680"), u"woman astronaut: light skin tone"_s, u"woman astronaut: light skin tone"_s})},
{u"woman astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F680"), u"woman astronaut: medium-light skin tone"_s, u"woman astronaut: medium-light skin tone"_s})},
{u"woman astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F680"), u"woman astronaut: medium skin tone"_s, u"woman astronaut: medium skin tone"_s})},
{u"woman astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F680"), u"woman astronaut: medium-dark skin tone"_s, u"woman astronaut: medium-dark skin tone"_s})},
{u"woman astronaut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F680"), u"woman astronaut: dark skin tone"_s, u"woman astronaut: dark skin tone"_s})},
{u"firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F692"), u"firefighter: light skin tone"_s, u"firefighter: light skin tone"_s})},
{u"firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F692"), u"firefighter: medium-light skin tone"_s, u"firefighter: medium-light skin tone"_s})},
{u"firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F692"), u"firefighter: medium skin tone"_s, u"firefighter: medium skin tone"_s})},
{u"firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F692"), u"firefighter: medium-dark skin tone"_s, u"firefighter: medium-dark skin tone"_s})},
{u"firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F692"), u"firefighter: dark skin tone"_s, u"firefighter: dark skin tone"_s})},
{u"man firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F692"), u"man firefighter: light skin tone"_s, u"man firefighter: light skin tone"_s})},
{u"man firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F692"), u"man firefighter: medium-light skin tone"_s, u"man firefighter: medium-light skin tone"_s})},
{u"man firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F692"), u"man firefighter: medium skin tone"_s, u"man firefighter: medium skin tone"_s})},
{u"man firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F692"), u"man firefighter: medium-dark skin tone"_s, u"man firefighter: medium-dark skin tone"_s})},
{u"man firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F692"), u"man firefighter: dark skin tone"_s, u"man firefighter: dark skin tone"_s})},
{u"woman firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F692"), u"woman firefighter: light skin tone"_s, u"woman firefighter: light skin tone"_s})},
{u"woman firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F692"), u"woman firefighter: medium-light skin tone"_s, u"woman firefighter: medium-light skin tone"_s})},
{u"woman firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F692"), u"woman firefighter: medium skin tone"_s, u"woman firefighter: medium skin tone"_s})},
{u"woman firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F692"), u"woman firefighter: medium-dark skin tone"_s, u"woman firefighter: medium-dark skin tone"_s})},
{u"woman firefighter"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F692"), u"woman firefighter: dark skin tone"_s, u"woman firefighter: dark skin tone"_s})},
{u"police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FB"), u"cop_tone1"_s, u"police officer: light skin tone"_s})},
{u"police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FC"), u"cop_tone2"_s, u"police officer: medium-light skin tone"_s})},
{u"police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FD"), u"cop_tone3"_s, u"police officer: medium skin tone"_s})},
{u"police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FE"), u"cop_tone4"_s, u"police officer: medium-dark skin tone"_s})},
{u"police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FF"), u"cop_tone5"_s, u"police officer: dark skin tone"_s})},
{u"man police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man police officer: light skin tone"_s, u"man police officer: light skin tone"_s})},
{u"man police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man police officer: medium-light skin tone"_s, u"man police officer: medium-light skin tone"_s})},
{u"man police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man police officer: medium skin tone"_s, u"man police officer: medium skin tone"_s})},
{u"man police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man police officer: medium-dark skin tone"_s, u"man police officer: medium-dark skin tone"_s})},
{u"man police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man police officer: dark skin tone"_s, u"man police officer: dark skin tone"_s})},
{u"woman police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman police officer: light skin tone"_s, u"woman police officer: light skin tone"_s})},
{u"woman police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman police officer: medium-light skin tone"_s, u"woman police officer: medium-light skin tone"_s})},
{u"woman police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman police officer: medium skin tone"_s, u"woman police officer: medium skin tone"_s})},
{u"woman police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman police officer: medium-dark skin tone"_s, u"woman police officer: medium-dark skin tone"_s})},
{u"woman police officer"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46E\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman police officer: dark skin tone"_s, u"woman police officer: dark skin tone"_s})},
{u"detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FB"), u"spy_tone1"_s, u"detective: light skin tone"_s})},
{u"detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FC"), u"spy_tone2"_s, u"detective: medium-light skin tone"_s})},
{u"detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FD"), u"spy_tone3"_s, u"detective: medium skin tone"_s})},
{u"detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FE"), u"spy_tone4"_s, u"detective: medium-dark skin tone"_s})},
{u"detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FF"), u"spy_tone5"_s, u"detective: dark skin tone"_s})},
{u"man detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man detective: light skin tone"_s, u"man detective: light skin tone"_s})},
{u"man detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man detective: medium-light skin tone"_s, u"man detective: medium-light skin tone"_s})},
{u"man detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man detective: medium skin tone"_s, u"man detective: medium skin tone"_s})},
{u"man detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man detective: medium-dark skin tone"_s, u"man detective: medium-dark skin tone"_s})},
{u"man detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man detective: dark skin tone"_s, u"man detective: dark skin tone"_s})},
{u"woman detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman detective: light skin tone"_s, u"woman detective: light skin tone"_s})},
{u"woman detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman detective: medium-light skin tone"_s, u"woman detective: medium-light skin tone"_s})},
{u"woman detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman detective: medium skin tone"_s, u"woman detective: medium skin tone"_s})},
{u"woman detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman detective: medium-dark skin tone"_s, u"woman detective: medium-dark skin tone"_s})},
{u"woman detective"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F575\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman detective: dark skin tone"_s, u"woman detective: dark skin tone"_s})},
{u"guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FB"), u"guardsman_tone1"_s, u"guard: light skin tone"_s})},
{u"guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FC"), u"guardsman_tone2"_s, u"guard: medium-light skin tone"_s})},
{u"guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FD"), u"guardsman_tone3"_s, u"guard: medium skin tone"_s})},
{u"guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FE"), u"guardsman_tone4"_s, u"guard: medium-dark skin tone"_s})},
{u"guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FF"), u"guardsman_tone5"_s, u"guard: dark skin tone"_s})},
{u"man guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man guard: light skin tone"_s, u"man guard: light skin tone"_s})},
{u"man guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man guard: medium-light skin tone"_s, u"man guard: medium-light skin tone"_s})},
{u"man guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man guard: medium skin tone"_s, u"man guard: medium skin tone"_s})},
{u"man guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man guard: medium-dark skin tone"_s, u"man guard: medium-dark skin tone"_s})},
{u"man guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man guard: dark skin tone"_s, u"man guard: dark skin tone"_s})},
{u"woman guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman guard: light skin tone"_s, u"woman guard: light skin tone"_s})},
{u"woman guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman guard: medium-light skin tone"_s, u"woman guard: medium-light skin tone"_s})},
{u"woman guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman guard: medium skin tone"_s, u"woman guard: medium skin tone"_s})},
{u"woman guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman guard: medium-dark skin tone"_s, u"woman guard: medium-dark skin tone"_s})},
{u"woman guard"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F482\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman guard: dark skin tone"_s, u"woman guard: dark skin tone"_s})},
{u"ninja"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F977\U0001F3FB"), u"ninja: light skin tone"_s, u"ninja: light skin tone"_s})},
{u"ninja"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F977\U0001F3FC"), u"ninja: medium-light skin tone"_s, u"ninja: medium-light skin tone"_s})},
{u"ninja"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F977\U0001F3FD"), u"ninja: medium skin tone"_s, u"ninja: medium skin tone"_s})},
{u"ninja"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F977\U0001F3FE"), u"ninja: medium-dark skin tone"_s, u"ninja: medium-dark skin tone"_s})},
{u"ninja"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F977\U0001F3FF"), u"ninja: dark skin tone"_s, u"ninja: dark skin tone"_s})},
{u"construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FB"), u"construction_worker_tone1"_s, u"construction worker: light skin tone"_s})},
{u"construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FC"), u"construction_worker_tone2"_s, u"construction worker: medium-light skin tone"_s})},
{u"construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FD"), u"construction_worker_tone3"_s, u"construction worker: medium skin tone"_s})},
{u"construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FE"), u"construction_worker_tone4"_s, u"construction worker: medium-dark skin tone"_s})},
{u"construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FF"), u"construction_worker_tone5"_s, u"construction worker: dark skin tone"_s})},
{u"man construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man construction worker: light skin tone"_s, u"man construction worker: light skin tone"_s})},
{u"man construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man construction worker: medium-light skin tone"_s, u"man construction worker: medium-light skin tone"_s})},
{u"man construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man construction worker: medium skin tone"_s, u"man construction worker: medium skin tone"_s})},
{u"man construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man construction worker: medium-dark skin tone"_s, u"man construction worker: medium-dark skin tone"_s})},
{u"man construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man construction worker: dark skin tone"_s, u"man construction worker: dark skin tone"_s})},
{u"woman construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman construction worker: light skin tone"_s, u"woman construction worker: light skin tone"_s})},
{u"woman construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman construction worker: medium-light skin tone"_s, u"woman construction worker: medium-light skin tone"_s})},
{u"woman construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman construction worker: medium skin tone"_s, u"woman construction worker: medium skin tone"_s})},
{u"woman construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman construction worker: medium-dark skin tone"_s, u"woman construction worker: medium-dark skin tone"_s})},
{u"woman construction worker"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F477\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman construction worker: dark skin tone"_s, u"woman construction worker: dark skin tone"_s})},
{u"person with crown"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC5\U0001F3FB"), u"person with crown: light skin tone"_s, u"person with crown: light skin tone"_s})},
{u"person with crown"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC5\U0001F3FC"), u"person with crown: medium-light skin tone"_s, u"person with crown: medium-light skin tone"_s})},
{u"person with crown"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC5\U0001F3FD"), u"person with crown: medium skin tone"_s, u"person with crown: medium skin tone"_s})},
{u"person with crown"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC5\U0001F3FE"), u"person with crown: medium-dark skin tone"_s, u"person with crown: medium-dark skin tone"_s})},
{u"person with crown"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC5\U0001F3FF"), u"person with crown: dark skin tone"_s, u"person with crown: dark skin tone"_s})},
{u"prince"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F934\U0001F3FB"), u"prince_tone1"_s, u"prince: light skin tone"_s})},
{u"prince"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F934\U0001F3FC"), u"prince_tone2"_s, u"prince: medium-light skin tone"_s})},
{u"prince"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F934\U0001F3FD"), u"prince_tone3"_s, u"prince: medium skin tone"_s})},
{u"prince"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F934\U0001F3FE"), u"prince_tone4"_s, u"prince: medium-dark skin tone"_s})},
{u"prince"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F934\U0001F3FF"), u"prince_tone5"_s, u"prince: dark skin tone"_s})},
{u"princess"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F478\U0001F3FB"), u"princess_tone1"_s, u"princess: light skin tone"_s})},
{u"princess"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F478\U0001F3FC"), u"princess_tone2"_s, u"princess: medium-light skin tone"_s})},
{u"princess"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F478\U0001F3FD"), u"princess_tone3"_s, u"princess: medium skin tone"_s})},
{u"princess"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F478\U0001F3FE"), u"princess_tone4"_s, u"princess: medium-dark skin tone"_s})},
{u"princess"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F478\U0001F3FF"), u"princess_tone5"_s, u"princess: dark skin tone"_s})},
{u"person wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FB"), u"man_with_turban_tone1"_s, u"person wearing turban: light skin tone"_s})},
{u"person wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FC"), u"man_with_turban_tone2"_s, u"person wearing turban: medium-light skin tone"_s})},
{u"person wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FD"), u"man_with_turban_tone3"_s, u"person wearing turban: medium skin tone"_s})},
{u"person wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FE"), u"man_with_turban_tone4"_s, u"person wearing turban: medium-dark skin tone"_s})},
{u"person wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FF"), u"man_with_turban_tone5"_s, u"person wearing turban: dark skin tone"_s})},
{u"man wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man wearing turban: light skin tone"_s, u"man wearing turban: light skin tone"_s})},
{u"man wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man wearing turban: medium-light skin tone"_s, u"man wearing turban: medium-light skin tone"_s})},
{u"man wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man wearing turban: medium skin tone"_s, u"man wearing turban: medium skin tone"_s})},
{u"man wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man wearing turban: medium-dark skin tone"_s, u"man wearing turban: medium-dark skin tone"_s})},
{u"man wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man wearing turban: dark skin tone"_s, u"man wearing turban: dark skin tone"_s})},
{u"woman wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman wearing turban: light skin tone"_s, u"woman wearing turban: light skin tone"_s})},
{u"woman wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman wearing turban: medium-light skin tone"_s, u"woman wearing turban: medium-light skin tone"_s})},
{u"woman wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman wearing turban: medium skin tone"_s, u"woman wearing turban: medium skin tone"_s})},
{u"woman wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman wearing turban: medium-dark skin tone"_s, u"woman wearing turban: medium-dark skin tone"_s})},
{u"woman wearing turban"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F473\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman wearing turban: dark skin tone"_s, u"woman wearing turban: dark skin tone"_s})},
{u"person with skullcap"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F472\U0001F3FB"), u"man_with_gua_pi_mao_tone1"_s, u"person with skullcap: light skin tone"_s})},
{u"person with skullcap"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F472\U0001F3FC"), u"man_with_gua_pi_mao_tone2"_s, u"person with skullcap: medium-light skin tone"_s})},
{u"person with skullcap"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F472\U0001F3FD"), u"man_with_gua_pi_mao_tone3"_s, u"person with skullcap: medium skin tone"_s})},
{u"person with skullcap"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F472\U0001F3FE"), u"man_with_gua_pi_mao_tone4"_s, u"person with skullcap: medium-dark skin tone"_s})},
{u"person with skullcap"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F472\U0001F3FF"), u"man_with_gua_pi_mao_tone5"_s, u"person with skullcap: dark skin tone"_s})},
{u"woman with headscarf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D5\U0001F3FB"), u"woman_with_headscarf_tone1"_s, u"woman with headscarf: light skin tone"_s})},
{u"woman with headscarf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D5\U0001F3FC"), u"woman_with_headscarf_tone2"_s, u"woman with headscarf: medium-light skin tone"_s})},
{u"woman with headscarf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D5\U0001F3FD"), u"woman_with_headscarf_tone3"_s, u"woman with headscarf: medium skin tone"_s})},
{u"woman with headscarf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D5\U0001F3FE"), u"woman_with_headscarf_tone4"_s, u"woman with headscarf: medium-dark skin tone"_s})},
{u"woman with headscarf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D5\U0001F3FF"), u"woman_with_headscarf_tone5"_s, u"woman with headscarf: dark skin tone"_s})},
{u"person in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FB"), u"man_in_tuxedo_tone1"_s, u"person in tuxedo: light skin tone"_s})},
{u"person in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FC"), u"man_in_tuxedo_tone2"_s, u"person in tuxedo: medium-light skin tone"_s})},
{u"person in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FD"), u"man_in_tuxedo_tone3"_s, u"person in tuxedo: medium skin tone"_s})},
{u"person in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FE"), u"man_in_tuxedo_tone4"_s, u"person in tuxedo: medium-dark skin tone"_s})},
{u"person in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FF"), u"man_in_tuxedo_tone5"_s, u"person in tuxedo: dark skin tone"_s})},
{u"man in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man in tuxedo: light skin tone"_s, u"man in tuxedo: light skin tone"_s})},
{u"man in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man in tuxedo: medium-light skin tone"_s, u"man in tuxedo: medium-light skin tone"_s})},
{u"man in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man in tuxedo: medium skin tone"_s, u"man in tuxedo: medium skin tone"_s})},
{u"man in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man in tuxedo: medium-dark skin tone"_s, u"man in tuxedo: medium-dark skin tone"_s})},
{u"man in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man in tuxedo: dark skin tone"_s, u"man in tuxedo: dark skin tone"_s})},
{u"woman in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman in tuxedo: light skin tone"_s, u"woman in tuxedo: light skin tone"_s})},
{u"woman in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman in tuxedo: medium-light skin tone"_s, u"woman in tuxedo: medium-light skin tone"_s})},
{u"woman in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman in tuxedo: medium skin tone"_s, u"woman in tuxedo: medium skin tone"_s})},
{u"woman in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman in tuxedo: medium-dark skin tone"_s, u"woman in tuxedo: medium-dark skin tone"_s})},
{u"woman in tuxedo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F935\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman in tuxedo: dark skin tone"_s, u"woman in tuxedo: dark skin tone"_s})},
{u"person with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FB"), u"bride_with_veil_tone1"_s, u"person with veil: light skin tone"_s})},
{u"person with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FC"), u"bride_with_veil_tone2"_s, u"person with veil: medium-light skin tone"_s})},
{u"person with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FD"), u"bride_with_veil_tone3"_s, u"person with veil: medium skin tone"_s})},
{u"person with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FE"), u"bride_with_veil_tone4"_s, u"person with veil: medium-dark skin tone"_s})},
{u"person with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FF"), u"bride_with_veil_tone5"_s, u"person with veil: dark skin tone"_s})},
{u"man with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man with veil: light skin tone"_s, u"man with veil: light skin tone"_s})},
{u"man with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man with veil: medium-light skin tone"_s, u"man with veil: medium-light skin tone"_s})},
{u"man with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man with veil: medium skin tone"_s, u"man with veil: medium skin tone"_s})},
{u"man with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man with veil: medium-dark skin tone"_s, u"man with veil: medium-dark skin tone"_s})},
{u"man with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man with veil: dark skin tone"_s, u"man with veil: dark skin tone"_s})},
{u"woman with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman with veil: light skin tone"_s, u"woman with veil: light skin tone"_s})},
{u"woman with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman with veil: medium-light skin tone"_s, u"woman with veil: medium-light skin tone"_s})},
{u"woman with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman with veil: medium skin tone"_s, u"woman with veil: medium skin tone"_s})},
{u"woman with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman with veil: medium-dark skin tone"_s, u"woman with veil: medium-dark skin tone"_s})},
{u"woman with veil"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F470\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman with veil: dark skin tone"_s, u"woman with veil: dark skin tone"_s})},
{u"pregnant woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F930\U0001F3FB"), u"pregnant_woman_tone1"_s, u"pregnant woman: light skin tone"_s})},
{u"pregnant woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F930\U0001F3FC"), u"pregnant_woman_tone2"_s, u"pregnant woman: medium-light skin tone"_s})},
{u"pregnant woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F930\U0001F3FD"), u"pregnant_woman_tone3"_s, u"pregnant woman: medium skin tone"_s})},
{u"pregnant woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F930\U0001F3FE"), u"pregnant_woman_tone4"_s, u"pregnant woman: medium-dark skin tone"_s})},
{u"pregnant woman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F930\U0001F3FF"), u"pregnant_woman_tone5"_s, u"pregnant woman: dark skin tone"_s})},
{u"pregnant man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC3\U0001F3FB"), u"pregnant man: light skin tone"_s, u"pregnant man: light skin tone"_s})},
{u"pregnant man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC3\U0001F3FC"), u"pregnant man: medium-light skin tone"_s, u"pregnant man: medium-light skin tone"_s})},
{u"pregnant man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC3\U0001F3FD"), u"pregnant man: medium skin tone"_s, u"pregnant man: medium skin tone"_s})},
{u"pregnant man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC3\U0001F3FE"), u"pregnant man: medium-dark skin tone"_s, u"pregnant man: medium-dark skin tone"_s})},
{u"pregnant man"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC3\U0001F3FF"), u"pregnant man: dark skin tone"_s, u"pregnant man: dark skin tone"_s})},
{u"pregnant person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC4\U0001F3FB"), u"pregnant person: light skin tone"_s, u"pregnant person: light skin tone"_s})},
{u"pregnant person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC4\U0001F3FC"), u"pregnant person: medium-light skin tone"_s, u"pregnant person: medium-light skin tone"_s})},
{u"pregnant person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC4\U0001F3FD"), u"pregnant person: medium skin tone"_s, u"pregnant person: medium skin tone"_s})},
{u"pregnant person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC4\U0001F3FE"), u"pregnant person: medium-dark skin tone"_s, u"pregnant person: medium-dark skin tone"_s})},
{u"pregnant person"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001FAC4\U0001F3FF"), u"pregnant person: dark skin tone"_s, u"pregnant person: dark skin tone"_s})},
{u"breast-feeding"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F931\U0001F3FB"), u"breast_feeding_tone1"_s, u"breast-feeding: light skin tone"_s})},
{u"breast-feeding"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F931\U0001F3FC"), u"breast_feeding_tone2"_s, u"breast-feeding: medium-light skin tone"_s})},
{u"breast-feeding"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F931\U0001F3FD"), u"breast_feeding_tone3"_s, u"breast-feeding: medium skin tone"_s})},
{u"breast-feeding"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F931\U0001F3FE"), u"breast_feeding_tone4"_s, u"breast-feeding: medium-dark skin tone"_s})},
{u"breast-feeding"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F931\U0001F3FF"), u"breast_feeding_tone5"_s, u"breast-feeding: dark skin tone"_s})},
{u"woman feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F37C"), u"woman feeding baby: light skin tone"_s, u"woman feeding baby: light skin tone"_s})},
{u"woman feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F37C"), u"woman feeding baby: medium-light skin tone"_s, u"woman feeding baby: medium-light skin tone"_s})},
{u"woman feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F37C"), u"woman feeding baby: medium skin tone"_s, u"woman feeding baby: medium skin tone"_s})},
{u"woman feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F37C"), u"woman feeding baby: medium-dark skin tone"_s, u"woman feeding baby: medium-dark skin tone"_s})},
{u"woman feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F37C"), u"woman feeding baby: dark skin tone"_s, u"woman feeding baby: dark skin tone"_s})},
{u"man feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F37C"), u"man feeding baby: light skin tone"_s, u"man feeding baby: light skin tone"_s})},
{u"man feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F37C"), u"man feeding baby: medium-light skin tone"_s, u"man feeding baby: medium-light skin tone"_s})},
{u"man feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F37C"), u"man feeding baby: medium skin tone"_s, u"man feeding baby: medium skin tone"_s})},
{u"man feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F37C"), u"man feeding baby: medium-dark skin tone"_s, u"man feeding baby: medium-dark skin tone"_s})},
{u"man feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F37C"), u"man feeding baby: dark skin tone"_s, u"man feeding baby: dark skin tone"_s})},
{u"person feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F37C"), u"person feeding baby: light skin tone"_s, u"person feeding baby: light skin tone"_s})},
{u"person feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F37C"), u"person feeding baby: medium-light skin tone"_s, u"person feeding baby: medium-light skin tone"_s})},
{u"person feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F37C"), u"person feeding baby: medium skin tone"_s, u"person feeding baby: medium skin tone"_s})},
{u"person feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F37C"), u"person feeding baby: medium-dark skin tone"_s, u"person feeding baby: medium-dark skin tone"_s})},
{u"person feeding baby"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F37C"), u"person feeding baby: dark skin tone"_s, u"person feeding baby: dark skin tone"_s})},
{u"baby angel"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F47C\U0001F3FB"), u"angel_tone1"_s, u"baby angel: light skin tone"_s})},
{u"baby angel"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F47C\U0001F3FC"), u"angel_tone2"_s, u"baby angel: medium-light skin tone"_s})},
{u"baby angel"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F47C\U0001F3FD"), u"angel_tone3"_s, u"baby angel: medium skin tone"_s})},
{u"baby angel"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F47C\U0001F3FE"), u"angel_tone4"_s, u"baby angel: medium-dark skin tone"_s})},
{u"baby angel"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F47C\U0001F3FF"), u"angel_tone5"_s, u"baby angel: dark skin tone"_s})},
{u"Santa Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F385\U0001F3FB"), u"santa_tone1"_s, u"Santa Claus: light skin tone"_s})},
{u"Santa Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F385\U0001F3FC"), u"santa_tone2"_s, u"Santa Claus: medium-light skin tone"_s})},
{u"Santa Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F385\U0001F3FD"), u"santa_tone3"_s, u"Santa Claus: medium skin tone"_s})},
{u"Santa Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F385\U0001F3FE"), u"santa_tone4"_s, u"Santa Claus: medium-dark skin tone"_s})},
{u"Santa Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F385\U0001F3FF"), u"santa_tone5"_s, u"Santa Claus: dark skin tone"_s})},
{u"Mrs. Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F936\U0001F3FB"), u"mrs_claus_tone1"_s, u"Mrs. Claus: light skin tone"_s})},
{u"Mrs. Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F936\U0001F3FC"), u"mrs_claus_tone2"_s, u"Mrs. Claus: medium-light skin tone"_s})},
{u"Mrs. Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F936\U0001F3FD"), u"mrs_claus_tone3"_s, u"Mrs. Claus: medium skin tone"_s})},
{u"Mrs. Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F936\U0001F3FE"), u"mrs_claus_tone4"_s, u"Mrs. Claus: medium-dark skin tone"_s})},
{u"Mrs. Claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F936\U0001F3FF"), u"mrs_claus_tone5"_s, u"Mrs. Claus: dark skin tone"_s})},
{u"mx claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F384"), u"mx claus: light skin tone"_s, u"mx claus: light skin tone"_s})},
{u"mx claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F384"), u"mx claus: medium-light skin tone"_s, u"mx claus: medium-light skin tone"_s})},
{u"mx claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F384"), u"mx claus: medium skin tone"_s, u"mx claus: medium skin tone"_s})},
{u"mx claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F384"), u"mx claus: medium-dark skin tone"_s, u"mx claus: medium-dark skin tone"_s})},
{u"mx claus"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F384"), u"mx claus: dark skin tone"_s, u"mx claus: dark skin tone"_s})},
{u"superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FB"), u"superhero_tone1"_s, u"superhero: light skin tone"_s})},
{u"superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FC"), u"superhero_tone2"_s, u"superhero: medium-light skin tone"_s})},
{u"superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FD"), u"superhero_tone3"_s, u"superhero: medium skin tone"_s})},
{u"superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FE"), u"superhero_tone4"_s, u"superhero: medium-dark skin tone"_s})},
{u"superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FF"), u"superhero_tone5"_s, u"superhero: dark skin tone"_s})},
{u"man superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man superhero: light skin tone"_s, u"man superhero: light skin tone"_s})},
{u"man superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man superhero: medium-light skin tone"_s, u"man superhero: medium-light skin tone"_s})},
{u"man superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man superhero: medium skin tone"_s, u"man superhero: medium skin tone"_s})},
{u"man superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man superhero: medium-dark skin tone"_s, u"man superhero: medium-dark skin tone"_s})},
{u"man superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man superhero: dark skin tone"_s, u"man superhero: dark skin tone"_s})},
{u"woman superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman superhero: light skin tone"_s, u"woman superhero: light skin tone"_s})},
{u"woman superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman superhero: medium-light skin tone"_s, u"woman superhero: medium-light skin tone"_s})},
{u"woman superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman superhero: medium skin tone"_s, u"woman superhero: medium skin tone"_s})},
{u"woman superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman superhero: medium-dark skin tone"_s, u"woman superhero: medium-dark skin tone"_s})},
{u"woman superhero"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B8\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman superhero: dark skin tone"_s, u"woman superhero: dark skin tone"_s})},
{u"supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FB"), u"supervillain_tone1"_s, u"supervillain: light skin tone"_s})},
{u"supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FC"), u"supervillain_tone2"_s, u"supervillain: medium-light skin tone"_s})},
{u"supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FD"), u"supervillain_tone3"_s, u"supervillain: medium skin tone"_s})},
{u"supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FE"), u"supervillain_tone4"_s, u"supervillain: medium-dark skin tone"_s})},
{u"supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FF"), u"supervillain_tone5"_s, u"supervillain: dark skin tone"_s})},
{u"man supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man supervillain: light skin tone"_s, u"man supervillain: light skin tone"_s})},
{u"man supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man supervillain: medium-light skin tone"_s, u"man supervillain: medium-light skin tone"_s})},
{u"man supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man supervillain: medium skin tone"_s, u"man supervillain: medium skin tone"_s})},
{u"man supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man supervillain: medium-dark skin tone"_s, u"man supervillain: medium-dark skin tone"_s})},
{u"man supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man supervillain: dark skin tone"_s, u"man supervillain: dark skin tone"_s})},
{u"woman supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman supervillain: light skin tone"_s, u"woman supervillain: light skin tone"_s})},
{u"woman supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman supervillain: medium-light skin tone"_s, u"woman supervillain: medium-light skin tone"_s})},
{u"woman supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman supervillain: medium skin tone"_s, u"woman supervillain: medium skin tone"_s})},
{u"woman supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman supervillain: medium-dark skin tone"_s, u"woman supervillain: medium-dark skin tone"_s})},
{u"woman supervillain"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9B9\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman supervillain: dark skin tone"_s, u"woman supervillain: dark skin tone"_s})},
{u"mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FB"), u"mage_tone1"_s, u"mage: light skin tone"_s})},
{u"mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FC"), u"mage_tone2"_s, u"mage: medium-light skin tone"_s})},
{u"mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FD"), u"mage_tone3"_s, u"mage: medium skin tone"_s})},
{u"mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FE"), u"mage_tone4"_s, u"mage: medium-dark skin tone"_s})},
{u"mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FF"), u"mage_tone5"_s, u"mage: dark skin tone"_s})},
{u"man mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man mage: light skin tone"_s, u"man mage: light skin tone"_s})},
{u"man mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man mage: medium-light skin tone"_s, u"man mage: medium-light skin tone"_s})},
{u"man mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man mage: medium skin tone"_s, u"man mage: medium skin tone"_s})},
{u"man mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man mage: medium-dark skin tone"_s, u"man mage: medium-dark skin tone"_s})},
{u"man mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man mage: dark skin tone"_s, u"man mage: dark skin tone"_s})},
{u"woman mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman mage: light skin tone"_s, u"woman mage: light skin tone"_s})},
{u"woman mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman mage: medium-light skin tone"_s, u"woman mage: medium-light skin tone"_s})},
{u"woman mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman mage: medium skin tone"_s, u"woman mage: medium skin tone"_s})},
{u"woman mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman mage: medium-dark skin tone"_s, u"woman mage: medium-dark skin tone"_s})},
{u"woman mage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D9\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman mage: dark skin tone"_s, u"woman mage: dark skin tone"_s})},
{u"fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FB"), u"fairy_tone1"_s, u"fairy: light skin tone"_s})},
{u"fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FC"), u"fairy_tone2"_s, u"fairy: medium-light skin tone"_s})},
{u"fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FD"), u"fairy_tone3"_s, u"fairy: medium skin tone"_s})},
{u"fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FE"), u"fairy_tone4"_s, u"fairy: medium-dark skin tone"_s})},
{u"fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FF"), u"fairy_tone5"_s, u"fairy: dark skin tone"_s})},
{u"man fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man fairy: light skin tone"_s, u"man fairy: light skin tone"_s})},
{u"man fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man fairy: medium-light skin tone"_s, u"man fairy: medium-light skin tone"_s})},
{u"man fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man fairy: medium skin tone"_s, u"man fairy: medium skin tone"_s})},
{u"man fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man fairy: medium-dark skin tone"_s, u"man fairy: medium-dark skin tone"_s})},
{u"man fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man fairy: dark skin tone"_s, u"man fairy: dark skin tone"_s})},
{u"woman fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman fairy: light skin tone"_s, u"woman fairy: light skin tone"_s})},
{u"woman fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman fairy: medium-light skin tone"_s, u"woman fairy: medium-light skin tone"_s})},
{u"woman fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman fairy: medium skin tone"_s, u"woman fairy: medium skin tone"_s})},
{u"woman fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman fairy: medium-dark skin tone"_s, u"woman fairy: medium-dark skin tone"_s})},
{u"woman fairy"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DA\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman fairy: dark skin tone"_s, u"woman fairy: dark skin tone"_s})},
{u"vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FB"), u"vampire_tone1"_s, u"vampire: light skin tone"_s})},
{u"vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FC"), u"vampire_tone2"_s, u"vampire: medium-light skin tone"_s})},
{u"vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FD"), u"vampire_tone3"_s, u"vampire: medium skin tone"_s})},
{u"vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FE"), u"vampire_tone4"_s, u"vampire: medium-dark skin tone"_s})},
{u"vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FF"), u"vampire_tone5"_s, u"vampire: dark skin tone"_s})},
{u"man vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man vampire: light skin tone"_s, u"man vampire: light skin tone"_s})},
{u"man vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man vampire: medium-light skin tone"_s, u"man vampire: medium-light skin tone"_s})},
{u"man vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man vampire: medium skin tone"_s, u"man vampire: medium skin tone"_s})},
{u"man vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man vampire: medium-dark skin tone"_s, u"man vampire: medium-dark skin tone"_s})},
{u"man vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man vampire: dark skin tone"_s, u"man vampire: dark skin tone"_s})},
{u"woman vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman vampire: light skin tone"_s, u"woman vampire: light skin tone"_s})},
{u"woman vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman vampire: medium-light skin tone"_s, u"woman vampire: medium-light skin tone"_s})},
{u"woman vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman vampire: medium skin tone"_s, u"woman vampire: medium skin tone"_s})},
{u"woman vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman vampire: medium-dark skin tone"_s, u"woman vampire: medium-dark skin tone"_s})},
{u"woman vampire"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DB\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman vampire: dark skin tone"_s, u"woman vampire: dark skin tone"_s})},
{u"merperson"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FB"), u"merperson_tone1"_s, u"merperson: light skin tone"_s})},
{u"merperson"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FC"), u"merperson_tone2"_s, u"merperson: medium-light skin tone"_s})},
{u"merperson"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FD"), u"merperson_tone3"_s, u"merperson: medium skin tone"_s})},
{u"merperson"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FE"), u"merperson_tone4"_s, u"merperson: medium-dark skin tone"_s})},
{u"merperson"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FF"), u"merperson_tone5"_s, u"merperson: dark skin tone"_s})},
{u"merman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"merman: light skin tone"_s, u"merman: light skin tone"_s})},
{u"merman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"merman: medium-light skin tone"_s, u"merman: medium-light skin tone"_s})},
{u"merman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"merman: medium skin tone"_s, u"merman: medium skin tone"_s})},
{u"merman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"merman: medium-dark skin tone"_s, u"merman: medium-dark skin tone"_s})},
{u"merman"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"merman: dark skin tone"_s, u"merman: dark skin tone"_s})},
{u"mermaid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"mermaid: light skin tone"_s, u"mermaid: light skin tone"_s})},
{u"mermaid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"mermaid: medium-light skin tone"_s, u"mermaid: medium-light skin tone"_s})},
{u"mermaid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"mermaid: medium skin tone"_s, u"mermaid: medium skin tone"_s})},
{u"mermaid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"mermaid: medium-dark skin tone"_s, u"mermaid: medium-dark skin tone"_s})},
{u"mermaid"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DC\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"mermaid: dark skin tone"_s, u"mermaid: dark skin tone"_s})},
{u"elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FB"), u"elf_tone1"_s, u"elf: light skin tone"_s})},
{u"elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FC"), u"elf_tone2"_s, u"elf: medium-light skin tone"_s})},
{u"elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FD"), u"elf_tone3"_s, u"elf: medium skin tone"_s})},
{u"elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FE"), u"elf_tone4"_s, u"elf: medium-dark skin tone"_s})},
{u"elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FF"), u"elf_tone5"_s, u"elf: dark skin tone"_s})},
{u"man elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man elf: light skin tone"_s, u"man elf: light skin tone"_s})},
{u"man elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man elf: medium-light skin tone"_s, u"man elf: medium-light skin tone"_s})},
{u"man elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man elf: medium skin tone"_s, u"man elf: medium skin tone"_s})},
{u"man elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man elf: medium-dark skin tone"_s, u"man elf: medium-dark skin tone"_s})},
{u"man elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man elf: dark skin tone"_s, u"man elf: dark skin tone"_s})},
{u"woman elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman elf: light skin tone"_s, u"woman elf: light skin tone"_s})},
{u"woman elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman elf: medium-light skin tone"_s, u"woman elf: medium-light skin tone"_s})},
{u"woman elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman elf: medium skin tone"_s, u"woman elf: medium skin tone"_s})},
{u"woman elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman elf: medium-dark skin tone"_s, u"woman elf: medium-dark skin tone"_s})},
{u"woman elf"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9DD\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman elf: dark skin tone"_s, u"woman elf: dark skin tone"_s})},
{u"person getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FB"), u"massage_tone1"_s, u"person getting massage: light skin tone"_s})},
{u"person getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FC"), u"massage_tone2"_s, u"person getting massage: medium-light skin tone"_s})},
{u"person getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FD"), u"massage_tone3"_s, u"person getting massage: medium skin tone"_s})},
{u"person getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FE"), u"massage_tone4"_s, u"person getting massage: medium-dark skin tone"_s})},
{u"person getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FF"), u"massage_tone5"_s, u"person getting massage: dark skin tone"_s})},
{u"man getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man getting massage: light skin tone"_s, u"man getting massage: light skin tone"_s})},
{u"man getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man getting massage: medium-light skin tone"_s, u"man getting massage: medium-light skin tone"_s})},
{u"man getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man getting massage: medium skin tone"_s, u"man getting massage: medium skin tone"_s})},
{u"man getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man getting massage: medium-dark skin tone"_s, u"man getting massage: medium-dark skin tone"_s})},
{u"man getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man getting massage: dark skin tone"_s, u"man getting massage: dark skin tone"_s})},
{u"woman getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman getting massage: light skin tone"_s, u"woman getting massage: light skin tone"_s})},
{u"woman getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman getting massage: medium-light skin tone"_s, u"woman getting massage: medium-light skin tone"_s})},
{u"woman getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman getting massage: medium skin tone"_s, u"woman getting massage: medium skin tone"_s})},
{u"woman getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman getting massage: medium-dark skin tone"_s, u"woman getting massage: medium-dark skin tone"_s})},
{u"woman getting massage"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F486\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman getting massage: dark skin tone"_s, u"woman getting massage: dark skin tone"_s})},
{u"person getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FB"), u"haircut_tone1"_s, u"person getting haircut: light skin tone"_s})},
{u"person getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FC"), u"haircut_tone2"_s, u"person getting haircut: medium-light skin tone"_s})},
{u"person getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FD"), u"haircut_tone3"_s, u"person getting haircut: medium skin tone"_s})},
{u"person getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FE"), u"haircut_tone4"_s, u"person getting haircut: medium-dark skin tone"_s})},
{u"person getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FF"), u"haircut_tone5"_s, u"person getting haircut: dark skin tone"_s})},
{u"man getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man getting haircut: light skin tone"_s, u"man getting haircut: light skin tone"_s})},
{u"man getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man getting haircut: medium-light skin tone"_s, u"man getting haircut: medium-light skin tone"_s})},
{u"man getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man getting haircut: medium skin tone"_s, u"man getting haircut: medium skin tone"_s})},
{u"man getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man getting haircut: medium-dark skin tone"_s, u"man getting haircut: medium-dark skin tone"_s})},
{u"man getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man getting haircut: dark skin tone"_s, u"man getting haircut: dark skin tone"_s})},
{u"woman getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman getting haircut: light skin tone"_s, u"woman getting haircut: light skin tone"_s})},
{u"woman getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman getting haircut: medium-light skin tone"_s, u"woman getting haircut: medium-light skin tone"_s})},
{u"woman getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman getting haircut: medium skin tone"_s, u"woman getting haircut: medium skin tone"_s})},
{u"woman getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman getting haircut: medium-dark skin tone"_s, u"woman getting haircut: medium-dark skin tone"_s})},
{u"woman getting haircut"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F487\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman getting haircut: dark skin tone"_s, u"woman getting haircut: dark skin tone"_s})},
{u"person walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FB"), u"walking_tone1"_s, u"person walking: light skin tone"_s})},
{u"person walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FC"), u"walking_tone2"_s, u"person walking: medium-light skin tone"_s})},
{u"person walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FD"), u"walking_tone3"_s, u"person walking: medium skin tone"_s})},
{u"person walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FE"), u"walking_tone4"_s, u"person walking: medium-dark skin tone"_s})},
{u"person walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FF"), u"walking_tone5"_s, u"person walking: dark skin tone"_s})},
{u"man walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man walking: light skin tone"_s, u"man walking: light skin tone"_s})},
{u"man walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man walking: medium-light skin tone"_s, u"man walking: medium-light skin tone"_s})},
{u"man walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man walking: medium skin tone"_s, u"man walking: medium skin tone"_s})},
{u"man walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man walking: medium-dark skin tone"_s, u"man walking: medium-dark skin tone"_s})},
{u"man walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man walking: dark skin tone"_s, u"man walking: dark skin tone"_s})},
{u"woman walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman walking: light skin tone"_s, u"woman walking: light skin tone"_s})},
{u"woman walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman walking: medium-light skin tone"_s, u"woman walking: medium-light skin tone"_s})},
{u"woman walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman walking: medium skin tone"_s, u"woman walking: medium skin tone"_s})},
{u"woman walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman walking: medium-dark skin tone"_s, u"woman walking: medium-dark skin tone"_s})},
{u"woman walking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B6\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman walking: dark skin tone"_s, u"woman walking: dark skin tone"_s})},
{u"person standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FB"), u"person standing: light skin tone"_s, u"person standing: light skin tone"_s})},
{u"person standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FC"), u"person standing: medium-light skin tone"_s, u"person standing: medium-light skin tone"_s})},
{u"person standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FD"), u"person standing: medium skin tone"_s, u"person standing: medium skin tone"_s})},
{u"person standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FE"), u"person standing: medium-dark skin tone"_s, u"person standing: medium-dark skin tone"_s})},
{u"person standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FF"), u"person standing: dark skin tone"_s, u"person standing: dark skin tone"_s})},
{u"man standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man standing: light skin tone"_s, u"man standing: light skin tone"_s})},
{u"man standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man standing: medium-light skin tone"_s, u"man standing: medium-light skin tone"_s})},
{u"man standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man standing: medium skin tone"_s, u"man standing: medium skin tone"_s})},
{u"man standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man standing: medium-dark skin tone"_s, u"man standing: medium-dark skin tone"_s})},
{u"man standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man standing: dark skin tone"_s, u"man standing: dark skin tone"_s})},
{u"woman standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman standing: light skin tone"_s, u"woman standing: light skin tone"_s})},
{u"woman standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman standing: medium-light skin tone"_s, u"woman standing: medium-light skin tone"_s})},
{u"woman standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman standing: medium skin tone"_s, u"woman standing: medium skin tone"_s})},
{u"woman standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman standing: medium-dark skin tone"_s, u"woman standing: medium-dark skin tone"_s})},
{u"woman standing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CD\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman standing: dark skin tone"_s, u"woman standing: dark skin tone"_s})},
{u"person kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FB"), u"person kneeling: light skin tone"_s, u"person kneeling: light skin tone"_s})},
{u"person kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FC"), u"person kneeling: medium-light skin tone"_s, u"person kneeling: medium-light skin tone"_s})},
{u"person kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FD"), u"person kneeling: medium skin tone"_s, u"person kneeling: medium skin tone"_s})},
{u"person kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FE"), u"person kneeling: medium-dark skin tone"_s, u"person kneeling: medium-dark skin tone"_s})},
{u"person kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FF"), u"person kneeling: dark skin tone"_s, u"person kneeling: dark skin tone"_s})},
{u"man kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man kneeling: light skin tone"_s, u"man kneeling: light skin tone"_s})},
{u"man kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man kneeling: medium-light skin tone"_s, u"man kneeling: medium-light skin tone"_s})},
{u"man kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man kneeling: medium skin tone"_s, u"man kneeling: medium skin tone"_s})},
{u"man kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man kneeling: medium-dark skin tone"_s, u"man kneeling: medium-dark skin tone"_s})},
{u"man kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man kneeling: dark skin tone"_s, u"man kneeling: dark skin tone"_s})},
{u"woman kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman kneeling: light skin tone"_s, u"woman kneeling: light skin tone"_s})},
{u"woman kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman kneeling: medium-light skin tone"_s, u"woman kneeling: medium-light skin tone"_s})},
{u"woman kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman kneeling: medium skin tone"_s, u"woman kneeling: medium skin tone"_s})},
{u"woman kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman kneeling: medium-dark skin tone"_s, u"woman kneeling: medium-dark skin tone"_s})},
{u"woman kneeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9CE\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman kneeling: dark skin tone"_s, u"woman kneeling: dark skin tone"_s})},
{u"person with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F9AF"), u"person with white cane: light skin tone"_s, u"person with white cane: light skin tone"_s})},
{u"person with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F9AF"), u"person with white cane: medium-light skin tone"_s, u"person with white cane: medium-light skin tone"_s})},
{u"person with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F9AF"), u"person with white cane: medium skin tone"_s, u"person with white cane: medium skin tone"_s})},
{u"person with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F9AF"), u"person with white cane: medium-dark skin tone"_s, u"person with white cane: medium-dark skin tone"_s})},
{u"person with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F9AF"), u"person with white cane: dark skin tone"_s, u"person with white cane: dark skin tone"_s})},
{u"man with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F9AF"), u"man with white cane: light skin tone"_s, u"man with white cane: light skin tone"_s})},
{u"man with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F9AF"), u"man with white cane: medium-light skin tone"_s, u"man with white cane: medium-light skin tone"_s})},
{u"man with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F9AF"), u"man with white cane: medium skin tone"_s, u"man with white cane: medium skin tone"_s})},
{u"man with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F9AF"), u"man with white cane: medium-dark skin tone"_s, u"man with white cane: medium-dark skin tone"_s})},
{u"man with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F9AF"), u"man with white cane: dark skin tone"_s, u"man with white cane: dark skin tone"_s})},
{u"woman with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F9AF"), u"woman with white cane: light skin tone"_s, u"woman with white cane: light skin tone"_s})},
{u"woman with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F9AF"), u"woman with white cane: medium-light skin tone"_s, u"woman with white cane: medium-light skin tone"_s})},
{u"woman with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F9AF"), u"woman with white cane: medium skin tone"_s, u"woman with white cane: medium skin tone"_s})},
{u"woman with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F9AF"), u"woman with white cane: medium-dark skin tone"_s, u"woman with white cane: medium-dark skin tone"_s})},
{u"woman with white cane"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F9AF"), u"woman with white cane: dark skin tone"_s, u"woman with white cane: dark skin tone"_s})},
{u"person in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F9BC"), u"person in motorized wheelchair: light skin tone"_s, u"person in motorized wheelchair: light skin tone"_s})},
{u"person in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F9BC"), u"person in motorized wheelchair: medium-light skin tone"_s, u"person in motorized wheelchair: medium-light skin tone"_s})},
{u"person in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F9BC"), u"person in motorized wheelchair: medium skin tone"_s, u"person in motorized wheelchair: medium skin tone"_s})},
{u"person in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F9BC"), u"person in motorized wheelchair: medium-dark skin tone"_s, u"person in motorized wheelchair: medium-dark skin tone"_s})},
{u"person in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F9BC"), u"person in motorized wheelchair: dark skin tone"_s, u"person in motorized wheelchair: dark skin tone"_s})},
{u"man in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F9BC"), u"man in motorized wheelchair: light skin tone"_s, u"man in motorized wheelchair: light skin tone"_s})},
{u"man in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F9BC"), u"man in motorized wheelchair: medium-light skin tone"_s, u"man in motorized wheelchair: medium-light skin tone"_s})},
{u"man in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F9BC"), u"man in motorized wheelchair: medium skin tone"_s, u"man in motorized wheelchair: medium skin tone"_s})},
{u"man in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F9BC"), u"man in motorized wheelchair: medium-dark skin tone"_s, u"man in motorized wheelchair: medium-dark skin tone"_s})},
{u"man in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F9BC"), u"man in motorized wheelchair: dark skin tone"_s, u"man in motorized wheelchair: dark skin tone"_s})},
{u"woman in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F9BC"), u"woman in motorized wheelchair: light skin tone"_s, u"woman in motorized wheelchair: light skin tone"_s})},
{u"woman in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F9BC"), u"woman in motorized wheelchair: medium-light skin tone"_s, u"woman in motorized wheelchair: medium-light skin tone"_s})},
{u"woman in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F9BC"), u"woman in motorized wheelchair: medium skin tone"_s, u"woman in motorized wheelchair: medium skin tone"_s})},
{u"woman in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F9BC"), u"woman in motorized wheelchair: medium-dark skin tone"_s, u"woman in motorized wheelchair: medium-dark skin tone"_s})},
{u"woman in motorized wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F9BC"), u"woman in motorized wheelchair: dark skin tone"_s, u"woman in motorized wheelchair: dark skin tone"_s})},
{u"person in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F9BD"), u"person in manual wheelchair: light skin tone"_s, u"person in manual wheelchair: light skin tone"_s})},
{u"person in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F9BD"), u"person in manual wheelchair: medium-light skin tone"_s, u"person in manual wheelchair: medium-light skin tone"_s})},
{u"person in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F9BD"), u"person in manual wheelchair: medium skin tone"_s, u"person in manual wheelchair: medium skin tone"_s})},
{u"person in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F9BD"), u"person in manual wheelchair: medium-dark skin tone"_s, u"person in manual wheelchair: medium-dark skin tone"_s})},
{u"person in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F9BD"), u"person in manual wheelchair: dark skin tone"_s, u"person in manual wheelchair: dark skin tone"_s})},
{u"man in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F9BD"), u"man in manual wheelchair: light skin tone"_s, u"man in manual wheelchair: light skin tone"_s})},
{u"man in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F9BD"), u"man in manual wheelchair: medium-light skin tone"_s, u"man in manual wheelchair: medium-light skin tone"_s})},
{u"man in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F9BD"), u"man in manual wheelchair: medium skin tone"_s, u"man in manual wheelchair: medium skin tone"_s})},
{u"man in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F9BD"), u"man in manual wheelchair: medium-dark skin tone"_s, u"man in manual wheelchair: medium-dark skin tone"_s})},
{u"man in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F9BD"), u"man in manual wheelchair: dark skin tone"_s, u"man in manual wheelchair: dark skin tone"_s})},
{u"woman in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F9BD"), u"woman in manual wheelchair: light skin tone"_s, u"woman in manual wheelchair: light skin tone"_s})},
{u"woman in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F9BD"), u"woman in manual wheelchair: medium-light skin tone"_s, u"woman in manual wheelchair: medium-light skin tone"_s})},
{u"woman in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F9BD"), u"woman in manual wheelchair: medium skin tone"_s, u"woman in manual wheelchair: medium skin tone"_s})},
{u"woman in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F9BD"), u"woman in manual wheelchair: medium-dark skin tone"_s, u"woman in manual wheelchair: medium-dark skin tone"_s})},
{u"woman in manual wheelchair"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F9BD"), u"woman in manual wheelchair: dark skin tone"_s, u"woman in manual wheelchair: dark skin tone"_s})},
{u"person running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FB"), u"runner_tone1"_s, u"person running: light skin tone"_s})},
{u"person running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FC"), u"runner_tone2"_s, u"person running: medium-light skin tone"_s})},
{u"person running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FD"), u"runner_tone3"_s, u"person running: medium skin tone"_s})},
{u"person running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FE"), u"runner_tone4"_s, u"person running: medium-dark skin tone"_s})},
{u"person running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FF"), u"runner_tone5"_s, u"person running: dark skin tone"_s})},
{u"man running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man running: light skin tone"_s, u"man running: light skin tone"_s})},
{u"man running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man running: medium-light skin tone"_s, u"man running: medium-light skin tone"_s})},
{u"man running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man running: medium skin tone"_s, u"man running: medium skin tone"_s})},
{u"man running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man running: medium-dark skin tone"_s, u"man running: medium-dark skin tone"_s})},
{u"man running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man running: dark skin tone"_s, u"man running: dark skin tone"_s})},
{u"woman running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman running: light skin tone"_s, u"woman running: light skin tone"_s})},
{u"woman running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman running: medium-light skin tone"_s, u"woman running: medium-light skin tone"_s})},
{u"woman running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman running: medium skin tone"_s, u"woman running: medium skin tone"_s})},
{u"woman running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman running: medium-dark skin tone"_s, u"woman running: medium-dark skin tone"_s})},
{u"woman running"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C3\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman running: dark skin tone"_s, u"woman running: dark skin tone"_s})},
{u"woman dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F483\U0001F3FB"), u"dancer_tone1"_s, u"woman dancing: light skin tone"_s})},
{u"woman dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F483\U0001F3FC"), u"dancer_tone2"_s, u"woman dancing: medium-light skin tone"_s})},
{u"woman dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F483\U0001F3FD"), u"dancer_tone3"_s, u"woman dancing: medium skin tone"_s})},
{u"woman dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F483\U0001F3FE"), u"dancer_tone4"_s, u"woman dancing: medium-dark skin tone"_s})},
{u"woman dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F483\U0001F3FF"), u"dancer_tone5"_s, u"woman dancing: dark skin tone"_s})},
{u"man dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F57A\U0001F3FB"), u"man_dancing_tone1"_s, u"man dancing: light skin tone"_s})},
{u"man dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F57A\U0001F3FC"), u"man_dancing_tone2"_s, u"man dancing: medium-light skin tone"_s})},
{u"man dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F57A\U0001F3FD"), u"man_dancing_tone3"_s, u"man dancing: medium skin tone"_s})},
{u"man dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F57A\U0001F3FE"), u"man_dancing_tone4"_s, u"man dancing: medium-dark skin tone"_s})},
{u"man dancing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F57A\U0001F3FF"), u"man_dancing_tone5"_s, u"man dancing: dark skin tone"_s})},
{u"person in suit levitating"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F574\U0001F3FB"), u"levitate_tone1"_s, u"person in suit levitating: light skin tone"_s})},
{u"person in suit levitating"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F574\U0001F3FC"), u"levitate_tone2"_s, u"person in suit levitating: medium-light skin tone"_s})},
{u"person in suit levitating"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F574\U0001F3FD"), u"levitate_tone3"_s, u"person in suit levitating: medium skin tone"_s})},
{u"person in suit levitating"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F574\U0001F3FE"), u"levitate_tone4"_s, u"person in suit levitating: medium-dark skin tone"_s})},
{u"person in suit levitating"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F574\U0001F3FF"), u"levitate_tone5"_s, u"person in suit levitating: dark skin tone"_s})},
{u"person in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FB"), u"person_in_steamy_room_tone1"_s, u"person in steamy room: light skin tone"_s})},
{u"person in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FC"), u"person_in_steamy_room_tone2"_s, u"person in steamy room: medium-light skin tone"_s})},
{u"person in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FD"), u"person_in_steamy_room_tone3"_s, u"person in steamy room: medium skin tone"_s})},
{u"person in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FE"), u"person_in_steamy_room_tone4"_s, u"person in steamy room: medium-dark skin tone"_s})},
{u"person in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FF"), u"person_in_steamy_room_tone5"_s, u"person in steamy room: dark skin tone"_s})},
{u"man in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man in steamy room: light skin tone"_s, u"man in steamy room: light skin tone"_s})},
{u"man in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man in steamy room: medium-light skin tone"_s, u"man in steamy room: medium-light skin tone"_s})},
{u"man in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man in steamy room: medium skin tone"_s, u"man in steamy room: medium skin tone"_s})},
{u"man in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man in steamy room: medium-dark skin tone"_s, u"man in steamy room: medium-dark skin tone"_s})},
{u"man in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man in steamy room: dark skin tone"_s, u"man in steamy room: dark skin tone"_s})},
{u"woman in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman in steamy room: light skin tone"_s, u"woman in steamy room: light skin tone"_s})},
{u"woman in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman in steamy room: medium-light skin tone"_s, u"woman in steamy room: medium-light skin tone"_s})},
{u"woman in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman in steamy room: medium skin tone"_s, u"woman in steamy room: medium skin tone"_s})},
{u"woman in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman in steamy room: medium-dark skin tone"_s, u"woman in steamy room: medium-dark skin tone"_s})},
{u"woman in steamy room"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D6\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman in steamy room: dark skin tone"_s, u"woman in steamy room: dark skin tone"_s})},
{u"person climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FB"), u"person_climbing_tone1"_s, u"person climbing: light skin tone"_s})},
{u"person climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FC"), u"person_climbing_tone2"_s, u"person climbing: medium-light skin tone"_s})},
{u"person climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FD"), u"person_climbing_tone3"_s, u"person climbing: medium skin tone"_s})},
{u"person climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FE"), u"person_climbing_tone4"_s, u"person climbing: medium-dark skin tone"_s})},
{u"person climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FF"), u"person_climbing_tone5"_s, u"person climbing: dark skin tone"_s})},
{u"man climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man climbing: light skin tone"_s, u"man climbing: light skin tone"_s})},
{u"man climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man climbing: medium-light skin tone"_s, u"man climbing: medium-light skin tone"_s})},
{u"man climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man climbing: medium skin tone"_s, u"man climbing: medium skin tone"_s})},
{u"man climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man climbing: medium-dark skin tone"_s, u"man climbing: medium-dark skin tone"_s})},
{u"man climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man climbing: dark skin tone"_s, u"man climbing: dark skin tone"_s})},
{u"woman climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman climbing: light skin tone"_s, u"woman climbing: light skin tone"_s})},
{u"woman climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman climbing: medium-light skin tone"_s, u"woman climbing: medium-light skin tone"_s})},
{u"woman climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman climbing: medium skin tone"_s, u"woman climbing: medium skin tone"_s})},
{u"woman climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman climbing: medium-dark skin tone"_s, u"woman climbing: medium-dark skin tone"_s})},
{u"woman climbing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D7\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman climbing: dark skin tone"_s, u"woman climbing: dark skin tone"_s})},
{u"horse racing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C7\U0001F3FB"), u"horse_racing_tone1"_s, u"horse racing: light skin tone"_s})},
{u"horse racing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C7\U0001F3FC"), u"horse_racing_tone2"_s, u"horse racing: medium-light skin tone"_s})},
{u"horse racing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C7\U0001F3FD"), u"horse_racing_tone3"_s, u"horse racing: medium skin tone"_s})},
{u"horse racing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C7\U0001F3FE"), u"horse_racing_tone4"_s, u"horse racing: medium-dark skin tone"_s})},
{u"horse racing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C7\U0001F3FF"), u"horse_racing_tone5"_s, u"horse racing: dark skin tone"_s})},
{u"snowboarder"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C2\U0001F3FB"), u"snowboarder_tone1"_s, u"snowboarder: light skin tone"_s})},
{u"snowboarder"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C2\U0001F3FC"), u"snowboarder_tone2"_s, u"snowboarder: medium-light skin tone"_s})},
{u"snowboarder"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C2\U0001F3FD"), u"snowboarder_tone3"_s, u"snowboarder: medium skin tone"_s})},
{u"snowboarder"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C2\U0001F3FE"), u"snowboarder_tone4"_s, u"snowboarder: medium-dark skin tone"_s})},
{u"snowboarder"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C2\U0001F3FF"), u"snowboarder_tone5"_s, u"snowboarder: dark skin tone"_s})},
{u"person golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FB"), u"person_golfing_tone1"_s, u"person golfing: light skin tone"_s})},
{u"person golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FC"), u"person_golfing_tone2"_s, u"person golfing: medium-light skin tone"_s})},
{u"person golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FD"), u"person_golfing_tone3"_s, u"person golfing: medium skin tone"_s})},
{u"person golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FE"), u"person_golfing_tone4"_s, u"person golfing: medium-dark skin tone"_s})},
{u"person golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FF"), u"person_golfing_tone5"_s, u"person golfing: dark skin tone"_s})},
{u"man golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man golfing: light skin tone"_s, u"man golfing: light skin tone"_s})},
{u"man golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man golfing: medium-light skin tone"_s, u"man golfing: medium-light skin tone"_s})},
{u"man golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man golfing: medium skin tone"_s, u"man golfing: medium skin tone"_s})},
{u"man golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man golfing: medium-dark skin tone"_s, u"man golfing: medium-dark skin tone"_s})},
{u"man golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man golfing: dark skin tone"_s, u"man golfing: dark skin tone"_s})},
{u"woman golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman golfing: light skin tone"_s, u"woman golfing: light skin tone"_s})},
{u"woman golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman golfing: medium-light skin tone"_s, u"woman golfing: medium-light skin tone"_s})},
{u"woman golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman golfing: medium skin tone"_s, u"woman golfing: medium skin tone"_s})},
{u"woman golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman golfing: medium-dark skin tone"_s, u"woman golfing: medium-dark skin tone"_s})},
{u"woman golfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CC\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman golfing: dark skin tone"_s, u"woman golfing: dark skin tone"_s})},
{u"person surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FB"), u"surfer_tone1"_s, u"person surfing: light skin tone"_s})},
{u"person surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FC"), u"surfer_tone2"_s, u"person surfing: medium-light skin tone"_s})},
{u"person surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FD"), u"surfer_tone3"_s, u"person surfing: medium skin tone"_s})},
{u"person surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FE"), u"surfer_tone4"_s, u"person surfing: medium-dark skin tone"_s})},
{u"person surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FF"), u"surfer_tone5"_s, u"person surfing: dark skin tone"_s})},
{u"man surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man surfing: light skin tone"_s, u"man surfing: light skin tone"_s})},
{u"man surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man surfing: medium-light skin tone"_s, u"man surfing: medium-light skin tone"_s})},
{u"man surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man surfing: medium skin tone"_s, u"man surfing: medium skin tone"_s})},
{u"man surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man surfing: medium-dark skin tone"_s, u"man surfing: medium-dark skin tone"_s})},
{u"man surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man surfing: dark skin tone"_s, u"man surfing: dark skin tone"_s})},
{u"woman surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman surfing: light skin tone"_s, u"woman surfing: light skin tone"_s})},
{u"woman surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman surfing: medium-light skin tone"_s, u"woman surfing: medium-light skin tone"_s})},
{u"woman surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman surfing: medium skin tone"_s, u"woman surfing: medium skin tone"_s})},
{u"woman surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman surfing: medium-dark skin tone"_s, u"woman surfing: medium-dark skin tone"_s})},
{u"woman surfing"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3C4\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman surfing: dark skin tone"_s, u"woman surfing: dark skin tone"_s})},
{u"person rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FB"), u"rowboat_tone1"_s, u"person rowing boat: light skin tone"_s})},
{u"person rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FC"), u"rowboat_tone2"_s, u"person rowing boat: medium-light skin tone"_s})},
{u"person rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FD"), u"rowboat_tone3"_s, u"person rowing boat: medium skin tone"_s})},
{u"person rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FE"), u"rowboat_tone4"_s, u"person rowing boat: medium-dark skin tone"_s})},
{u"person rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FF"), u"rowboat_tone5"_s, u"person rowing boat: dark skin tone"_s})},
{u"man rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man rowing boat: light skin tone"_s, u"man rowing boat: light skin tone"_s})},
{u"man rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man rowing boat: medium-light skin tone"_s, u"man rowing boat: medium-light skin tone"_s})},
{u"man rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man rowing boat: medium skin tone"_s, u"man rowing boat: medium skin tone"_s})},
{u"man rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man rowing boat: medium-dark skin tone"_s, u"man rowing boat: medium-dark skin tone"_s})},
{u"man rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man rowing boat: dark skin tone"_s, u"man rowing boat: dark skin tone"_s})},
{u"woman rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman rowing boat: light skin tone"_s, u"woman rowing boat: light skin tone"_s})},
{u"woman rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman rowing boat: medium-light skin tone"_s, u"woman rowing boat: medium-light skin tone"_s})},
{u"woman rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman rowing boat: medium skin tone"_s, u"woman rowing boat: medium skin tone"_s})},
{u"woman rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman rowing boat: medium-dark skin tone"_s, u"woman rowing boat: medium-dark skin tone"_s})},
{u"woman rowing boat"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6A3\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman rowing boat: dark skin tone"_s, u"woman rowing boat: dark skin tone"_s})},
{u"person swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FB"), u"swimmer_tone1"_s, u"person swimming: light skin tone"_s})},
{u"person swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FC"), u"swimmer_tone2"_s, u"person swimming: medium-light skin tone"_s})},
{u"person swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FD"), u"swimmer_tone3"_s, u"person swimming: medium skin tone"_s})},
{u"person swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FE"), u"swimmer_tone4"_s, u"person swimming: medium-dark skin tone"_s})},
{u"person swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FF"), u"swimmer_tone5"_s, u"person swimming: dark skin tone"_s})},
{u"man swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man swimming: light skin tone"_s, u"man swimming: light skin tone"_s})},
{u"man swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man swimming: medium-light skin tone"_s, u"man swimming: medium-light skin tone"_s})},
{u"man swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man swimming: medium skin tone"_s, u"man swimming: medium skin tone"_s})},
{u"man swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man swimming: medium-dark skin tone"_s, u"man swimming: medium-dark skin tone"_s})},
{u"man swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man swimming: dark skin tone"_s, u"man swimming: dark skin tone"_s})},
{u"woman swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman swimming: light skin tone"_s, u"woman swimming: light skin tone"_s})},
{u"woman swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman swimming: medium-light skin tone"_s, u"woman swimming: medium-light skin tone"_s})},
{u"woman swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman swimming: medium skin tone"_s, u"woman swimming: medium skin tone"_s})},
{u"woman swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman swimming: medium-dark skin tone"_s, u"woman swimming: medium-dark skin tone"_s})},
{u"woman swimming"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CA\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman swimming: dark skin tone"_s, u"woman swimming: dark skin tone"_s})},
{u"person bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FB"), u"basketball_player_tone1"_s, u"person bouncing ball: light skin tone"_s})},
{u"person bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FC"), u"basketball_player_tone2"_s, u"person bouncing ball: medium-light skin tone"_s})},
{u"person bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FD"), u"basketball_player_tone3"_s, u"person bouncing ball: medium skin tone"_s})},
{u"person bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FE"), u"basketball_player_tone4"_s, u"person bouncing ball: medium-dark skin tone"_s})},
{u"person bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FF"), u"basketball_player_tone5"_s, u"person bouncing ball: dark skin tone"_s})},
{u"man bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man bouncing ball: light skin tone"_s, u"man bouncing ball: light skin tone"_s})},
{u"man bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man bouncing ball: medium-light skin tone"_s, u"man bouncing ball: medium-light skin tone"_s})},
{u"man bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man bouncing ball: medium skin tone"_s, u"man bouncing ball: medium skin tone"_s})},
{u"man bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man bouncing ball: medium-dark skin tone"_s, u"man bouncing ball: medium-dark skin tone"_s})},
{u"man bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man bouncing ball: dark skin tone"_s, u"man bouncing ball: dark skin tone"_s})},
{u"woman bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman bouncing ball: light skin tone"_s, u"woman bouncing ball: light skin tone"_s})},
{u"woman bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman bouncing ball: medium-light skin tone"_s, u"woman bouncing ball: medium-light skin tone"_s})},
{u"woman bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman bouncing ball: medium skin tone"_s, u"woman bouncing ball: medium skin tone"_s})},
{u"woman bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman bouncing ball: medium-dark skin tone"_s, u"woman bouncing ball: medium-dark skin tone"_s})},
{u"woman bouncing ball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U000026F9\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman bouncing ball: dark skin tone"_s, u"woman bouncing ball: dark skin tone"_s})},
{u"person lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FB"), u"lifter_tone1"_s, u"person lifting weights: light skin tone"_s})},
{u"person lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FC"), u"lifter_tone2"_s, u"person lifting weights: medium-light skin tone"_s})},
{u"person lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FD"), u"lifter_tone3"_s, u"person lifting weights: medium skin tone"_s})},
{u"person lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FE"), u"lifter_tone4"_s, u"person lifting weights: medium-dark skin tone"_s})},
{u"person lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FF"), u"lifter_tone5"_s, u"person lifting weights: dark skin tone"_s})},
{u"man lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man lifting weights: light skin tone"_s, u"man lifting weights: light skin tone"_s})},
{u"man lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man lifting weights: medium-light skin tone"_s, u"man lifting weights: medium-light skin tone"_s})},
{u"man lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man lifting weights: medium skin tone"_s, u"man lifting weights: medium skin tone"_s})},
{u"man lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man lifting weights: medium-dark skin tone"_s, u"man lifting weights: medium-dark skin tone"_s})},
{u"man lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man lifting weights: dark skin tone"_s, u"man lifting weights: dark skin tone"_s})},
{u"woman lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman lifting weights: light skin tone"_s, u"woman lifting weights: light skin tone"_s})},
{u"woman lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman lifting weights: medium-light skin tone"_s, u"woman lifting weights: medium-light skin tone"_s})},
{u"woman lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman lifting weights: medium skin tone"_s, u"woman lifting weights: medium skin tone"_s})},
{u"woman lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman lifting weights: medium-dark skin tone"_s, u"woman lifting weights: medium-dark skin tone"_s})},
{u"woman lifting weights"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3CB\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman lifting weights: dark skin tone"_s, u"woman lifting weights: dark skin tone"_s})},
{u"person biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FB"), u"bicyclist_tone1"_s, u"person biking: light skin tone"_s})},
{u"person biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FC"), u"bicyclist_tone2"_s, u"person biking: medium-light skin tone"_s})},
{u"person biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FD"), u"bicyclist_tone3"_s, u"person biking: medium skin tone"_s})},
{u"person biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FE"), u"bicyclist_tone4"_s, u"person biking: medium-dark skin tone"_s})},
{u"person biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FF"), u"bicyclist_tone5"_s, u"person biking: dark skin tone"_s})},
{u"man biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man biking: light skin tone"_s, u"man biking: light skin tone"_s})},
{u"man biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man biking: medium-light skin tone"_s, u"man biking: medium-light skin tone"_s})},
{u"man biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man biking: medium skin tone"_s, u"man biking: medium skin tone"_s})},
{u"man biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man biking: medium-dark skin tone"_s, u"man biking: medium-dark skin tone"_s})},
{u"man biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man biking: dark skin tone"_s, u"man biking: dark skin tone"_s})},
{u"woman biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman biking: light skin tone"_s, u"woman biking: light skin tone"_s})},
{u"woman biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman biking: medium-light skin tone"_s, u"woman biking: medium-light skin tone"_s})},
{u"woman biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman biking: medium skin tone"_s, u"woman biking: medium skin tone"_s})},
{u"woman biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman biking: medium-dark skin tone"_s, u"woman biking: medium-dark skin tone"_s})},
{u"woman biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B4\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman biking: dark skin tone"_s, u"woman biking: dark skin tone"_s})},
{u"person mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FB"), u"mountain_bicyclist_tone1"_s, u"person mountain biking: light skin tone"_s})},
{u"person mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FC"), u"mountain_bicyclist_tone2"_s, u"person mountain biking: medium-light skin tone"_s})},
{u"person mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FD"), u"mountain_bicyclist_tone3"_s, u"person mountain biking: medium skin tone"_s})},
{u"person mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FE"), u"mountain_bicyclist_tone4"_s, u"person mountain biking: medium-dark skin tone"_s})},
{u"person mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FF"), u"mountain_bicyclist_tone5"_s, u"person mountain biking: dark skin tone"_s})},
{u"man mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man mountain biking: light skin tone"_s, u"man mountain biking: light skin tone"_s})},
{u"man mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man mountain biking: medium-light skin tone"_s, u"man mountain biking: medium-light skin tone"_s})},
{u"man mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man mountain biking: medium skin tone"_s, u"man mountain biking: medium skin tone"_s})},
{u"man mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man mountain biking: medium-dark skin tone"_s, u"man mountain biking: medium-dark skin tone"_s})},
{u"man mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man mountain biking: dark skin tone"_s, u"man mountain biking: dark skin tone"_s})},
{u"woman mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman mountain biking: light skin tone"_s, u"woman mountain biking: light skin tone"_s})},
{u"woman mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman mountain biking: medium-light skin tone"_s, u"woman mountain biking: medium-light skin tone"_s})},
{u"woman mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman mountain biking: medium skin tone"_s, u"woman mountain biking: medium skin tone"_s})},
{u"woman mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman mountain biking: medium-dark skin tone"_s, u"woman mountain biking: medium-dark skin tone"_s})},
{u"woman mountain biking"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6B5\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman mountain biking: dark skin tone"_s, u"woman mountain biking: dark skin tone"_s})},
{u"person cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FB"), u"cartwheel_tone1"_s, u"person cartwheeling: light skin tone"_s})},
{u"person cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FC"), u"cartwheel_tone2"_s, u"person cartwheeling: medium-light skin tone"_s})},
{u"person cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FD"), u"cartwheel_tone3"_s, u"person cartwheeling: medium skin tone"_s})},
{u"person cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FE"), u"cartwheel_tone4"_s, u"person cartwheeling: medium-dark skin tone"_s})},
{u"person cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FF"), u"cartwheel_tone5"_s, u"person cartwheeling: dark skin tone"_s})},
{u"man cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man cartwheeling: light skin tone"_s, u"man cartwheeling: light skin tone"_s})},
{u"man cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man cartwheeling: medium-light skin tone"_s, u"man cartwheeling: medium-light skin tone"_s})},
{u"man cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man cartwheeling: medium skin tone"_s, u"man cartwheeling: medium skin tone"_s})},
{u"man cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man cartwheeling: medium-dark skin tone"_s, u"man cartwheeling: medium-dark skin tone"_s})},
{u"man cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man cartwheeling: dark skin tone"_s, u"man cartwheeling: dark skin tone"_s})},
{u"woman cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman cartwheeling: light skin tone"_s, u"woman cartwheeling: light skin tone"_s})},
{u"woman cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman cartwheeling: medium-light skin tone"_s, u"woman cartwheeling: medium-light skin tone"_s})},
{u"woman cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman cartwheeling: medium skin tone"_s, u"woman cartwheeling: medium skin tone"_s})},
{u"woman cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman cartwheeling: medium-dark skin tone"_s, u"woman cartwheeling: medium-dark skin tone"_s})},
{u"woman cartwheeling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F938\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman cartwheeling: dark skin tone"_s, u"woman cartwheeling: dark skin tone"_s})},
{u"person playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FB"), u"water_polo_tone1"_s, u"person playing water polo: light skin tone"_s})},
{u"person playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FC"), u"water_polo_tone2"_s, u"person playing water polo: medium-light skin tone"_s})},
{u"person playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FD"), u"water_polo_tone3"_s, u"person playing water polo: medium skin tone"_s})},
{u"person playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FE"), u"water_polo_tone4"_s, u"person playing water polo: medium-dark skin tone"_s})},
{u"person playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FF"), u"water_polo_tone5"_s, u"person playing water polo: dark skin tone"_s})},
{u"man playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man playing water polo: light skin tone"_s, u"man playing water polo: light skin tone"_s})},
{u"man playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man playing water polo: medium-light skin tone"_s, u"man playing water polo: medium-light skin tone"_s})},
{u"man playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man playing water polo: medium skin tone"_s, u"man playing water polo: medium skin tone"_s})},
{u"man playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man playing water polo: medium-dark skin tone"_s, u"man playing water polo: medium-dark skin tone"_s})},
{u"man playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man playing water polo: dark skin tone"_s, u"man playing water polo: dark skin tone"_s})},
{u"woman playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman playing water polo: light skin tone"_s, u"woman playing water polo: light skin tone"_s})},
{u"woman playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman playing water polo: medium-light skin tone"_s, u"woman playing water polo: medium-light skin tone"_s})},
{u"woman playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman playing water polo: medium skin tone"_s, u"woman playing water polo: medium skin tone"_s})},
{u"woman playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman playing water polo: medium-dark skin tone"_s, u"woman playing water polo: medium-dark skin tone"_s})},
{u"woman playing water polo"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93D\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman playing water polo: dark skin tone"_s, u"woman playing water polo: dark skin tone"_s})},
{u"person playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FB"), u"handball_tone1"_s, u"person playing handball: light skin tone"_s})},
{u"person playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FC"), u"handball_tone2"_s, u"person playing handball: medium-light skin tone"_s})},
{u"person playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FD"), u"handball_tone3"_s, u"person playing handball: medium skin tone"_s})},
{u"person playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FE"), u"handball_tone4"_s, u"person playing handball: medium-dark skin tone"_s})},
{u"person playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FF"), u"handball_tone5"_s, u"person playing handball: dark skin tone"_s})},
{u"man playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man playing handball: light skin tone"_s, u"man playing handball: light skin tone"_s})},
{u"man playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man playing handball: medium-light skin tone"_s, u"man playing handball: medium-light skin tone"_s})},
{u"man playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man playing handball: medium skin tone"_s, u"man playing handball: medium skin tone"_s})},
{u"man playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man playing handball: medium-dark skin tone"_s, u"man playing handball: medium-dark skin tone"_s})},
{u"man playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man playing handball: dark skin tone"_s, u"man playing handball: dark skin tone"_s})},
{u"woman playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman playing handball: light skin tone"_s, u"woman playing handball: light skin tone"_s})},
{u"woman playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman playing handball: medium-light skin tone"_s, u"woman playing handball: medium-light skin tone"_s})},
{u"woman playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman playing handball: medium skin tone"_s, u"woman playing handball: medium skin tone"_s})},
{u"woman playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman playing handball: medium-dark skin tone"_s, u"woman playing handball: medium-dark skin tone"_s})},
{u"woman playing handball"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F93E\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman playing handball: dark skin tone"_s, u"woman playing handball: dark skin tone"_s})},
{u"person juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FB"), u"juggling_tone1"_s, u"person juggling: light skin tone"_s})},
{u"person juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FC"), u"juggling_tone2"_s, u"person juggling: medium-light skin tone"_s})},
{u"person juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FD"), u"juggling_tone3"_s, u"person juggling: medium skin tone"_s})},
{u"person juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FE"), u"juggling_tone4"_s, u"person juggling: medium-dark skin tone"_s})},
{u"person juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FF"), u"juggling_tone5"_s, u"person juggling: dark skin tone"_s})},
{u"man juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man juggling: light skin tone"_s, u"man juggling: light skin tone"_s})},
{u"man juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man juggling: medium-light skin tone"_s, u"man juggling: medium-light skin tone"_s})},
{u"man juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man juggling: medium skin tone"_s, u"man juggling: medium skin tone"_s})},
{u"man juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man juggling: medium-dark skin tone"_s, u"man juggling: medium-dark skin tone"_s})},
{u"man juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man juggling: dark skin tone"_s, u"man juggling: dark skin tone"_s})},
{u"woman juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman juggling: light skin tone"_s, u"woman juggling: light skin tone"_s})},
{u"woman juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman juggling: medium-light skin tone"_s, u"woman juggling: medium-light skin tone"_s})},
{u"woman juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman juggling: medium skin tone"_s, u"woman juggling: medium skin tone"_s})},
{u"woman juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman juggling: medium-dark skin tone"_s, u"woman juggling: medium-dark skin tone"_s})},
{u"woman juggling"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F939\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman juggling: dark skin tone"_s, u"woman juggling: dark skin tone"_s})},
{u"person in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FB"), u"person_in_lotus_position_tone1"_s, u"person in lotus position: light skin tone"_s})},
{u"person in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FC"), u"person_in_lotus_position_tone2"_s, u"person in lotus position: medium-light skin tone"_s})},
{u"person in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FD"), u"person_in_lotus_position_tone3"_s, u"person in lotus position: medium skin tone"_s})},
{u"person in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FE"), u"person_in_lotus_position_tone4"_s, u"person in lotus position: medium-dark skin tone"_s})},
{u"person in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FF"), u"person_in_lotus_position_tone5"_s, u"person in lotus position: dark skin tone"_s})},
{u"man in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FB\U0000200D\U00002642\U0000FE0F"), u"man in lotus position: light skin tone"_s, u"man in lotus position: light skin tone"_s})},
{u"man in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FC\U0000200D\U00002642\U0000FE0F"), u"man in lotus position: medium-light skin tone"_s, u"man in lotus position: medium-light skin tone"_s})},
{u"man in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FD\U0000200D\U00002642\U0000FE0F"), u"man in lotus position: medium skin tone"_s, u"man in lotus position: medium skin tone"_s})},
{u"man in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FE\U0000200D\U00002642\U0000FE0F"), u"man in lotus position: medium-dark skin tone"_s, u"man in lotus position: medium-dark skin tone"_s})},
{u"man in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FF\U0000200D\U00002642\U0000FE0F"), u"man in lotus position: dark skin tone"_s, u"man in lotus position: dark skin tone"_s})},
{u"woman in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FB\U0000200D\U00002640\U0000FE0F"), u"woman in lotus position: light skin tone"_s, u"woman in lotus position: light skin tone"_s})},
{u"woman in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FC\U0000200D\U00002640\U0000FE0F"), u"woman in lotus position: medium-light skin tone"_s, u"woman in lotus position: medium-light skin tone"_s})},
{u"woman in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FD\U0000200D\U00002640\U0000FE0F"), u"woman in lotus position: medium skin tone"_s, u"woman in lotus position: medium skin tone"_s})},
{u"woman in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FE\U0000200D\U00002640\U0000FE0F"), u"woman in lotus position: medium-dark skin tone"_s, u"woman in lotus position: medium-dark skin tone"_s})},
{u"woman in lotus position"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D8\U0001F3FF\U0000200D\U00002640\U0000FE0F"), u"woman in lotus position: dark skin tone"_s, u"woman in lotus position: dark skin tone"_s})},
{u"person taking bath"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6C0\U0001F3FB"), u"bath_tone1"_s, u"person taking bath: light skin tone"_s})},
{u"person taking bath"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6C0\U0001F3FC"), u"bath_tone2"_s, u"person taking bath: medium-light skin tone"_s})},
{u"person taking bath"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6C0\U0001F3FD"), u"bath_tone3"_s, u"person taking bath: medium skin tone"_s})},
{u"person taking bath"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6C0\U0001F3FE"), u"bath_tone4"_s, u"person taking bath: medium-dark skin tone"_s})},
{u"person taking bath"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6C0\U0001F3FF"), u"bath_tone5"_s, u"person taking bath: dark skin tone"_s})},
{u"person in bed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6CC\U0001F3FB"), u"person_in_bed_tone1"_s, u"person in bed: light skin tone"_s})},
{u"person in bed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6CC\U0001F3FC"), u"person_in_bed_tone2"_s, u"person in bed: medium-light skin tone"_s})},
{u"person in bed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6CC\U0001F3FD"), u"person_in_bed_tone3"_s, u"person in bed: medium skin tone"_s})},
{u"person in bed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6CC\U0001F3FE"), u"person_in_bed_tone4"_s, u"person in bed: medium-dark skin tone"_s})},
{u"person in bed"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F6CC\U0001F3FF"), u"person_in_bed_tone5"_s, u"person in bed: dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"), u"people holding hands: light skin tone"_s, u"people holding hands: light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"), u"people holding hands: light skin tone, medium-light skin tone"_s, u"people holding hands: light skin tone, medium-light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"), u"people holding hands: light skin tone, medium skin tone"_s, u"people holding hands: light skin tone, medium skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"), u"people holding hands: light skin tone, medium-dark skin tone"_s, u"people holding hands: light skin tone, medium-dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"), u"people holding hands: light skin tone, dark skin tone"_s, u"people holding hands: light skin tone, dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"), u"people holding hands: medium-light skin tone, light skin tone"_s, u"people holding hands: medium-light skin tone, light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"), u"people holding hands: medium-light skin tone"_s, u"people holding hands: medium-light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"), u"people holding hands: medium-light skin tone, medium skin tone"_s, u"people holding hands: medium-light skin tone, medium skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"), u"people holding hands: medium-light skin tone, medium-dark skin tone"_s, u"people holding hands: medium-light skin tone, medium-dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"), u"people holding hands: medium-light skin tone, dark skin tone"_s, u"people holding hands: medium-light skin tone, dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"), u"people holding hands: medium skin tone, light skin tone"_s, u"people holding hands: medium skin tone, light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"), u"people holding hands: medium skin tone, medium-light skin tone"_s, u"people holding hands: medium skin tone, medium-light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"), u"people holding hands: medium skin tone"_s, u"people holding hands: medium skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"), u"people holding hands: medium skin tone, medium-dark skin tone"_s, u"people holding hands: medium skin tone, medium-dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"), u"people holding hands: medium skin tone, dark skin tone"_s, u"people holding hands: medium skin tone, dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"), u"people holding hands: medium-dark skin tone, light skin tone"_s, u"people holding hands: medium-dark skin tone, light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"), u"people holding hands: medium-dark skin tone, medium-light skin tone"_s, u"people holding hands: medium-dark skin tone, medium-light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"), u"people holding hands: medium-dark skin tone, medium skin tone"_s, u"people holding hands: medium-dark skin tone, medium skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"), u"people holding hands: medium-dark skin tone"_s, u"people holding hands: medium-dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"), u"people holding hands: medium-dark skin tone, dark skin tone"_s, u"people holding hands: medium-dark skin tone, dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FB"), u"people holding hands: dark skin tone, light skin tone"_s, u"people holding hands: dark skin tone, light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FC"), u"people holding hands: dark skin tone, medium-light skin tone"_s, u"people holding hands: dark skin tone, medium-light skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FD"), u"people holding hands: dark skin tone, medium skin tone"_s, u"people holding hands: dark skin tone, medium skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FE"), u"people holding hands: dark skin tone, medium-dark skin tone"_s, u"people holding hands: dark skin tone, medium-dark skin tone"_s})},
{u"people holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F9D1\U0001F3FF"), u"people holding hands: dark skin tone"_s, u"people holding hands: dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46D\U0001F3FB"), u"women holding hands: light skin tone"_s, u"women holding hands: light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"), u"women holding hands: light skin tone, medium-light skin tone"_s, u"women holding hands: light skin tone, medium-light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"), u"women holding hands: light skin tone, medium skin tone"_s, u"women holding hands: light skin tone, medium skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"), u"women holding hands: light skin tone, medium-dark skin tone"_s, u"women holding hands: light skin tone, medium-dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"), u"women holding hands: light skin tone, dark skin tone"_s, u"women holding hands: light skin tone, dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"), u"women holding hands: medium-light skin tone, light skin tone"_s, u"women holding hands: medium-light skin tone, light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46D\U0001F3FC"), u"women holding hands: medium-light skin tone"_s, u"women holding hands: medium-light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"), u"women holding hands: medium-light skin tone, medium skin tone"_s, u"women holding hands: medium-light skin tone, medium skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"), u"women holding hands: medium-light skin tone, medium-dark skin tone"_s, u"women holding hands: medium-light skin tone, medium-dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"), u"women holding hands: medium-light skin tone, dark skin tone"_s, u"women holding hands: medium-light skin tone, dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"), u"women holding hands: medium skin tone, light skin tone"_s, u"women holding hands: medium skin tone, light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"), u"women holding hands: medium skin tone, medium-light skin tone"_s, u"women holding hands: medium skin tone, medium-light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46D\U0001F3FD"), u"women holding hands: medium skin tone"_s, u"women holding hands: medium skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"), u"women holding hands: medium skin tone, medium-dark skin tone"_s, u"women holding hands: medium skin tone, medium-dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"), u"women holding hands: medium skin tone, dark skin tone"_s, u"women holding hands: medium skin tone, dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"), u"women holding hands: medium-dark skin tone, light skin tone"_s, u"women holding hands: medium-dark skin tone, light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"), u"women holding hands: medium-dark skin tone, medium-light skin tone"_s, u"women holding hands: medium-dark skin tone, medium-light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"), u"women holding hands: medium-dark skin tone, medium skin tone"_s, u"women holding hands: medium-dark skin tone, medium skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46D\U0001F3FE"), u"women holding hands: medium-dark skin tone"_s, u"women holding hands: medium-dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FF"), u"women holding hands: medium-dark skin tone, dark skin tone"_s, u"women holding hands: medium-dark skin tone, dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FB"), u"women holding hands: dark skin tone, light skin tone"_s, u"women holding hands: dark skin tone, light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FC"), u"women holding hands: dark skin tone, medium-light skin tone"_s, u"women holding hands: dark skin tone, medium-light skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FD"), u"women holding hands: dark skin tone, medium skin tone"_s, u"women holding hands: dark skin tone, medium skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F469\U0001F3FE"), u"women holding hands: dark skin tone, medium-dark skin tone"_s, u"women holding hands: dark skin tone, medium-dark skin tone"_s})},
{u"women holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46D\U0001F3FF"), u"women holding hands: dark skin tone"_s, u"women holding hands: dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46B\U0001F3FB"), u"woman and man holding hands: light skin tone"_s, u"woman and man holding hands: light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"woman and man holding hands: light skin tone, medium-light skin tone"_s, u"woman and man holding hands: light skin tone, medium-light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"woman and man holding hands: light skin tone, medium skin tone"_s, u"woman and man holding hands: light skin tone, medium skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"woman and man holding hands: light skin tone, medium-dark skin tone"_s, u"woman and man holding hands: light skin tone, medium-dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"woman and man holding hands: light skin tone, dark skin tone"_s, u"woman and man holding hands: light skin tone, dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"woman and man holding hands: medium-light skin tone, light skin tone"_s, u"woman and man holding hands: medium-light skin tone, light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46B\U0001F3FC"), u"woman and man holding hands: medium-light skin tone"_s, u"woman and man holding hands: medium-light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"woman and man holding hands: medium-light skin tone, medium skin tone"_s, u"woman and man holding hands: medium-light skin tone, medium skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"woman and man holding hands: medium-light skin tone, medium-dark skin tone"_s, u"woman and man holding hands: medium-light skin tone, medium-dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"woman and man holding hands: medium-light skin tone, dark skin tone"_s, u"woman and man holding hands: medium-light skin tone, dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"woman and man holding hands: medium skin tone, light skin tone"_s, u"woman and man holding hands: medium skin tone, light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"woman and man holding hands: medium skin tone, medium-light skin tone"_s, u"woman and man holding hands: medium skin tone, medium-light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46B\U0001F3FD"), u"woman and man holding hands: medium skin tone"_s, u"woman and man holding hands: medium skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"woman and man holding hands: medium skin tone, medium-dark skin tone"_s, u"woman and man holding hands: medium skin tone, medium-dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"woman and man holding hands: medium skin tone, dark skin tone"_s, u"woman and man holding hands: medium skin tone, dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"woman and man holding hands: medium-dark skin tone, light skin tone"_s, u"woman and man holding hands: medium-dark skin tone, light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"woman and man holding hands: medium-dark skin tone, medium-light skin tone"_s, u"woman and man holding hands: medium-dark skin tone, medium-light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"woman and man holding hands: medium-dark skin tone, medium skin tone"_s, u"woman and man holding hands: medium-dark skin tone, medium skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46B\U0001F3FE"), u"woman and man holding hands: medium-dark skin tone"_s, u"woman and man holding hands: medium-dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"woman and man holding hands: medium-dark skin tone, dark skin tone"_s, u"woman and man holding hands: medium-dark skin tone, dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"woman and man holding hands: dark skin tone, light skin tone"_s, u"woman and man holding hands: dark skin tone, light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"woman and man holding hands: dark skin tone, medium-light skin tone"_s, u"woman and man holding hands: dark skin tone, medium-light skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"woman and man holding hands: dark skin tone, medium skin tone"_s, u"woman and man holding hands: dark skin tone, medium skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"woman and man holding hands: dark skin tone, medium-dark skin tone"_s, u"woman and man holding hands: dark skin tone, medium-dark skin tone"_s})},
{u"woman and man holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46B\U0001F3FF"), u"woman and man holding hands: dark skin tone"_s, u"woman and man holding hands: dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46C\U0001F3FB"), u"men holding hands: light skin tone"_s, u"men holding hands: light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"men holding hands: light skin tone, medium-light skin tone"_s, u"men holding hands: light skin tone, medium-light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"men holding hands: light skin tone, medium skin tone"_s, u"men holding hands: light skin tone, medium skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"men holding hands: light skin tone, medium-dark skin tone"_s, u"men holding hands: light skin tone, medium-dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"men holding hands: light skin tone, dark skin tone"_s, u"men holding hands: light skin tone, dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"men holding hands: medium-light skin tone, light skin tone"_s, u"men holding hands: medium-light skin tone, light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46C\U0001F3FC"), u"men holding hands: medium-light skin tone"_s, u"men holding hands: medium-light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"men holding hands: medium-light skin tone, medium skin tone"_s, u"men holding hands: medium-light skin tone, medium skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"men holding hands: medium-light skin tone, medium-dark skin tone"_s, u"men holding hands: medium-light skin tone, medium-dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"men holding hands: medium-light skin tone, dark skin tone"_s, u"men holding hands: medium-light skin tone, dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"men holding hands: medium skin tone, light skin tone"_s, u"men holding hands: medium skin tone, light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"men holding hands: medium skin tone, medium-light skin tone"_s, u"men holding hands: medium skin tone, medium-light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46C\U0001F3FD"), u"men holding hands: medium skin tone"_s, u"men holding hands: medium skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"men holding hands: medium skin tone, medium-dark skin tone"_s, u"men holding hands: medium skin tone, medium-dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"men holding hands: medium skin tone, dark skin tone"_s, u"men holding hands: medium skin tone, dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"men holding hands: medium-dark skin tone, light skin tone"_s, u"men holding hands: medium-dark skin tone, light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"men holding hands: medium-dark skin tone, medium-light skin tone"_s, u"men holding hands: medium-dark skin tone, medium-light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"men holding hands: medium-dark skin tone, medium skin tone"_s, u"men holding hands: medium-dark skin tone, medium skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46C\U0001F3FE"), u"men holding hands: medium-dark skin tone"_s, u"men holding hands: medium-dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FF"), u"men holding hands: medium-dark skin tone, dark skin tone"_s, u"men holding hands: medium-dark skin tone, dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FB"), u"men holding hands: dark skin tone, light skin tone"_s, u"men holding hands: dark skin tone, light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FC"), u"men holding hands: dark skin tone, medium-light skin tone"_s, u"men holding hands: dark skin tone, medium-light skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FD"), u"men holding hands: dark skin tone, medium skin tone"_s, u"men holding hands: dark skin tone, medium skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U0001F91D\U0000200D\U0001F468\U0001F3FE"), u"men holding hands: dark skin tone, medium-dark skin tone"_s, u"men holding hands: dark skin tone, medium-dark skin tone"_s})},
{u"men holding hands"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F46C\U0001F3FF"), u"men holding hands: dark skin tone"_s, u"men holding hands: dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F48F\U0001F3FB"), u"kiss: light skin tone"_s, u"kiss: light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F48F\U0001F3FC"), u"kiss: medium-light skin tone"_s, u"kiss: medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F48F\U0001F3FD"), u"kiss: medium skin tone"_s, u"kiss: medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F48F\U0001F3FE"), u"kiss: medium-dark skin tone"_s, u"kiss: medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F48F\U0001F3FF"), u"kiss: dark skin tone"_s, u"kiss: dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"), u"kiss: person, person, light skin tone, medium-light skin tone"_s, u"kiss: person, person, light skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"), u"kiss: person, person, light skin tone, medium skin tone"_s, u"kiss: person, person, light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"), u"kiss: person, person, light skin tone, medium-dark skin tone"_s, u"kiss: person, person, light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"), u"kiss: person, person, light skin tone, dark skin tone"_s, u"kiss: person, person, light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"), u"kiss: person, person, medium-light skin tone, light skin tone"_s, u"kiss: person, person, medium-light skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"), u"kiss: person, person, medium-light skin tone, medium skin tone"_s, u"kiss: person, person, medium-light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"), u"kiss: person, person, medium-light skin tone, medium-dark skin tone"_s, u"kiss: person, person, medium-light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"), u"kiss: person, person, medium-light skin tone, dark skin tone"_s, u"kiss: person, person, medium-light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"), u"kiss: person, person, medium skin tone, light skin tone"_s, u"kiss: person, person, medium skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"), u"kiss: person, person, medium skin tone, medium-light skin tone"_s, u"kiss: person, person, medium skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"), u"kiss: person, person, medium skin tone, medium-dark skin tone"_s, u"kiss: person, person, medium skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"), u"kiss: person, person, medium skin tone, dark skin tone"_s, u"kiss: person, person, medium skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"), u"kiss: person, person, medium-dark skin tone, light skin tone"_s, u"kiss: person, person, medium-dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"), u"kiss: person, person, medium-dark skin tone, medium-light skin tone"_s, u"kiss: person, person, medium-dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"), u"kiss: person, person, medium-dark skin tone, medium skin tone"_s, u"kiss: person, person, medium-dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FF"), u"kiss: person, person, medium-dark skin tone, dark skin tone"_s, u"kiss: person, person, medium-dark skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FB"), u"kiss: person, person, dark skin tone, light skin tone"_s, u"kiss: person, person, dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FC"), u"kiss: person, person, dark skin tone, medium-light skin tone"_s, u"kiss: person, person, dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FD"), u"kiss: person, person, dark skin tone, medium skin tone"_s, u"kiss: person, person, dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F9D1\U0001F3FE"), u"kiss: person, person, dark skin tone, medium-dark skin tone"_s, u"kiss: person, person, dark skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: woman, man, light skin tone"_s, u"kiss: woman, man, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: woman, man, light skin tone, medium-light skin tone"_s, u"kiss: woman, man, light skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: woman, man, light skin tone, medium skin tone"_s, u"kiss: woman, man, light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: woman, man, light skin tone, medium-dark skin tone"_s, u"kiss: woman, man, light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: woman, man, light skin tone, dark skin tone"_s, u"kiss: woman, man, light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: woman, man, medium-light skin tone, light skin tone"_s, u"kiss: woman, man, medium-light skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: woman, man, medium-light skin tone"_s, u"kiss: woman, man, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: woman, man, medium-light skin tone, medium skin tone"_s, u"kiss: woman, man, medium-light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: woman, man, medium-light skin tone, medium-dark skin tone"_s, u"kiss: woman, man, medium-light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: woman, man, medium-light skin tone, dark skin tone"_s, u"kiss: woman, man, medium-light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: woman, man, medium skin tone, light skin tone"_s, u"kiss: woman, man, medium skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: woman, man, medium skin tone, medium-light skin tone"_s, u"kiss: woman, man, medium skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: woman, man, medium skin tone"_s, u"kiss: woman, man, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: woman, man, medium skin tone, medium-dark skin tone"_s, u"kiss: woman, man, medium skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: woman, man, medium skin tone, dark skin tone"_s, u"kiss: woman, man, medium skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: woman, man, medium-dark skin tone, light skin tone"_s, u"kiss: woman, man, medium-dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: woman, man, medium-dark skin tone, medium-light skin tone"_s, u"kiss: woman, man, medium-dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: woman, man, medium-dark skin tone, medium skin tone"_s, u"kiss: woman, man, medium-dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: woman, man, medium-dark skin tone"_s, u"kiss: woman, man, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: woman, man, medium-dark skin tone, dark skin tone"_s, u"kiss: woman, man, medium-dark skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: woman, man, dark skin tone, light skin tone"_s, u"kiss: woman, man, dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: woman, man, dark skin tone, medium-light skin tone"_s, u"kiss: woman, man, dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: woman, man, dark skin tone, medium skin tone"_s, u"kiss: woman, man, dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: woman, man, dark skin tone, medium-dark skin tone"_s, u"kiss: woman, man, dark skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: woman, man, dark skin tone"_s, u"kiss: woman, man, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: man, man, light skin tone"_s, u"kiss: man, man, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: man, man, light skin tone, medium-light skin tone"_s, u"kiss: man, man, light skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: man, man, light skin tone, medium skin tone"_s, u"kiss: man, man, light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: man, man, light skin tone, medium-dark skin tone"_s, u"kiss: man, man, light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: man, man, light skin tone, dark skin tone"_s, u"kiss: man, man, light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: man, man, medium-light skin tone, light skin tone"_s, u"kiss: man, man, medium-light skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: man, man, medium-light skin tone"_s, u"kiss: man, man, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: man, man, medium-light skin tone, medium skin tone"_s, u"kiss: man, man, medium-light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: man, man, medium-light skin tone, medium-dark skin tone"_s, u"kiss: man, man, medium-light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: man, man, medium-light skin tone, dark skin tone"_s, u"kiss: man, man, medium-light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: man, man, medium skin tone, light skin tone"_s, u"kiss: man, man, medium skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: man, man, medium skin tone, medium-light skin tone"_s, u"kiss: man, man, medium skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: man, man, medium skin tone"_s, u"kiss: man, man, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: man, man, medium skin tone, medium-dark skin tone"_s, u"kiss: man, man, medium skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: man, man, medium skin tone, dark skin tone"_s, u"kiss: man, man, medium skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: man, man, medium-dark skin tone, light skin tone"_s, u"kiss: man, man, medium-dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: man, man, medium-dark skin tone, medium-light skin tone"_s, u"kiss: man, man, medium-dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: man, man, medium-dark skin tone, medium skin tone"_s, u"kiss: man, man, medium-dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: man, man, medium-dark skin tone"_s, u"kiss: man, man, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: man, man, medium-dark skin tone, dark skin tone"_s, u"kiss: man, man, medium-dark skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FB"), u"kiss: man, man, dark skin tone, light skin tone"_s, u"kiss: man, man, dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FC"), u"kiss: man, man, dark skin tone, medium-light skin tone"_s, u"kiss: man, man, dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FD"), u"kiss: man, man, dark skin tone, medium skin tone"_s, u"kiss: man, man, dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FE"), u"kiss: man, man, dark skin tone, medium-dark skin tone"_s, u"kiss: man, man, dark skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F468\U0001F3FF"), u"kiss: man, man, dark skin tone"_s, u"kiss: man, man, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FB"), u"kiss: woman, woman, light skin tone"_s, u"kiss: woman, woman, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FC"), u"kiss: woman, woman, light skin tone, medium-light skin tone"_s, u"kiss: woman, woman, light skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FD"), u"kiss: woman, woman, light skin tone, medium skin tone"_s, u"kiss: woman, woman, light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FE"), u"kiss: woman, woman, light skin tone, medium-dark skin tone"_s, u"kiss: woman, woman, light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FF"), u"kiss: woman, woman, light skin tone, dark skin tone"_s, u"kiss: woman, woman, light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FB"), u"kiss: woman, woman, medium-light skin tone, light skin tone"_s, u"kiss: woman, woman, medium-light skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FC"), u"kiss: woman, woman, medium-light skin tone"_s, u"kiss: woman, woman, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FD"), u"kiss: woman, woman, medium-light skin tone, medium skin tone"_s, u"kiss: woman, woman, medium-light skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FE"), u"kiss: woman, woman, medium-light skin tone, medium-dark skin tone"_s, u"kiss: woman, woman, medium-light skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FF"), u"kiss: woman, woman, medium-light skin tone, dark skin tone"_s, u"kiss: woman, woman, medium-light skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FB"), u"kiss: woman, woman, medium skin tone, light skin tone"_s, u"kiss: woman, woman, medium skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FC"), u"kiss: woman, woman, medium skin tone, medium-light skin tone"_s, u"kiss: woman, woman, medium skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FD"), u"kiss: woman, woman, medium skin tone"_s, u"kiss: woman, woman, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FE"), u"kiss: woman, woman, medium skin tone, medium-dark skin tone"_s, u"kiss: woman, woman, medium skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FF"), u"kiss: woman, woman, medium skin tone, dark skin tone"_s, u"kiss: woman, woman, medium skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FB"), u"kiss: woman, woman, medium-dark skin tone, light skin tone"_s, u"kiss: woman, woman, medium-dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FC"), u"kiss: woman, woman, medium-dark skin tone, medium-light skin tone"_s, u"kiss: woman, woman, medium-dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FD"), u"kiss: woman, woman, medium-dark skin tone, medium skin tone"_s, u"kiss: woman, woman, medium-dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FE"), u"kiss: woman, woman, medium-dark skin tone"_s, u"kiss: woman, woman, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FF"), u"kiss: woman, woman, medium-dark skin tone, dark skin tone"_s, u"kiss: woman, woman, medium-dark skin tone, dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FB"), u"kiss: woman, woman, dark skin tone, light skin tone"_s, u"kiss: woman, woman, dark skin tone, light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FC"), u"kiss: woman, woman, dark skin tone, medium-light skin tone"_s, u"kiss: woman, woman, dark skin tone, medium-light skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FD"), u"kiss: woman, woman, dark skin tone, medium skin tone"_s, u"kiss: woman, woman, dark skin tone, medium skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FE"), u"kiss: woman, woman, dark skin tone, medium-dark skin tone"_s, u"kiss: woman, woman, dark skin tone, medium-dark skin tone"_s})},
{u"kiss"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F48B\U0000200D\U0001F469\U0001F3FF"), u"kiss: woman, woman, dark skin tone"_s, u"kiss: woman, woman, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F491\U0001F3FB"), u"couple with heart: light skin tone"_s, u"couple with heart: light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F491\U0001F3FC"), u"couple with heart: medium-light skin tone"_s, u"couple with heart: medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F491\U0001F3FD"), u"couple with heart: medium skin tone"_s, u"couple with heart: medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F491\U0001F3FE"), u"couple with heart: medium-dark skin tone"_s, u"couple with heart: medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F491\U0001F3FF"), u"couple with heart: dark skin tone"_s, u"couple with heart: dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"), u"couple with heart: person, person, light skin tone, medium-light skin tone"_s, u"couple with heart: person, person, light skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"), u"couple with heart: person, person, light skin tone, medium skin tone"_s, u"couple with heart: person, person, light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"), u"couple with heart: person, person, light skin tone, medium-dark skin tone"_s, u"couple with heart: person, person, light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"), u"couple with heart: person, person, light skin tone, dark skin tone"_s, u"couple with heart: person, person, light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"), u"couple with heart: person, person, medium-light skin tone, light skin tone"_s, u"couple with heart: person, person, medium-light skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"), u"couple with heart: person, person, medium-light skin tone, medium skin tone"_s, u"couple with heart: person, person, medium-light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"), u"couple with heart: person, person, medium-light skin tone, medium-dark skin tone"_s, u"couple with heart: person, person, medium-light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"), u"couple with heart: person, person, medium-light skin tone, dark skin tone"_s, u"couple with heart: person, person, medium-light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"), u"couple with heart: person, person, medium skin tone, light skin tone"_s, u"couple with heart: person, person, medium skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"), u"couple with heart: person, person, medium skin tone, medium-light skin tone"_s, u"couple with heart: person, person, medium skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"), u"couple with heart: person, person, medium skin tone, medium-dark skin tone"_s, u"couple with heart: person, person, medium skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"), u"couple with heart: person, person, medium skin tone, dark skin tone"_s, u"couple with heart: person, person, medium skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"), u"couple with heart: person, person, medium-dark skin tone, light skin tone"_s, u"couple with heart: person, person, medium-dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"), u"couple with heart: person, person, medium-dark skin tone, medium-light skin tone"_s, u"couple with heart: person, person, medium-dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"), u"couple with heart: person, person, medium-dark skin tone, medium skin tone"_s, u"couple with heart: person, person, medium-dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FF"), u"couple with heart: person, person, medium-dark skin tone, dark skin tone"_s, u"couple with heart: person, person, medium-dark skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FB"), u"couple with heart: person, person, dark skin tone, light skin tone"_s, u"couple with heart: person, person, dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FC"), u"couple with heart: person, person, dark skin tone, medium-light skin tone"_s, u"couple with heart: person, person, dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FD"), u"couple with heart: person, person, dark skin tone, medium skin tone"_s, u"couple with heart: person, person, dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F9D1\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F9D1\U0001F3FE"), u"couple with heart: person, person, dark skin tone, medium-dark skin tone"_s, u"couple with heart: person, person, dark skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: woman, man, light skin tone"_s, u"couple with heart: woman, man, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: woman, man, light skin tone, medium-light skin tone"_s, u"couple with heart: woman, man, light skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: woman, man, light skin tone, medium skin tone"_s, u"couple with heart: woman, man, light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: woman, man, light skin tone, medium-dark skin tone"_s, u"couple with heart: woman, man, light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: woman, man, light skin tone, dark skin tone"_s, u"couple with heart: woman, man, light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: woman, man, medium-light skin tone, light skin tone"_s, u"couple with heart: woman, man, medium-light skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: woman, man, medium-light skin tone"_s, u"couple with heart: woman, man, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: woman, man, medium-light skin tone, medium skin tone"_s, u"couple with heart: woman, man, medium-light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: woman, man, medium-light skin tone, medium-dark skin tone"_s, u"couple with heart: woman, man, medium-light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: woman, man, medium-light skin tone, dark skin tone"_s, u"couple with heart: woman, man, medium-light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: woman, man, medium skin tone, light skin tone"_s, u"couple with heart: woman, man, medium skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: woman, man, medium skin tone, medium-light skin tone"_s, u"couple with heart: woman, man, medium skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: woman, man, medium skin tone"_s, u"couple with heart: woman, man, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: woman, man, medium skin tone, medium-dark skin tone"_s, u"couple with heart: woman, man, medium skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: woman, man, medium skin tone, dark skin tone"_s, u"couple with heart: woman, man, medium skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: woman, man, medium-dark skin tone, light skin tone"_s, u"couple with heart: woman, man, medium-dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: woman, man, medium-dark skin tone, medium-light skin tone"_s, u"couple with heart: woman, man, medium-dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: woman, man, medium-dark skin tone, medium skin tone"_s, u"couple with heart: woman, man, medium-dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: woman, man, medium-dark skin tone"_s, u"couple with heart: woman, man, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: woman, man, medium-dark skin tone, dark skin tone"_s, u"couple with heart: woman, man, medium-dark skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: woman, man, dark skin tone, light skin tone"_s, u"couple with heart: woman, man, dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: woman, man, dark skin tone, medium-light skin tone"_s, u"couple with heart: woman, man, dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: woman, man, dark skin tone, medium skin tone"_s, u"couple with heart: woman, man, dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: woman, man, dark skin tone, medium-dark skin tone"_s, u"couple with heart: woman, man, dark skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: woman, man, dark skin tone"_s, u"couple with heart: woman, man, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: man, man, light skin tone"_s, u"couple with heart: man, man, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: man, man, light skin tone, medium-light skin tone"_s, u"couple with heart: man, man, light skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: man, man, light skin tone, medium skin tone"_s, u"couple with heart: man, man, light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: man, man, light skin tone, medium-dark skin tone"_s, u"couple with heart: man, man, light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: man, man, light skin tone, dark skin tone"_s, u"couple with heart: man, man, light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: man, man, medium-light skin tone, light skin tone"_s, u"couple with heart: man, man, medium-light skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: man, man, medium-light skin tone"_s, u"couple with heart: man, man, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: man, man, medium-light skin tone, medium skin tone"_s, u"couple with heart: man, man, medium-light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: man, man, medium-light skin tone, medium-dark skin tone"_s, u"couple with heart: man, man, medium-light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: man, man, medium-light skin tone, dark skin tone"_s, u"couple with heart: man, man, medium-light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: man, man, medium skin tone, light skin tone"_s, u"couple with heart: man, man, medium skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: man, man, medium skin tone, medium-light skin tone"_s, u"couple with heart: man, man, medium skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: man, man, medium skin tone"_s, u"couple with heart: man, man, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: man, man, medium skin tone, medium-dark skin tone"_s, u"couple with heart: man, man, medium skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: man, man, medium skin tone, dark skin tone"_s, u"couple with heart: man, man, medium skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: man, man, medium-dark skin tone, light skin tone"_s, u"couple with heart: man, man, medium-dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: man, man, medium-dark skin tone, medium-light skin tone"_s, u"couple with heart: man, man, medium-dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: man, man, medium-dark skin tone, medium skin tone"_s, u"couple with heart: man, man, medium-dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: man, man, medium-dark skin tone"_s, u"couple with heart: man, man, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: man, man, medium-dark skin tone, dark skin tone"_s, u"couple with heart: man, man, medium-dark skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FB"), u"couple with heart: man, man, dark skin tone, light skin tone"_s, u"couple with heart: man, man, dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FC"), u"couple with heart: man, man, dark skin tone, medium-light skin tone"_s, u"couple with heart: man, man, dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FD"), u"couple with heart: man, man, dark skin tone, medium skin tone"_s, u"couple with heart: man, man, dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FE"), u"couple with heart: man, man, dark skin tone, medium-dark skin tone"_s, u"couple with heart: man, man, dark skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F468\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F468\U0001F3FF"), u"couple with heart: man, man, dark skin tone"_s, u"couple with heart: man, man, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"), u"couple with heart: woman, woman, light skin tone"_s, u"couple with heart: woman, woman, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"), u"couple with heart: woman, woman, light skin tone, medium-light skin tone"_s, u"couple with heart: woman, woman, light skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"), u"couple with heart: woman, woman, light skin tone, medium skin tone"_s, u"couple with heart: woman, woman, light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"), u"couple with heart: woman, woman, light skin tone, medium-dark skin tone"_s, u"couple with heart: woman, woman, light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FB\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"), u"couple with heart: woman, woman, light skin tone, dark skin tone"_s, u"couple with heart: woman, woman, light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"), u"couple with heart: woman, woman, medium-light skin tone, light skin tone"_s, u"couple with heart: woman, woman, medium-light skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"), u"couple with heart: woman, woman, medium-light skin tone"_s, u"couple with heart: woman, woman, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"), u"couple with heart: woman, woman, medium-light skin tone, medium skin tone"_s, u"couple with heart: woman, woman, medium-light skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"), u"couple with heart: woman, woman, medium-light skin tone, medium-dark skin tone"_s, u"couple with heart: woman, woman, medium-light skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FC\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"), u"couple with heart: woman, woman, medium-light skin tone, dark skin tone"_s, u"couple with heart: woman, woman, medium-light skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"), u"couple with heart: woman, woman, medium skin tone, light skin tone"_s, u"couple with heart: woman, woman, medium skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"), u"couple with heart: woman, woman, medium skin tone, medium-light skin tone"_s, u"couple with heart: woman, woman, medium skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"), u"couple with heart: woman, woman, medium skin tone"_s, u"couple with heart: woman, woman, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"), u"couple with heart: woman, woman, medium skin tone, medium-dark skin tone"_s, u"couple with heart: woman, woman, medium skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FD\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"), u"couple with heart: woman, woman, medium skin tone, dark skin tone"_s, u"couple with heart: woman, woman, medium skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"), u"couple with heart: woman, woman, medium-dark skin tone, light skin tone"_s, u"couple with heart: woman, woman, medium-dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"), u"couple with heart: woman, woman, medium-dark skin tone, medium-light skin tone"_s, u"couple with heart: woman, woman, medium-dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"), u"couple with heart: woman, woman, medium-dark skin tone, medium skin tone"_s, u"couple with heart: woman, woman, medium-dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"), u"couple with heart: woman, woman, medium-dark skin tone"_s, u"couple with heart: woman, woman, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FE\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"), u"couple with heart: woman, woman, medium-dark skin tone, dark skin tone"_s, u"couple with heart: woman, woman, medium-dark skin tone, dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FB"), u"couple with heart: woman, woman, dark skin tone, light skin tone"_s, u"couple with heart: woman, woman, dark skin tone, light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FC"), u"couple with heart: woman, woman, dark skin tone, medium-light skin tone"_s, u"couple with heart: woman, woman, dark skin tone, medium-light skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FD"), u"couple with heart: woman, woman, dark skin tone, medium skin tone"_s, u"couple with heart: woman, woman, dark skin tone, medium skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FE"), u"couple with heart: woman, woman, dark skin tone, medium-dark skin tone"_s, u"couple with heart: woman, woman, dark skin tone, medium-dark skin tone"_s})},
{u"couple with heart"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F469\U0001F3FF\U0000200D\U00002764\U0000FE0F\U0000200D\U0001F469\U0001F3FF"), u"couple with heart: woman, woman, dark skin tone"_s, u"couple with heart: woman, woman, dark skin tone"_s})},
{u"light skin tone"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3FB"), u"tone1"_s, u"light skin tone"_s})},
{u"medium-light skin tone"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3FC"), u"tone2"_s, u"medium-light skin tone"_s})},
{u"medium skin tone"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3FD"), u"tone3"_s, u"medium skin tone"_s})},
{u"medium-dark skin tone"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3FE"), u"tone4"_s, u"medium-dark skin tone"_s})},
{u"dark skin tone"_s, QVariant::fromValue(Emoji{QString::fromUtf8("\U0001F3FF"), u"tone5"_s, u"dark skin tone"_s})},
